#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "include.h"

/* This file contains subroutines for generating bases of states.  */

/*  The variable MAXDONE is the maximum number of different possible
    bases that can be constructed.                                  */

#define MAXDONE 10000

/*  Some of the bases can be stored in static memory.
    MAX_BASE gives the number of partype set aside for 
    this purpose.  Any overflow is stored in malloc'ed 
    memory.  */

#ifndef MAX_BASE
#define MAX_BASE 1000000
#endif

#define PRINTIT 0 /*print basis size & timing information in makebasis */

/*  initbasis(x,ht,npx,kt,cyclic,o,multi) returns the basis x
    for a given o = -1,+1, ht[i]=integer, i=0 to i<HINDEX is winding
    number for the ith dimension.  kt is 2*K, and npx is the 
    particle number truncation.  multi is the desired lattice 
    subgroup of symmetries. o=0 implies no orientation reversal symmetry.
    if kt=0, for a basis with helicity only. For the first entry,
    nsectors must be set to zero.

    if kt<0 then abs(kt) is an upper limit for the momenta.
    
    cyclic is a flag indicating if a string is closed (TRUE)
    or open (FALSE).

    The basis should be initialized to x.nsectors=0
*/

void initbasis(sectors *x,int *ht, int npx, int kt, 
	       int cyclic, int o, int multi){
  int i;
  for(i=0; i<HINDEX; i++)x->ht[i]=ht[i];
  x->npmax=npx; x->kt=kt; x->cyclic=cyclic; 
  x->o=o; x->multi=multi;
}

/* Initialize array pointers, clearing any previously stored values.
   Deallocation is in reverse order of deallocation.
*/

void freebasis(sectors *x){
  int i;
  if(x->nsectors>0){
    for(i=x->nsectors-1; i>=0; i--)
      free(x->innerno[i]);
    for(i=x->nsectors-1; i>=0; i--){
      free(x->zeros[i]);
      free(x->inner[i]);
    }
  }
  x->nsectors=0;
}

/* makebasis(x) finds the inner product list and the actual
     basis of states.  */

void makebasis(sectors *x){
  int npmin,np, i,in;
  size_t ist;
  partype *temp;
#if PRINTIT
  clock_t t1=clock();
#endif

  /* Complain if a basis has already been constructed. */
  if(x->nsectors>0){
    fprintf(stderr,"Basis already filled in makebasis\n");
    goto error;
  }  
  
  if(x->cyclic && x->kt<0){
    fprintf(stderr,"Bases with cyclic=%i and kt=%i not yet"
	    " implimented, exiting\n",x->cyclic,x->kt);
    goto error;
  }

  /* find the the minumum number of possible particles */
  for(i=0, npmin=0; i<HINDEX; i++) npmin += abs(x->ht[i]);
  if(HSHIFT>0 && x->kt>0 && npmin == 0)npmin=2;
  /* printf("minimum number of particles np=%i\n",np); */
  /* loop through the bases and find the inner products of states
     and don't count zero norm states in x->length               */
  x->length=0;
  /* We allocate all of the basis memory first */
  for(i=0, np=npmin; np<=x->npmax; np+=2, i++)
    x->basis[i] = mahk(x->ht,np,x->kt,x->cyclic,abs(x->o),
		       subgroup(x->multi));
  for(i=0, np=npmin; np<=x->npmax; np+=2, i++){
    x->np[i] = np;
    x->inner[i]=malloc(x->basis[i]->length*sizeof(unsigned char));
    x->zeros[i]=malloc(x->basis[i]->length*sizeof(size_t));
    x->nzeros[i] = 0;
    x->innerno[i]=NULL;
    if(NULL == x->inner[i] || NULL == x->zeros[i]){
      fprintf(stderr,"Not enough memory in makebasis for inner"
	     " product, exiting.  length=%i\n",x->basis[i]->length);
      goto error;
    }
#if PRINTIT 
    printf("     length: %i for %i particles \n",(int) x->basis[i]->length,np);
#endif
    for(ist=0, temp=x->basis[i]->locate; ist<x->basis[i]->length; 
	temp+=np, ist++){
      in=innerproduct(temp,temp,np,x->cyclic,x->o,x->multi);
      if(in >= (1 << (8*sizeof(x->inner[0][0])))){
	fprintf(stderr,"Inner product =%i is too large\n",in);
	goto error;
      }
      x->inner[i][ist]=in;
      if(0 == (x->inner[i][ist])){
	x->zeros[i][x->nzeros[i]++]=ist;
      } else
	x->length++;
    }
    x->zeros[i]=realloc(x->zeros[i],x->nzeros[i]*sizeof(size_t));
  }
  x->nsectors=i;
  if(x->length==0){
    fprintf(stderr,"Empty basis\n");
    goto error;
  }
  /*  Print out the results */
#if PRINTIT
  printf("    total length: %i (nonzero norm states)\n",(int) x->length);
#endif
#if 1==0  /* debugging stuff */
  printf("the basis is (with inner product):\n\n");
  for(i=0; i<x->nsectors; i++)
    for(ist=0; ist<x->basis[i]->length; ist++){
      int k;
      for(k=0; k<x->np[i]; k++) printf("%i %i  ",
       (x->basis[i]->locate[ist*x->np[i]+k]&((1<<HSHIFT)-1))*2+1,
       (x->basis[i]->locate[ist*x->np[i]+k]>>HSHIFT));
      printf(" %i",(int) x->inner[i][ist]);
      printf("\n");
    }
#endif
#if PRINTIT
  printf("    Total of %f CPU seconds to calculate basis\n",
	 (float)(clock()-t1)/(float) CLOCKS_PER_SEC);
#endif
  return;

error:
  fprintf(stderr,"Error in makebasis, ht=");
  for(i=0; i<HINDEX; i++)fprintf(stderr," %i",x->ht[i]);
  fprintf(stderr,", np=%i, kt=%i, cyclic=%i, o=%i, multi=%i, exiting\n",
	  x->npmax,x->kt,x->cyclic,x->o,x->multi);
  freebasis(x);
  return;
}

/*
  rebasis(x,i) rebuilds an existing basis for nonzero winding
  number or PHASE_CONVENTION==3. It also constructs the
  associated real-valued 1/sqrt(norm) array.  
  For l=0, the basis does not change for the first two 
  phase conventions, but it can change for the third.
  For l>0, the first phase convention does not work.  
  Only does multi=0.
  
  We must be careful in the case where a state has a nonzero
  inner product with its orientation reversal:  we must avoid
  including it twice (once for +o and once for -o) in the 
  basis.  We take the convention that we choose the +o version
  unless it vanishes (which can happen for nonzero n), else
  we take the -o state. 
*/

void rebasis(sectors *x, element *p1a){
  int i,oflag,winding=iswinding(x->ht);
  size_t ist;
  partype *temp;
  element eps=1.0e-12, (*whichinner)(partype *, partype *, int,
				     int, element *);
  if(x->nsectors==0){
    fprintf(stderr,"Empty basis in rebasis\n");
    goto error;
  } else if(PHASE_CONVENTION!=3 && !winding)
    return; /* don't need to do anything */ 
  else if(x->multi!=0){
    fprintf(stderr,"Nonzero multi=%i\n",x->multi);
    goto error;
  } else if(PHASE_CONVENTION==2){
    whichinner=innerproductcu;
  } else if(PHASE_CONVENTION==3 || (PHASE_CONVENTION==0 && winding)){
    whichinner=innerproductcv;
  } else {
    fprintf(stderr,"Can't find right phase convention in rebasis\n");
    goto error;
  }
  x->length=0;
  for(i=0; i<x->nsectors; i++){
    x->nzeros[i] = 0;
    x->innerno[i]=realloc(x->innerno[i],x->basis[i]->length*sizeof(element));
    x->zeros[i]=realloc(x->zeros[i],x->basis[i]->length*sizeof(size_t));
    for(ist=0, temp=x->basis[i]->locate; ist<x->basis[i]->length; 
	temp+=x->np[i], ist++){
      x->innerno[i][ist]=whichinner(temp,temp,x->np[i],x->o,p1a);
      oflag=x->o==-1 && 0==innerproduct(temp,temp,x->np[i],1,-1,0)
	&& fabs(whichinner(temp,temp,x->np[i],1,p1a))>eps;
     /* Print out inner products */
#if 0
      printf("norms: i=%i np=%i ist=%i  %i %e\n",i,x->np[i],ist,
	     x->inner[i][ist],x->innerno[i][ist]);
#endif
      if(fabs(x->innerno[i][ist])<=eps || oflag){
	x->innerno[i][ist]=0.0;
	x->inner[i][ist]=0;
	x->zeros[i][x->nzeros[i]++]=ist;
      } else {
	x->inner[i][ist]=ceil(x->innerno[i][ist]-eps);
	x->innerno[i][ist]=1/sqrt(x->innerno[i][ist]);
	x->length++;
      }
    }
    x->zeros[i]=realloc(x->zeros[i],x->nzeros[i]*sizeof(size_t));
  }
  return;
  
error:
  fprintf(stderr,"Error in rebasis, exiting\n");
  freebasis(x);
  return;
}

/* functions hltoi and hitol map a particle to and from a list
   of helicities.  The momentum is left untouched.
   hitol returns the momentum alone. */

#ifndef hltoi  /* There is also a macro version defined in include.h */

partype hltoi(partype x, int *hlist){
  int i;
  for(i=0; i<HINDEX; i++)
    if(hlist[i])break;
#if 0 /* debugging print */
  if(i==HINDEX){
    fprintf(stderr,"Illegal helicity values:  ");
    for(i=0; i<HINDEX; i++)fprintf(stderr,"%i ",hlist[i]);
    fprintf(stderr,"\nin function hltoi, exiting.\n");
    abort();
  } 
#endif
  return ((2*i+((hlist[i]+1)>>1))<<HSHIFT)+ (x & KMASK);
}
#endif
#ifndef hitol /* There is also a macro version defined in include.h */

partype hitol(partype x,int *hlist){
#if HINDEX==1
  hlist[0]=
#else
#if HINDEX==2
  hlist[0]=hlist[1]=0;
#else
  memset(hlist,0,HINDEX*sizeof(int));
#endif
  hlist[x>>(HSHIFT+1)] = 
#endif
    x&(1<<HSHIFT)?1:-1;
  /* return the momentum part of x */
  return x & KMASK;
}

#endif

/*  Find the inner product of two states, with flag for 
   cyclic permutations and, if o!=0, orientation reversals and
   transverse lattice symmetry multiplet multi 
   (see group.c for definitions).  
   This returns the number of nonzero contractions and 
   o = +1(-1) is even(odd) orientation reversal */

int innerproduct(partype *left, partype *right, int np, 
		 int cyclic, int o,int multi){
  int ss=0,s1,i,j,m;
  latype p;
  partype leftleft[2*NPMAX],*sp,*backtop=NULL,*top,backwards[2*NPMAX],c[NPMAX];
  if(cyclic){
    memcpy(leftleft,left,np*sizeof(partype));
    memcpy(top=leftleft+np,left,np*sizeof(partype));
    if(o){
      for(i=0; i<np; i++)backwards[i]=left[np-i-1];
      memcpy(backtop=backwards+np,backwards,np*sizeof(partype));
    }
    for(p=0; p<subgroup(multi)->length; p++){
      for(j=0; j<np; j++)c[j]=partrans(right+j,subgroup(multi)->locate[p]);
      s1=0;
      for(sp=leftleft; sp<top; sp++){
        for(m=0; m<np; m++)if(sp[m]!=c[m])goto b1;
	s1++;
      b1: ;
      }
      if(o)
	for(sp=backwards; sp<backtop; sp++){
          for(m=0; m<np; m++)if(sp[m]!=c[m])goto b2;
	  s1+=o;
	b2: ;
	}
      ss+=multiplet(multi)[p]*s1;
    }
  } else {
    if(o)for(i=0; i<np; i++)backwards[i]=left[np-i-1];
    for(p=0; p<subgroup(multi)->length; p++){
      for(j=0; j<np; j++)c[j]=partrans(right+j,subgroup(multi)->locate[p]);
      s1=0;
      for(m=0; m<np; m++)if(left[m]!=c[m])goto b3;
      s1++;
    b3:
      if(o && !memcmp(backwards,c,np*sizeof(partype)))s1+=o;
      ss+=multiplet(multi)[p]*s1;
    }
  }
  return ss;
}


/*  Find the inner product of two states for nonzero winding
    number n, transverse momentum 2*pi*l/n, and definite 
    orientation symmetry o for the 
    unboosted state.  Here, we use the convention that
    orientation reversal does not cyclically permute the trace
    (this is a phase convention).  Zero norm states
    remain zero norm.  The three forms below correspond to
    the three phase convenctions used in interactionc,
    interactioncu, and interactioncv.              */

element innerproductc(partype *left, partype *right, int np, 
		       int o, element *p1a){
  int l[HINDEX];
  int i;
  for(i=0; i<HINDEX; i++)
    l[i]=(int) (p1a[i]+0.5);
  if(iswinding(l)){
    fprintf(stderr,"Nonzero P_1 & n not implimented" 
	    " for first phase convention\n");
    return -1.0;
  } else
    return (element) innerproduct(left,right,np,1,o,0);
}

element innerproductcu(partype *left, partype *right, int np, 
		       int o, element *p1a){
  int i,w=innerproduct(left,right,np,1==1,0,0),ss=0,v=0;
  element total=0.0;
  int l[HINDEX];
  /*  element lx[HINDEX],bx[HINDEX],lt[HINDEX],bt[HINDEX],thetal,thetab;*/
  partype backwards[2*NPMAX];
  size_t length=np*sizeof(partype);
  /* check for l/w an integer */
  for(i=0; i<HINDEX; i++){
    l[i]=(int) (p1a[i]+0.5);
    if(l[i]%w != 0)return 0.0;
  }
  /* check if o=0  */
  if(o==0)return (element) w;
  for(i=0; i<np; i++)backwards[np-i-1]=right[i];
  memcpy(backwards+np,backwards,length);
  for(i=0; i<np; i++)
    if(0==memcmp(left,backwards+i,length)){v=i; ss++;}
  if(ss>0){
    for(i=0; i<HINDEX; i++)total+=(element) l[i];
#if 1==0
    printf("total=2 pi %e %i/%i (l, v and p)\n",total,v,np);
#endif
    total*= 2.0*(atan(1.) * 4.)*(element) v/(element) np;
    return (element) w+o * (element)ss * cos(total);
  } else
    return (element) w;
}

element innerproductcv(partype *left, partype *right, int np, 
		       int o, element *p1a){
  int i,j,w=innerproduct(left,right,np,1,0,0),v=0,phase[HINDEX],ss=0;
  element total=0.0;
  int hl[HINDEX],hb[HINDEX];
  partype backwards[2*NPMAX];
  size_t length=np*sizeof(partype);
  int l[HINDEX];
  /* check for l/w an integer */
  for(i=0; i<HINDEX; i++){
    l[i]=(int) (p1a[i]+0.5);
    if(l[i]%w != 0)return 0.0;
  }
  /* check if o=0  */
  if(o==0)return (element) w;
  for(i=0; i<np; i++)backwards[np-i-1]=right[i];
  memcpy(backwards+np,backwards,length);
  for(i=0; i<np; i++)
    if(0==memcmp(left,backwards+i,length)){v=i; ss++;}
  if(ss!=0){
    (void) hitol(left[0],hl); (void) hitol(backwards[v],hb);
    for(i=0; i<HINDEX; i++)phase[i]=(hb[i]-hl[i])/2;
    for(j=0; j<v; j++){
      (void) hitol(backwards[j],hb);
      for(i=0; i<HINDEX; i++)phase[i]+=hb[i];
    }
    for(i=0; i<HINDEX; i++)total+=((element) phase[i])*p1a[i];
#if 1==0
    printf("total=%i %e (shift and P_1a)\n",phase[0],p1a[0]);
#endif
      return (element) w+o*(element) ss*cos(total);
  } else
    return (element) w;
}

/*  give state canonical order with flag for
    cyclic permuations and, if o!=0, orientation reversals
    and transverse lattice symmetry multiplet multi 
    (see group.c for definitions).

    Speed is increased by using pointers for the loops and
    by using memcpy.
    on vpp, vectorization is not allowed for pointer variables.
    However, vectorization is still suppressed since the loops
    reference the functions memcmp and memcpy.  There is 
    probably a fundamental limit to vectorization for this
    procedure since the "best" can not be found in parallel.   

    It is not a good idea to use memcmp() in the following 
    because if partype spans two bytes, memcmp() can assign
    an incorrect order for the two states.

*/

void wellorder(partype *state, int np, 
	       int cyclic, int o, latlist *sub){
  int j;
  partype best[NPMAX],statestate[2*NPMAX],backwards[2*NPMAX],
    *sp,*top;
  latype p;
  int m;
  memcpy(best,state,np*sizeof(partype));
  if(cyclic){
    for(p=0; p<sub->length; p++){
      for(j=0; j<np; j++)statestate[j]=partrans(state+j,sub->locate[p]);
      memcpy(top=statestate+np,statestate,np*sizeof(partype));
      for(sp=statestate; sp<top; sp++)
	for(m=0; m<np; m++)
	  if(best[m]!=sp[m]){
	    if(best[m]>sp[m])
	      memcpy(best,sp,np*sizeof(partype));
	    break;
	  }
      if(o!=0){
	for(j=0; j<np; j++)backwards[j]=statestate[np-j-1];
	memcpy(top=backwards+np,backwards,np*sizeof(partype));
	for(sp=backwards; sp<top; sp++)
	  for(m=0; m<np; m++)
	    if(best[m]!=sp[m]){
	      if(best[m]>sp[m])
		memcpy(best,sp,np*sizeof(partype));
	      break;
	    }
      }
    }
  } else {
    for(p=0; p<sub->length; p++){
      for(j=0; j<np; j++)statestate[j]=partrans(state+j,sub->locate[p]);
	for(m=0; m<np; m++)
	  if(best[m]!=statestate[m]){
	    if(best[m]>statestate[m])
	      memcpy(best,statestate,np*sizeof(partype));
	    break;
	  }
      if(o!=0){
	for(j=0; j<np; j++)backwards[j]=statestate[np-j-1];
	for(m=0; m<np; m++)
	  if(best[m]!=backwards[m]){
	    if(best[m]>backwards[m])
	      memcpy(best,backwards,np*sizeof(partype));
	    break;
	  }
      }
    }
  }
  memcpy(state,best,np*sizeof(partype));
}

/*Add ly terms y to a sorted list x of length lx and max length maxlx. */

void addterms(element *x, int *lx,int maxlx,element *y,int ny){
  int i,j,k;
  for(i=0; i<ny; i++){
    for(j=0; j<*lx && x[j]<y[i]; j++);
    if(*lx<maxlx)(*lx)++;
    for(k=*lx-1; k>j; k--)x[k]=x[k-1];
    if(j<maxlx)x[j]=y[i];
  }
  return;
}

/* Add state to sorted basis; canonical order is assumed and
   operation is not done if it is a copy. Returns new length of basis.
   The speed of this routine is the limiting factor for
   the generation of large bases. 

   It is not a good idea to use memcmp() in the following 
   because if partype spans two bytes, memcmp() can assign
   an incorrect order for the two states.                       */

size_t addstate(partype *basis, partype *state,size_t nst, size_t np){
  size_t i,l=0,u=nst,length=np*sizeof(partype);
  int j;
  partype *bp;
/*     fprintf(stderr,"    add state to %i:  ",(int) nst);
       for(j=0;j<np;j++)fprintf(stderr,"%i ",(int) state[j]);
       fprintf(stderr,"\n");  */
  while(u-l>0){
    i=((u-l)>>1)+l;
    bp=basis+i*np;
    for(j=0; j<np; j++)if(state[j]!=bp[j])goto b1;
    return nst;
  b1:
    if(state[j]<bp[j])  
      u=i;
    else
      l=i+1; 
  }
/*  printf("values u=%i, l=%i\n",u,l); */
/* The following line is the limiting time factor
   in the generation of a large basis.*/
  memmove(basis+(u+1)*np,basis+u*np,(nst-u)*length);
  memcpy(basis+u*np,state,length);
  return nst+1;
}

/* Return position of state in basis (length nst) assuming the
   state is in the basis and the basis is sorted.  

   It is not a good idea to use memcmp() in the following 
   because if partype spans two bytes, memcmp() can assign
   an incorrect order for the two states.    
*/

size_t findnext(partype *basis, partype *state,size_t nst, size_t np){
  size_t i,l=0,u=1;
  int j;
  partype *bp;
  for(;u<nst;){
    bp=basis+u*np;
    for(j=0; j<np; j++)if(state[j]!=bp[j])goto b1;
    return u;
  b1:
    if(state[j]>bp[j]){  
      l=u+1; 
      u<<=1;
    }else
      break;
  }
  if(u>nst)u=nst;
  while(u-l>0){
    i=((u-l)>>1)+l;
    bp=basis+i*np;
    for(j=0; j<np; j++)if(state[j]!=bp[j])goto b2;
    return i;
  b2:
    if(state[j]<bp[j])  
      u=i;
    else
      l=i+1;
  }
  fprintf(stderr,"findnext():  Can't find state in basis, p=%i, exiting\n",np);
  fprintf(stderr,"  State: ");
  for(i=0; i<np; i++) fprintf(stderr," (%i,%i)",
			      (state[i]&KMASK)*2+1,state[i]>>HSHIFT);
  fprintf(stderr,"  or: ");
  for(i=0; i<np; i++) fprintf(stderr," %i",(int) state[i]);
  fprintf(stderr,"\n");
  fprintf(stderr,"  in basis (first 50): \n");
  for(j=0; j<nst && j<50; j++){
    fprintf(stderr,"  ");
    for(i=0; i<np; i++) 
      fprintf(stderr," (%i,%i)",
	      (basis[j*np+i]&KMASK)*2+1,basis[j*np+i]>>HSHIFT);
    fprintf(stderr,"\n");
  }
  exit(111);
  return 0;
}



/*  mahk(ht,np,kt,multi) makes a basis of states of momentum kt, 
    "helicities" ht, and np particles assuming symmetry under 
    orientation reversals and lattice symmetries associated with
    multiplet multi (multi=0 corresponds to the trivial subgroup:
    no lattice symmetries).  It returns a pointer to the array,
    and the number of states in that array.
    Here, 2*kt is the "harmonic resolution" K.
    Internal storage has all momenta right bit shifted
    by one and "helicites" 0,1. 
    o is flag to indicate if orientation reversals are included. 

    mahk efficiently handles cases where "helicities" are negative.  This
    will give a factor of HINDEX! * 2^HINDEX improvement in memory and 
    speed. 

    Memory allocation is an important issue in systems like the
    CRAY where memory fragmentation is a problem.  The first
    MAX_BASE, in units of sizeof(partype), are stored in static
    memory.  Any overflow is malloc'ed.  The global variable max_base
    is the amount of memory used.

*/ 

#if MAX_BASE>0
static partype stbase[MAX_BASE];
#else
static partype *stbase=NULL;
#endif
static kbase b={stbase,0};
static int nhkdone=0;
static struct {int ht[HINDEX]; int np; int kt; latlist *subgroup; 
  int o; int cyclic; kbase b; int ismalloc;} hkdone[MAXDONE];

kbase *mahk(int *ht, int np, int kt, 
		   int cyclic, int o, latlist *sub){
  int i,j,kk,l,hh1,hlist[HINDEX],abswind=0,reverse;
  latype g,p;
  partype hh,kt2;
  size_t ns=0;
  partype *s=NULL,z[NPMAX];
  kbase *y;
/*   Check if kt is is small enough */
#if HSHIFT>0
  if((abs(kt)-np+1)>>1 > KMASK){
    fprintf(stderr,"K = %i in mahk is too large,\n"
	    "because HSHIFT=%i, exiting.\n",kt,HSHIFT);
    return NULL;
  } 
#else
  if(kt>0)fprintf(stderr,"kt=%i is nonzero in mahk!\n",kt);
#endif
/* Check if o=0 or 1 */
  if(abs(o)>1){
    fprintf(stderr,"o=%i is invalid in mahk, exiting.\n",o);
    return NULL;
  }
/* Check that the subgroup of lattice symmetries 
   leaves the winding number invariant.  */
  if(!isinvariantvector(ht,sub)){
    fprintf(stderr,"Subgroup of lattice symmetries does not leave\n"
	    "winding number (");
    for(i=0; i<HINDEX; i++)fprintf(stderr," %i",ht[i]);
    fprintf(stderr,") invariant in mahk(...), exiting.\n");
    return NULL ;
  }

/*    Check if it has already been done */
  for(i=0; i<nhkdone; i++){
    for(j=0; j<HINDEX && hkdone[i].ht[j]==ht[j]; j++);
    if(j==HINDEX && sub->length == hkdone[i].subgroup->length){
      for(g=0; g<sub->length &&  
	    hkdone[i].subgroup->locate[g]==sub->locate[g]; g++);
      if(g==sub->length && hkdone[i].kt==kt && hkdone[i].np==np &&
	 hkdone[i].cyclic==cyclic && (hkdone[i].o==o || np<3)){
	/*      printf("g=%i, sub->length=%i\n",g,sub->length); */
	/*      printf("already done\n");  */
	return &hkdone[i].b;
      }
    }
  } 
#if 1==0 /* debugging print */
        fprintf(stderr,"starting ht =");
           for(i=0;i<HINDEX;i++)fprintf(stderr," %2i",ht[i]);
           fprintf(stderr,", np = %i, K = %i\n", (int) np, (int) kt); 
#endif 
  if(nhkdone==MAXDONE-1){
    fprintf(stderr,"memory for hkdone[] too small, exiting\n");
    return NULL;
  } 
/*      zero particle case */
  if(np==0)
    if(!iswinding(ht))
      ns=1;
    else{
      fprintf(stderr,"illegal zero particle value for ht in mahk, exiting\n");
      return NULL;
    }
  /*      One particle case */
  else if(np==1){
    abswind=0;
    for(l=0; l<HINDEX; l++)abswind += abs(ht[l]);
    if(abswind!=1){
      fprintf(stderr,"illegal one particle value for ht in mahk, exiting\n");
      return NULL;
    }
    if(kt>0){
      if(kt%2==0){
	fprintf(stderr,"illegal one particle value for kt in mahk, exiting\n");
	return NULL;
      }	
      ns=1;
      s=malloc(sizeof(partype));
      kt2=(kt>>1);
      if(kt2>=1<<HSHIFT){
	fprintf(stderr,"K too large for HSHIFT, exiting\n");
	return NULL;
      }	
      s[0]=hltoi(kt2,ht);
    } else {
      ns=(abs(kt)+1)>>1;
      s=malloc(ns*sizeof(partype));
      for(kt2=0; kt2<ns; kt2++)
	s[kt2]=hltoi(kt2,ht);
    }
  } else{
    /*      Many particle case done with recusion.  */
    /* When HSHIFT=0, kk only takes on the value 0  */
#if HSHIFT>0
    for(kk=1; kk<abs(kt)-np+2; kk+=2){
      if(kk>>1 >= 1<<HSHIFT){
	fprintf(stderr,"K too large for HSHIFT in mahk, exiting\n");
	return NULL;
      }	
#else
    kk=0;{
#endif

	/* hh can not be used as the loop variable because
	   of possible overflow when hh = (2*HINDEX)<<HSHIFT  */
      for(hh1=0; hh1<2*HINDEX; hh1++){
	hh=hh1<<HSHIFT;
	(void) hitol(hh,hlist);
	/*  Make sure that we haven't already used this index assuming
	    equivalence under the specified subgroup.                  */
	for(g=0; g<sub->length && hh>=partrans(&hh,sub->locate[g]); g++);
	/* Find the minimum number of particles necessary for the winding number */
	abswind=0;
	for(l=0; l<HINDEX; l++)abswind += abs(hlist[l]=ht[l]-hlist[l]);
	if(abswind <= np-1 && g==sub->length){
	  p=latorder(hlist);
	  /*  here, we assume multi=0 is the trivial lattice subgroup:
	      that is, we assume no lattice symmetries for
	      states with fewer particles since they have, in general, nonzero winding.  
	      However, we assume orientation reversal and cyclic symmetries.  */
	  y=mahk(hlist,np-1,(kt>0)?kt-kk:kt+kk,1==1,1,subgroup(0));
	  
	  /* add a particle to the end of each state along with its 
             orientation reversal (but not if it's both cyclic and o-symmetric)
             along with all cyclic permutations of the state.  */
	  for(reverse=0; (o&&cyclic)?(reverse<1):(reverse<2); reverse++)
	    for(j=0; j<np-1; j++){
	      s = realloc(s,(ns+y->length)*np*sizeof(partype));   
	      if(s==NULL){
		fprintf(stderr,"2 no memory left in mahk, exiting\n");
		return NULL;
	      } 
	      for(i=0; i < y->length; i++){
		/* transform the helicity indices back again and copy
		   onto z[] with a cyclic permutation, reversal. */
		for(l=0; l<np-1; l++)
		  z[reverse?np-2-l:l]=
		    partrans(y->locate+i*(np-1)+((j+l)%(np-1)),p);
		z[np-1]=(kk>>1)+hh;
#if 1==0 /* debugging print */
		fprintf(stderr,"                z: ");
		for(l=0;l<np;l++)fprintf(stderr,"%i ",(int) z[l]>>HSHIFT);
		fprintf(stderr,"\n"); 
#endif
		wellorder(z,np,cyclic,o,sub);
		ns=addstate(s,z,ns,np);
	      } 
	    }
        } 
      }
#if 1==0 /* debugging print */ 
      fprintf(stderr,"ns = %i \n",ns);
#endif
    } 
    s=realloc(s,ns*np*sizeof(partype));
  }
  
  /* Here we see if the results can be stored in static memory
     otherwise, they are malloc'ed.  */
  if(b.length+ns*np<MAX_BASE){
    memcpy(b.locate+b.length,s,ns*np*sizeof(partype));
    hkdone[nhkdone].b.locate=b.locate+b.length;
    b.length+=ns*np;
    hkdone[nhkdone].ismalloc=1;
    free(s);
  } else {
    hkdone[nhkdone].b.locate=s; 
    hkdone[nhkdone].ismalloc=0;
  }
  hkdone[nhkdone].b.length=ns;
  max_base+=ns*np;
  for(i=0; i<HINDEX; i++) hkdone[nhkdone].ht[i]=ht[i];
  hkdone[nhkdone].subgroup=sub;
  hkdone[nhkdone].o=abs(o); 
  hkdone[nhkdone].cyclic=cyclic;
  hkdone[nhkdone].np=np; hkdone[nhkdone].kt=kt;
  return &hkdone[nhkdone++].b;
}

/*  Free up memory used by mahk.
    Deallocation is in reverse order of allocation.  */

void freemahk(int j){
  while(nhkdone>j)
    if(hkdone[--nhkdone].ismalloc)
      free(hkdone[nhkdone].b.locate);
    else
      b.length-=hkdone[nhkdone].np*hkdone[nhkdone].b.length;
}

/*  resthk(bb,x,npx,kt,cyclic,o,subgroup) finds the remaining npy-particle
    states in the basis bb with a possible nonzero matrix
    element with npx-particle state x.
    It returns states in the basis occuring after or equal to x.
    It returns the resulting list of locations.  
   
    It is a good idea if the calling routine has already set
    malloc'ed enough memory for y.
*/

void resthk(cbase *y, sectors *bb, int npy, partype *x, int npx){
  /* Maximum number of particles in the interaction */
  int i,j,kk,l,hlist[HINDEX],nix,niy,
    k[2*NPMAX],hk[2*NPMAX][HINDEX],ktx=0;
  size_t request;
  partype z[NPMAX];
  kbase *s=NULL;
  if(bb->kt<0 && bb->cyclic){
    fprintf(stderr,"Can't do this in resthk, exiting\n");
    y->locate=NULL;
    return;
  }
  /*  printf("npx = %i, npy = %i, nix = %i, niy = %i\n",npx,npy,nix,niy); */
  y->length=0;
  for(i=0; i<npx; i++)k[i]=2*hitol(x[i],hk[i])+1;
  if(bb->cyclic){
    memcpy(k+npx,k,npx*sizeof(int));
    memcpy(hk[npx],hk[0],npx*HINDEX*sizeof(int));
  }
  /* first, do interior of string (or whole string if cyclic) */
  /* nix = number of interacting particles in starting state */
  /* niy = number of interacting particles in final state */
  if(npx==1 && npy==1){
    nix=1; niy=1;
  } else {
    nix=(4+npx-npy)/2; niy=(4+npy-npx)/2; 
  }
  if(nix<1 || niy<1) return;
  for(j=0; (bb->cyclic)?(j<npx):(j+nix-1<npx); j++){    
    kk=0;
    for(l=0; l<HINDEX; l++)hlist[l]=0;
    for(i=0; i<nix; i++){
      kk+=k[j+i];
      for(l=0; l<HINDEX; l++)hlist[l]+=hk[j+i][l];
    }  
    /*  for(l=0; l<HINDEX; l++)printf(" %i",hlist[l]); */
    /* Get a basis with no symmetries */
    s=mahk(hlist,niy,kk,0,0,subgroup(0));
    /* see if y is big enough.  Note that y->chunk may not 
       necessarily be a multiple of npy.  */
    request=(y->length+s->length)*npy;
    if(request>y->chunk){
      y->chunk=request;
      y->locate=realloc(y->locate,y->chunk*sizeof(partype)); 
      if(y->locate==NULL){
	fprintf(stderr,"No memory left in resthk, exiting\n");
	return;
      }
    }
    /* printf("length of mahk result is %i\n",s->length); */
    for(i=0; i<s->length; i++){     
      for(l=0; l<npx-nix; l++)
	z[(l+j+niy)%npy] = x[(l+j+nix)%npx];
      for(l=0; l<niy; l++)
	z[(j+l)%npy] = s->locate[i*niy+l];
      wellorder(z,npy,bb->cyclic,bb->o,subgroup(bb->multi));
      if(npy==npx)
	for(l=0; l<npx; l++){
	  if(z[l]<x[l])goto noadd1;
	  else if(z[l]>x[l])break;
	}
      y->length=addstate(y->locate,z,y->length,npy);
    noadd1: 
      ;
    }
  } 
  if(!bb->cyclic && bb->kt<0){
    for(j=0; j<npx; j++)ktx+=k[j];
    if(npx==0 && npy==0){
      y->length=1;
      return;
    }
    nix=(2+npx-npy)/2; 
    niy=(2+npy-npx)/2;
    for(j=0;;){    
      kk=abs(bb->kt)-ktx;
      for(l=0; l<HINDEX; l++)hlist[l]=0;
      for(i=0; i<nix; i++){
	kk+=k[j+i];
	for(l=0; l<HINDEX; l++)hlist[l]+=hk[j+i][l];
      }  
      /*  for(l=0; l<HINDEX; l++)printf(" %i",hlist[l]); */
      /* Get a basis with no symmetries */
      s=mahk(hlist,niy,-kk,0,0,subgroup(0));
      /* see if y is big enough.  Note that y->chunk may not 
	 necessarily be a multiple of npy.  */
      request=(y->length+s->length)*npy;
      if(request>y->chunk){
	y->chunk=request;
	y->locate=realloc(y->locate,y->chunk*sizeof(partype)); 
	if(y->locate==NULL){
	  fprintf(stderr,"No memory left in resthk, exiting\n");
	  return;
	}
      }
      for(i=0; i<s->length; i++){     
	for(l=0; l<npx-nix; l++)
	  z[j?l:l+niy] = x[j?l:l+nix];
	for(l=0; l<niy; l++)
	  z[j+l] = s->locate[i*niy+l];
	wellorder(z,npy,bb->cyclic,bb->o,subgroup(bb->multi));
	if(npy==npx)
	  for(l=0; l<npx; l++){
	    if(z[l]<x[l])goto noadd2;
	    else if(z[l]>x[l]) break;
	  }
	y->length=addstate(y->locate,z,y->length,npy);
      noadd2:
	;
#if 1==0 /* debugging print */
	if(npy==4&&z[0]==1&&z[1]==32&&z[2]==0&&z[3]==33){
	  printf(" left state:");
	  for(l=0; l<npx; l++)printf(" %i",x[l]);
	  printf("\n");
	}
#endif
      }
      if(npx-nix<1 || j)break; else j=npx-nix; 
    } 
  }
#if 0 /* debugging print */
  if(npx==1){
  printf("     findnext result for %i-particle state:\n     ",npx);
  for(i=0; i<npx; i++) printf(" %i",(x[i]&~(~0<<HSHIFT))*2+1);
  for(i=0; i<npx; i++) printf(" %i",x[i]>>HSHIFT);
  printf("  or: ");
  for(i=0; i<npx; i++) printf(" %i",(int) x[i]);
  printf("\n     found %i %i-particle states",y->length,npy);
  printf("\n");
  }
#endif
  return;
}

/* return a flag to say if a basis is winding  */

int iswinding(int *ht){
  int i;
  for(i=0; i<HINDEX && ht[i]==0; i++);
  return i<HINDEX;
}

#if 0 /* The inclusion of the option cyclic in mak() is not yet correct */

/*  mak(np,kt) makes a basis of states of momentum kt 
    and np particles assuming symmetry under orientation 
    reversals and flips.  It returns a pointer to the array,
    and the number of states in that array.  Here, kt is an odd
    integer. Internal storage has all numbers right bit shifted
    by one.  o is a flag for orientation reversals */

static int nkdone=0;
static struct {int np; int kt; kbase b; int cyclic; int o;} 
kdone[MAXDONE];

kbase *mak(int np, int kt, int cyclic, int o){
  int i,j,kk,l;
  size_t ns=0,tchunk=0;
  partype *s=NULL, z[NPMAX];
  kbase *y;
/*   Check if kt is is small enough */
  if(kt>0 && kt-np+1 >  (1<<(8*sizeof(partype)+1))-1){
    fprintf(stderr,"K = %i in mak is too large for %i particles, exiting.\n",kt,np);
    return NULL;
  } 
/* Check if o=0 or 1 */
  if(o!=0 && o!=1){
    fprintf(stderr,"o=%i is invalid in mahk, exiting.\n",o);
    return NULL;
  }
/*    Check if it has already been done */
  for(i=0; i<nkdone; i++)
    if(kdone[i].kt==kt && kdone[i].np==np && (kdone[i].o==o || np<3)) return &kdone[i].b;
/*        fprintf(stderr,"starting np,kt = %i %i \n", (int) np, (int) kt); */
  if(nkdone==MAXDONE-1){
    fprintf(stderr,"memory for kdone[] too small, exiting\n");
    return NULL;
  } 
 if(np==1){
/*      One particle case */
    s=malloc(1*1*sizeof(partype));
    s[0] = (partype) kt>>1;
    kdone[nkdone].b.length=1;
   }
  else{
/*      Many particle case done with recusion  */
    for(kk=1; kk<kt-np+2; kk+=2){
      y=mak(np-1,kt-kk,cyclic,o);
      for(i=0; i < y->length; i++){
/*          fprintf(stderr,"  At np,kt,kk,i =  %i %i %i %i\n",np,kt,kk,i); */
	for(j=0; j<np-1; j++){
	  if(ns==tchunk){
	    s = realloc(s,(tchunk+=CHUNK)*np*sizeof(partype)); 
	    if(s==NULL){
	      fprintf(stderr,"2 no memory left in mak, exiting\n");
	      return NULL;
	    } 
	  } 
	  for(l=0; l<np-1; l++) 
	    z[l]=(y->locate)[i*(np-1)+((j+l)%(np-1))];
	  z[np-1]=(partype) kk>>1;
/*            fprintf(stderr,"                z: ");
              for(l=0;l<np;l++)fprintf(stderr,"%i ",(int) (z[l]<<1)+1);
              fprintf(stderr,"\n");  */
          wellorder(z,np,cyclic,o,0);
          ns=addstate(s,z,ns,np);
        } 
      } 
    } 
    s=realloc(s,ns*np*sizeof(partype));   
    kdone[nkdone].b.length=ns;
  }    
  kdone[nkdone].b.locate=s;
  kdone[nkdone].np=np; kdone[nkdone].kt=kt;
  kdone[nkdone].o=o;
  return &kdone[nkdone++].b;
}

/*  Free up memory used by mak  */

void free_mak(int j){
  int i;
  for(i=nkdone-1; i>=j; i--) 
    free(kdone[i].b.locate);
  nkdone=j;
}

#endif


