#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "include.h"


/*  This file contains subroutines associated with the
    group structure of the lattice symmetries. 
    The first multiplets must be:
      the trivial subgroup
      even transverse parity (full)
      odd transverse parity  (full)
      singlet representation of the full group
      various other representations of the full group.  
*/

/*  subgroup and multiplet return a pointer to constant objects.
    Thus, other code can safely use pointers to specify subgroups
    and multiplets.
    Other code must not write onto these objects.   */


#if HINDEX == 1

/******************************************************************

     Multiplets are as follows:
     Charge conjugation: C = P_1 O, where O is orientation reversal
     In the following P_1 is parity and E is the identity.

   group   |   Multiplet                           
   element | 0  1  2  3   
  ---------|------------
   E       | 1  1  1  1  
   P_1     |    1 -1  1                      
  ----------------------

*******************************************************************/

latlist *subgroup(int mult){
  const static latype trivial=0, a[NGROUP]={0,1};
  static latlist 
    x[NMULTIPLETS]={{&trivial,1},{a,2},{a,2},{a,2}};
  return x+mult;
}

const int *multiplet(int mult){
  const static int a[NMULTIPLETS][NGROUP]={{1,0},{1,1},{1,-1},{1,1}};
  return a[mult];
}

#elif HINDEX == 2


/*******************************************************************

     Multiplets are as follows:
     Charge conjugation: C = R^2 O, where O is orientation reversal
     In the following P_1 and P_2 are parity, R is a rotation,
     and E is the identity.
     Note that:  P_2 = P_1 R^2 and R^2 = P_1 P_2.

  group   |                     Multiplet                           
  element | 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 
 ---------|---------------------------------------------------------
  E       | 1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1 
  P_1     |          1 -1  1 -1  1 -1  1 -1                         
  P_2     |          1 -1  1 -1 -1  1  1 -1              1 -1       
  R^2     |    1 -1  1  1  1  1 -1 -1  1  1  1 -1  1 -1             
          |                                                         
  P_1 R   |          1 -1 -1  1              1  1 -1 -1        1 -1 
  R       |          1  1 -1 -1                                     
  R^3     |          1  1 -1 -1                                     
  P_1 R^3 |          1 -1 -1  1              1 -1 -1  1             
 -------------------------------------------------------------------
  J_z^P_1 |         0+ 0- 2+ 2- 1+ 1-

  The multiplets have the following use:

  9,  8    P_perp = (c,0),   P_2 = +1
  7, 10    P_perp = (c,0),   P_2 = -1
  11,12    P_perp = (c,c), P_1 R = +1
  14,13    P_perp = (c,c), P_1 R = -1
  15       n = (c,0)         P_2 = +1        
  16       n = (c,0)         P_2 = -1
  17       n = (c,c)       P_1 R = +1 
  18       n = (c,c)       P_1 R = -1

********************************************************************/

latlist *subgroup(int mult){
  /*int i;
  const static int flag=0;
  const static latype b[NMULTIPLETS]={1,2,2,8,8,8,8,4,4}; */
  const static latype trivial[1]={0};
  const static latype a[NGROUP]={0,1,2,3,4,5,6,7}, p[2]={0,(1<<HINDEX)-1}, 
	        f1[4]={0,1,2,3},f2[4]={0,3,4,7},
                g1[2]={0,2},g2[2]={0,4};
  static latlist x[NMULTIPLETS]={{trivial,1},{p,2},{p,2},
	                 {a,8},{a,8},{a,8},{a,8},
                         {f1,4},{f1,4},{f1,4},{f1,4},
			 {f2,4},{f2,4},{f2,4},{f2,4},
                         {g1,2},{g1,2},{g2,2},{g2,2}};
  return x+mult;
}

const int *multiplet(int mult){
  const static int trivial=1;
  const static int a1[2]={1,1},a2[2]={1,-1};
  const static int
      a3[NGROUP]={1,1,1,1, 1,1,1,1}, a4[NGROUP]={1,-1,-1,1, -1,1,1,-1},
      a5[NGROUP]={1,1,1,1, -1,-1,-1,-1}, a6[NGROUP]={1,-1,-1,1, 1,-1,-1,1},
      a7[4]={1,1,-1,-1}, a8[4]={1,-1,1,-1},
      a9[4]={1,1,1,1}, a10[4]={1,-1,-1,1};
  const static int *x[NMULTIPLETS]={&trivial,a1,a2,
	  a3,a4,a5,a6,a7,a8,a9,a10,
          a9,a8,a7,a10,
	  a1,a2,a1,a2};
  return x[mult];
}

#else

#endif

/*  Factorial function for integers  */
#if 0 /* not used */
latype factorial(int n){
  static int x[]={1,1,2,6,24,120};
  if(n<6) return x[n]; else return n*factorial(n-1);
}
#endif

/* permutations of HINDEX objects */

#if HINDEX==1
static int permutations[]={0}; 
#elif HINDEX==2
static int permutations[]={0,1,1,0}; 
#else
int *permutations=NULL;
#endif

/*  Permutations of n integers.  Returns a pointer to a list
    of all possible permutations  */

#if 0  /* no longer used */

#define PERMLIMITS 15

int *perms(int n){
  int i,j,k;
  static int ndone[PERMLIMITS+1];
  static int *x[PERMLIMITS+1];
#if 0 /* debugging print */
  if(n>PERMLIMITS){
    fprintf(stderr,"Cannot do n=%i in permutations, exiting\n",n);
    return NULL;
  } 
#endif
  if(!ndone[n]){
    x[n]=malloc(factorial(n)*n*sizeof(int));
    if(x[n]==NULL){
      fprintf(stderr,"no memory left in permutations, exiting\n");
      return NULL;
    } 
    for(i=0;i<factorial(n-1);i++)
      for(j=0; j<n; j++){
        for(k=0; k<j; k++)
          x[n][(i*n+n-1-j)*n+k]=perms(n-1)[i*(n-1)+k];
        x[n][(i*n+n-1-j)*n+j]=n-1;
        for(k=j+1; k<n; k++)
          x[n][(i*n+n-1-j)*n+k]=perms(n-1)[i*(n-1)+k-1];
      } 
    ndone[n]=1;
  } 
  return x[n];
}

#endif /* no longer use perms */

/*  The function partrans modifies the "helicity" indices
    to undo the transformation defined by the p-th 
    lattice symmetry group element.  The momentum is left untouched.   */

partype partrans(partype *z, latype p){
  int i,hlist[HINDEX],h2[HINDEX];
  int *perm=permutations+(p>>HINDEX)*HINDEX;
  (void) hitol(*z,hlist);
  for(i=0; i<HINDEX; i++)h2[perm[i]]=hlist[i];
  for(i=0; i<HINDEX; i++)h2[i] *= 1-2*((int) (p>>i)&1);
#if 0 /* debugging print */
   fprintf(stderr,"In partrans, old hlist =");
   for(i=0; i<HINDEX; i++)fprintf(stderr," %2i",hlist[i]);
   fprintf(stderr,", new hlist =");
   for(i=0; i<HINDEX; i++)fprintf(stderr," %2i",h2[i]);
   fprintf(stderr,",  x= %i,",x->flips);
   for(i=0; i<HINDEX; i++)fprintf(stderr," %i",x->permutes[i]);
   fprintf(stderr,".\n");
#endif
   return hltoi(*z,h2);
}

/*  The function latorder permutes and flips the various winding number
    indices into a canonical order based on the lattice symmetry group, 
    returning the associated lattice group element
    needed to undo the transformation.  */

latype latorder(int *hlist){
  int i,h1[HINDEX],h2[HINDEX],h3[HINDEX];
  latype p,pbest=0;
  for(i=0; i<HINDEX; i++)h1[i]=hlist[i];
  for(p=1; p<NGROUP; p++){
    for(i=0; i<HINDEX; i++)
      h2[i] = h1[i]*(1-2*((p>>i)&1));
    for(i=0; i<HINDEX; i++)
      h3[i]=h2[permutations[(p>>HINDEX)*HINDEX+i]];
    for(i=0; i<HINDEX && h3[i]==hlist[i]; i++);
    if(i<HINDEX && h3[i]>hlist[i]){
         for(i=0; i<HINDEX; i++)hlist[i]=h3[i];
	 pbest=p;
    }  
  } 
  return pbest;
}

/* Check to see if an integer vector is invariant
   under a given subgroup of the lattice rotation group.
   it returns a boolean value. */

int isinvariantvector(int *n,latlist *sub){
  int i,j,n2[HINDEX];
  int *perm;
  for(i=0; i<sub->length; i++){
    perm=permutations+(sub->locate[i]>>HINDEX)*HINDEX;
    for(j=0; j<HINDEX; j++)
      n2[perm[j]]=n[j];
    for(j=0; j<HINDEX; j++)
      n2[j]*=1-2*((sub->locate[i]>>j)&1);
#if 0 /* debug print */
    printf("  element %i n=(%i,%i) n2=(%i,%i)\n",
	   sub->locate[i],n[0],n[1],n2[0],n2[1]);
#endif
    for(j=0;j<HINDEX;j++)
      if(n[j]!=n2[j])return 0;
  }
  return 1;
}

/*This checks that the n is invariant.*/

int fisinvariantvector(element *n, latlist *sub){
  int i,j,k;
  int *perm;
  element n2[HINDEX];
  for(i=0; i<sub->length; i++){
    perm=permutations+(sub->locate[i]>>HINDEX)*HINDEX;
    for(j=0; j<HINDEX; j++)
      n2[perm[j]]=n[j];
    for(j=0; j<HINDEX; j++){
      k=(1-2*((sub->locate[i]>>j)&1));
      n2[j]*=k;
    }
#if 1 /* debug print */
    printf("  element %i n=(%g,%g) n2=(%g,%g)\n",
	   sub->locate[i],n[0],n[1],n2[0],n2[1]);
#endif
    for(j=0; j<HINDEX; j++)
    if(fabs(n[j]-n2[j])>1.0e-10)return 0;
  }
  return 1;
}












