#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "include.h"
#if defined(USE_MPI)
#include "run.h"
#include "usempi.h"
#endif

/* print timing information and size of basis. */
#define PRINTIT 0

#define SIZE_PRINT  "%i" /* Print format for size_t, either i or li */

/*  The subroutines in this file calculate the Hamiltonian 
    for both winding modes 
    and ordinary bound state spectra.  There are two forms:
    one returns an array.  The other outputs
    the nonzero matrix elements to an array.  */



/*  hamels(couplings,x) 
    returns the nonzero matrix elements of the Hamiltonian (upper triangle)
    for a given set of couplings and basis. 
    for a given 0 = -1,+1 and  kt is 2*K.  The nonzero matrix elements
    are returned.

    This include code that checks only nonzero matrix elements.       */

void hamels(elslist *els,sectors *x,element *couplings){
  int i,j,k,jrest,npi,npj,start[NPMAX],ndivroot;
  latype p;
  size_t ist,jst,izero,jzero,nonzero=0,request;
  element y1[NPARAMS],y2[NPARAMS],yt[NPARAMS],kfactor,*divroot=NULL;
  partype c[NPMAX],gc[NGROUP][NPMAX], *ipart, *jpart, *restpart;
  latlist *subg;
  cbase rest;
#if PRINTIT
  clock_t t1=clock();
#endif
#if defined(USE_MPI) && !ONE_PROCESSOR_SPECTRUM
  int nproc,myrank;
  MPI_Comm_size(MPI_COMM_WORLD,&nproc);
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);  
#endif

  subg=subgroup(x->multi);

  els->n=x->length;
  if(els->n>els->jchunk){
    els->jchunk=els->n;
    if(NULL==(els->j=realloc(els->j,(els->jchunk+1)*sizeof(size_t)))){
      fprintf(stderr,"Problem allocating for els->j, els->n=%i\n",els->n);
      goto memerror;
    }
  }
  rest.chunk=els->n*x->np[x->nsectors-1];
  if(NULL==(rest.locate=malloc(rest.chunk*sizeof(partype)))){
    fprintf(stderr,"Problem allocating for rest, rest.chunk=%i\n",
	    rest.chunk);
    goto memerror;
  }

  /* Allocate and initialize the vector used to supply the norm */

  if(x->kt)kfactor=((element) abs(x->kt))* 0.5; else kfactor=1.0;
  /*  a look-up table for kfactor/sqrt(integers)  */
  ndivroot=x->npmax*2*subg->length; 
  ndivroot=ndivroot*ndivroot+1;
  if(NULL==(divroot=malloc(ndivroot*sizeof(element)))){
    fprintf(stderr,"Problem allocating for divroot, ndivroot=%i\n",ndivroot);
    goto memerror;
  }
  for(i=1;i<ndivroot;i++)divroot[i]=kfactor/sqrt((element) i);
 
/* Find the position of the beginning of each sector */
  for(i=1, start[0]=0; i<x->nsectors; i++)
    start[i]=start[i-1]+x->basis[i-1]->length - x->nzeros[i-1];

/*  Actually calculate matrix elements of the Hamiltonian  */
/*  "i" refers to the left state and "j" refers to the right state
    i, j           sector number
    ist, jst       current position in the sector
    izero,jzero    Current number of null norm states in sector.
    jrest          current position in "rest"
    npi, npj       number of particles in the current sector
    ipart, jpart   pointer to the current state in the sector
    c[]            is the orientation reversal of the right state  
    gc[p][]        various lattice transforms applied to the left state  */

/* Redesign where two rows are calculated at once and the
   result is then copied onto els arrays.  This uses a bit
   more memory and one has to copy two arrays instead of 
   one, but it is still as fast.  */
 
/* loop through i sectors  */
  for(i=0; i<x->nsectors; i++){
    npi=x->np[i];
    ipart=x->basis[i]->locate;
    /* loop through states in each i sector  */
    for(izero=0, ist=0; ist<x->basis[i]->length; ipart+=npi, ist++){
      /* if a state has zero norm, then goto the next row  */
      if(izero < x->nzeros[i] && x->zeros[i][izero] == ist){
	izero++;
	continue;
      }
      k=ist+start[i]-izero;
      els->j[k]=nonzero;  /* els->j is filled even if we skip row */

#if defined(USE_MPI) && !ONE_PROCESSOR_SPECTRUM
 /* If this is not in range, go to the next row */
      if(ist%nproc != myrank)
	continue;
#endif
      /*  Calculate all lattice transformations of the left state */
      for(p=0; p<subg->length; p++)
	for(k=0; k<npi; k++)
	  gc[p][k]=partrans(ipart+k,subg->locate[p]);
      /*  Loop through j sectors  */
      for(j=i; j<x->nsectors && j<i+2; j++){
	npj=x->np[j];	
	jpart=x->basis[j]->locate+((i==j)?ist*npj:0);
	/*  calculate states with nonzero product with i */
	resthk(&rest,x,npj,ipart,npi);
	jzero=(i==j)?izero:0; jst=(i==j)?ist:0;
	/*  See if there is enough memory to store next row */
	request=nonzero+rest.length;
	if(request>=els->tchunk){
	  els->tchunk=request;
	  els->i=realloc(els->i,els->tchunk*sizeof(size_t));
	  if(NULL==els->i)
	    goto memerror;
	  els->e=realloc(els->e,els->tchunk*sizeof(element));
	  if(NULL==els->e)
	    goto memerror;
	}
	/*  loop through j states in each sector  */
	for(jrest=0, restpart=rest.locate; 
	    jrest<rest.length; 
	    jst++, jpart+=npj, jrest++, restpart+=npj){
	  /*  see if j state matches the state in "rest"
	      findnext is MUCH faster, as it should be.  */ 
#if 1
	  k=findnext(jpart,restpart,x->basis[j]->length-jst,npj);
	  jst+=k; jpart+=k*npj;
#else
	  while(memcmp(restpart,jpart,npj*sizeof(partype))!=0){
	    jst++; jpart+=npj;
	  }
#endif
	  /*  if the state has zero norm, goto next state */ 
	  if(x->inner[j][jst] == 0)continue;
	  while(jzero<x->nzeros[j]&&x->zeros[j][jzero]<jst)jzero++;
	  for(k=0; k<NPARAMS; k++)yt[k]=0.0;
	  for(k=0; k<npj; k++) c[k]=jpart[npj-k-1];
	  for(p=0; p<subg->length; p++){
	    interaction(y1,gc[p],jpart,npi,npj);
	    interaction(y2,gc[p],c,npi,npj);
/*  The following multiplications are all times 1 or -1, 
    so the following could also be speeded up considererabley.  */
	    for(k=0; k<NPARAMS; k++) 
	      yt[k] += (y1[k]+(element) x->o*y2[k])*
		multiplet(x->multi)[p];
	  }
	  for(k=0; k<NPARAMS && fabs(yt[k])<1.0e-14; k++);
/* if zero, go to next state in rest */
	  if(k==NPARAMS) continue;
/*This is the 'c' convention with zero for the first matrix element */
	  els->e[nonzero]=0.0;
	  for(k=0; k<NPARAMS; k++)els->e[nonzero] += couplings[k]*yt[k];
	  els->i[nonzero] = jst+start[j]-jzero;
#if 0  /* Debugging: the following should give identical results */
	      els->e[nonzero] *= kfactor/sqrt((element) 
			         (x->inner[i][ist]*x->inner[j][jst]));
#else
	      els->e[nonzero] *= divroot[x->inner[i][ist]*x->inner[j][jst]];
#endif
	  nonzero++;
	}
      }
    }
  }
  els->j[els->n]=nonzero;
  /*if(ist+start[i-1]-izero!=jst+x->start[j-1]-jzero || 
     jst+start[j-1]-jzero != x->length)
    fprintf(stderr,"Error in state count in hamelsc\n"
	    "ishift=%i jshift=%i supposed to be %i\n",
	    ist+start[i-1]-izero,jst+start[j-1]-jzero,x->length);*/
#if PRINTIT
  printf("Total of %f CPU seconds to calculate %i  matrix elements\n",
         (float) (clock()-t1)/(float) CLOCKS_PER_SEC,nonzero); 
#endif
  free(divroot);
  free(rest.locate);
  return;

memerror:
  fprintf(stderr,"Ran out of memory in hamels(), nonzero=%i\n",nonzero);
  free(divroot);
  free(rest.locate);
  els->j=NULL; /* This is an error condition */
  return;
}

/*  hamelsc(els,x,y,couplings) 
    returns the nonzero matrix elements of the Hamiltonian (upper triangle)
    for a given set of couplings and basis. 
    for a given 0 = -1,+1 and  kt is 2*K.  The nonzero matrix elements
    are returned. x is the basis with desired transverse parity and o
    while y is a basis with the opposite transverse parity and o.
    x->multi and y->multi must be 1 or 2.

    This include code that checks only nonzero matrix elements.       */


void hamelsc(elslist *els, sectors *x, sectors *y,element *couplings){
  int i,j,k,jrest,npi,npj,opi,opj,winding=iswinding(x->ht),
    start[NPMAX],ndivroot;
  latype p;
  size_t ist,jst,izero[2],jzero[2],nonzero=0,nz[2],request;
  sectors *xy[2];
  complex_e z0[NGROUP][NPARAMS],z1[NGROUP][NPARAMS];
  element kfactor,*divroot=NULL,zt[NPARAMS]/*,test[NPARAMS]*/;
  partype c[NPMAX],gc[NGROUP][NPMAX],*ipart,*jpart,*restpart;
  latlist *subg;
  size_t *els_i[2];
  element *els_e[2];
  cbase rest;
  void (*whichinteraction)(complex_e *, partype *, partype *,int,int);
#if PRINTIT
  clock_t t1=clock();
#endif
#if defined(USE_MPI) && !ONE_PROCESSOR_SPECTRUM
  int nproc,myrank;
  MPI_Comm_size(MPI_COMM_WORLD,&nproc);
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);  
#endif

  xy[0]=x; xy[1]=y;
  subg=subgroup(x->multi);

  /*  check for nonzero winding number */
  if(x->kt<0 && !x->cyclic)
    whichinteraction=interactiontens;
  else if((PHASE_CONVENTION==1 || PHASE_CONVENTION==0) && !winding){
    whichinteraction=interactionc;
  } else if(PHASE_CONVENTION==2){
    whichinteraction=interactioncu;
  } else if(PHASE_CONVENTION==3 || (PHASE_CONVENTION==0 && winding)){
    whichinteraction=interactioncv;
  } else {
    fprintf(stderr,"Can't find right phase convention in hamelsc\n");
    goto error;
  }
  /*  printf("xy[0]->o=%i xy[1]->o=%i\n",xy[0]->o,xy[1]->o);*/
  /*  Check that everything is OK  */
  if(x->o != -y->o || x->kt != y->kt
     || x->nsectors != y->nsectors){
    fprintf(stderr,"Bad input basis for hamelsc\n");
    goto error;
  }
  /* Check that both multiplets have the same subgroup */
  if(subg->length!=subgroup(y->multi)->length){
      fprintf(stderr,"wrong multiplets in hamelsc, exiting\n");
      goto error;
    }
  for(i=0; i<subg->length; i++)
    if(subg->locate[i]!=subgroup(y->multi)->locate[i]){
      fprintf(stderr,"wrong multiplets in hamelsc, exiting\n");
      goto error;
    }

  /* Allocate memory for els->j, temp_i, temp_e, and rest.  
     These should all be a row long (or less).  */

  els->n=x->length+y->length;
  if(els->n>els->jchunk){
    els->jchunk=els->n;
    if(NULL==(els->j=realloc(els->j,(els->jchunk+1)*sizeof(size_t))))
      goto memerror;
  }
  rest.chunk=els->n*x->np[x->nsectors-1];
  if(NULL==(rest.locate=malloc(rest.chunk*sizeof(partype)))){
    fprintf(stderr,"Problem allocating for rest, rest.chunk=%i\n",
	    rest.chunk);
    goto memerror;
  }

  for(k=0; k<2; k++){
    els_i[k]=malloc(els->n*sizeof(size_t));
    els_e[k]=malloc(els->n*sizeof(element));
    if(NULL==els_i[k] || NULL==els_e[k]){
      fprintf(stderr,"Problem allocating els_i and els_e in hamelsc\n");
      goto memerror;
    }
  }

  /* Allocate and initialize the vector used to supply the norm */

  if(x->kt)kfactor=((element) abs(x->kt))* 0.5; else kfactor=1.0;
  /*  a look-up table for kfactor/sqrt(integers)  */
  ndivroot=x->npmax*2*subg->length; 
  ndivroot=ndivroot*ndivroot+1;
  if(NULL==(divroot=malloc(ndivroot*sizeof(element)))){
    fprintf(stderr,"Problem allocating for divroot, ndivroot=%i\n",ndivroot);
    goto memerror;
  }
  for(i=1;i<ndivroot;i++)divroot[i]=kfactor/sqrt((element) i);
  
  /* Find the position of the beginning of each sector */
  for(i=1, start[0]=0; i<x->nsectors; i++)
    start[i]=start[i-1]+x->basis[i-1]->length - x->nzeros[i-1]+
      y->basis[i-1]->length - y->nzeros[i-1];

/*  Actually calculate matrix elements of the Hamiltonian  */
/*  "i" refers to the left state and "j" refers to the right state
   i, j           sector number
   izero[],jzero[]Current number of null norm states in sector.
   jrest          current position in "rest"
   npi, npj       number of particles in the current sector
   ipart, jpart   pointers to the current states in the sector
   c[]            is the orientation reversal of the right state
   in[] jn[]      are the inner product for the states.
   wzi[] wzj[]    current position in the list of zeros.      */
/* loop through i sectors  */
  for(i=0; i<x->nsectors; i++){
    npi=x->np[i];
    ipart=x->basis[i]->locate;
    izero[0]=izero[1]=0;
/* loop through states in each i sector  */
    for(ist=0; ist<x->basis[i]->length; ipart+=npi, ist++){
      
#if defined(USE_MPI) && !ONE_PROCESSOR_SPECTRUM
/*  If neither row is done by this processor, go to the next row  */
/*    but we don't forget to update izero[] and fill els->j !!!!!*/
      if(ist % nproc != myrank){
	for(opi=0; opi<2; opi++){
	  if(xy[opi]->inner[i][ist]==0){
	    izero[opi]++;
	  }else
	    els->j[2*ist+start[i]-izero[0]-izero[1]+opi]=nonzero;
	}
	continue;
      }
#endif
      nz[0]=0;
      nz[1]=0;
/*  Calculate lattice transformations of the left state */
      for(p=0; p<subg->length; p++)
	for(k=0; k<npi; k++)
	  gc[p][k]=partrans(ipart+k,subg->locate[p]);
/*  Loop through j sectors, here we assume 4-point interactions  */
      for(j=i; j<x->nsectors && j<i+2; j++){
	npj=x->np[j];
	jpart=x->basis[j]->locate+((i==j)?ist*npj:0);
/*  calculate states with possible nonzero product with i */
	resthk(&rest,x,npj,ipart,npi);
	jst=(i==j)?ist:0; 
	for(opi=0; opi<2; opi++)
	  jzero[opi]=(i==j)?izero[opi]:0;
/*  loop through j states in each sector  */
	for(jrest=0, restpart=rest.locate; jrest<rest.length; 
	    jst++, jpart+=npj, jrest++, restpart+=npj){
/*  see if j state matches the state in "rest"
    This should be the dominant time user in a large
    calculation.  findnext(...) is much faster... */
#if 1
	  k=findnext(jpart,restpart,x->basis[j]->length-jst,npj);
	  jst+=k; jpart+=k*npj;
#else
	  while(memcmp(restpart,jpart,npj*sizeof(partype))!=0){
	    jst++; jpart+=npj;
	  }
#endif
/* Update jzero[...] without explicitly looking at all inner products. */
	  for(opj=0; opj<2; opj++)
	    while(jzero[opj]<xy[opj]->nzeros[j] && 
		  xy[opj]->zeros[j][jzero[opj]]<jst)jzero[opj]++; 
	  for(k=0; k<npj; k++)c[k]=jpart[npj-k-1];
	  for(p=0; p<subg->length; p++){
	    whichinteraction(z0[p],gc[p],jpart,npi,npj);
	    whichinteraction(z1[p],gc[p],c,npi,npj);
	  }
	  for(opj=0; opj<2; opj++){
/* go on to the next one if the state doesn't exist */
	    if(xy[opj]->inner[j][jst]==0){
	      jzero[opj]++;
	      continue;
	    }
	    for(opi=0; opi<=((i==j && ist==jst)?opj:1); opi++){
/* go on to the next one if the state doesn't exist */
	      if(xy[opi]->inner[i][ist]==0)continue;
	      for(k=0; k<NPARAMS; k++)zt[k]=0.0;
	      for(p=0; p<subg->length; p++)
/*  The following multiplications are all times 1 or -1, 
    so the following could also be speeded up considerably.  */
		for(k=0; k<NPARAMS; k++)
		  if(opj==opi)
		    zt[k]+=(z0[p][k].re+xy[opj]->o*z1[p][k].re)*
		      multiplet(xy[opi]->multi)[p];
		  else if(opi<opj)
		    zt[k]-=(z0[p][k].im+xy[opj]->o*z1[p][k].im)*
		      multiplet(xy[opi]->multi)[p];
		  else
		    zt[k]+=(z0[p][k].im+xy[opj]->o*z1[p][k].im)*
		      multiplet(xy[opi]->multi)[p];
/*  Check that the matrix element is nonzero */
/*  This is, in fact, almost always the case */
	      for(k=0; k<NPARAMS; k++)
		if(fabs(zt[k])>1.0e-14)break;
/* go to next opi loop if it is zero */
	      if(k==NPARAMS)continue;
/*This is the 'c' convention with zero for the first matrix element */
	      els_i[opi][nz[opi]]=
		2*jst+opj+start[j]-jzero[0]-jzero[1];
	      els_e[opi][nz[opi]]=0.0;
	      for(k=0; k<NPARAMS; k++)
		els_e[opi][nz[opi]] += couplings[k]*zt[k];
	      if(x->cyclic && (winding || PHASE_CONVENTION==3)){
		els_e[opi][nz[opi]] *= 
		  kfactor*xy[opi]->innerno[i][ist]*xy[opj]->innerno[j][jst];
	      } else{
		els_e[opi][nz[opi]] *= 
		  divroot[xy[opi]->inner[i][ist]*xy[opj]->inner[j][jst]];
	      }
	      nz[opi]++;
	    }
	  }
	}
      }
/* Add two rows to list of nonzero matrix elements  */
      for(opi=0; opi<2; opi++){
	if(xy[opi]->inner[i][ist]==0){
	  izero[opi]++;
	  /* do next iteration of opi loop   */
	  continue;
	}
	k=2*ist+opi+start[i]-izero[0]-izero[1];
	els->j[k]=nonzero;
	request=nonzero+nz[opi];
	if(request>els->tchunk){
	  els->tchunk=request;
	  if(NULL==(els->i=realloc(els->i,els->tchunk*sizeof(size_t))))
	    goto memerror;
	  if(NULL==(els->e=realloc(els->e,els->tchunk*sizeof(element))))
	    goto memerror;
	}
#if 0
  fprintf(stderr,
	  "nonzero=%i nz=(%i,%i) ist=%i opi=%i\n",
	  nonzero,nz[0],nz[1],ist,opi);
#endif
	memcpy(els->e+nonzero,els_e[opi],nz[opi]*sizeof(element));
	memcpy(els->i+nonzero,els_i[opi],nz[opi]*sizeof(size_t));
	nonzero+=nz[opi];  /* it is important that this is updated last */
      }
    }
  }
  els->j[els->n]=nonzero;
#if PRINTIT
  printf("Total of %f CPU seconds to calculate %i matrix elements\n",
         (float) (clock()-t1)/(float) CLOCKS_PER_SEC,nonzero);
#endif
  free(divroot);
  for(k=0; k<2; k++){
    free(els_e[k]);
    free(els_i[k]);
  }
  free(rest.locate);
  return;

  /* Error handling */

memerror:
  fprintf(stderr,"Ran out of memory in hamelsc(),\n"
	  "nonzero=%i nz=(%i,%i), temp_e(temp_i) is  %i(%i) bytes\n",
	  nonzero,nz[0],nz[1],sizeof(element),sizeof(size_t));
error:
  free(divroot);
  for(k=0; k<2; k++){
    free(els_e[k]);
    free(els_i[k]);
  }
  free(rest.locate);
  els->j=NULL;  /* this is an error condition */ 
  return;
}


#if 0   /* we don't use this anymore since it doesn't work well */

/* This routine finds the starting and ending rows of the
   Hamiltonian to be calculated by this processor so that 
   each processor has the same amount of work to do.  */

void startstop(elslist *els){
  int nproc,myrank;
  MPI_Comm_size(MPI_COMM_WORLD,&nproc);
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);

  /* In case of roundoff problems, do the very first
     and last by hand:  */
  if(myrank==0)
    els->start=0;
  else
    els->start=(double) els->n*
      (1.0-sqrt(1.0-(double) myrank/(double) nproc));
  if(myrank==nproc-1)
    els->stop=els->n;
  else
    els->stop=(double) els->n*
      (1.0-sqrt(1.0-(double) (myrank+1)/(double) nproc));
}
#endif

/* diagnostic function to print some nonzero matrix elements...
   j=0 prints all elements */

void printels(elslist *x, int j){
  int i,length=x->j[x->n];
  if(j==0 || 2*j>=length){
    printf("Printing all nonzero matrix elements:");
    for(i=0; i<length; i++){
      printf("\n %i %f " SIZE_PRINT " ",i,x->e[i],x->i[i]);
      if(i<x->n)printf("  " SIZE_PRINT "",x->j[i]);
    }
  } else { 
    printf("Printing some nonzero matrix elements:");
    for(i=0; i<j; i++){
      printf("\n  %i %f " SIZE_PRINT " ",i,x->e[i],x->i[i]);
      if(i<x->n)printf("  " SIZE_PRINT "",x->j[i]);
    }
    printf("\n  .... \n  " SIZE_PRINT " ....  " SIZE_PRINT "",
	   x->n-1,x->j[x->n-1]);
    for(i=length-j; i<length; i++){
      printf("\n  %i %f " SIZE_PRINT " ",i,x->e[i],x->i[i]);
      if(i<x->n)printf("  " SIZE_PRINT "",x->j[i]);
    }
  }
  printf("\n");
  }


/***********************************************************************/

/*  makeham(x,couplings) returns the full Hamiltonian matrix
    for a given basis x.  This subroutine is mostly for debugging
    purposes and not at all fast or efficient. */

element *makeham(sectors *x, element *couplings){
  int i,j,k;
  latype p;
  size_t ist,jst,iist,jjst=0;
  element y1[NPARAMS],y2[NPARAMS],yt[NPARAMS],kfactor;
  static element *locate=NULL;
  partype c[NPMAX],g[NPMAX],*temp,*temp2;
#if PRINTIT
  clock_t t1=clock();
#endif
  if(x->kt)kfactor=((element) abs(x->kt))* 0.5; else kfactor=1.0;
  locate = realloc(locate,x->length*x->length*sizeof(element));
  if(locate == NULL){
    fprintf(stderr,"Not enough memory in makeham(x,o,kt) for Hamiltonian"
	    "matrix, exiting\n");
    return NULL;
  }
  for(iist=0; iist < x->length*x->length; iist++)
    locate[iist]=(element) 0.0;
 /*  Actually calculate matrix elements of the Hamiltonian  */
  for(iist=0, i=0; i<x->nsectors; i++)
    for(ist=0; ist<x->basis[i]->length; ist++){
      temp=x->basis[i]->locate+ist*x->np[i];
      if(x->inner[i][ist] != 0){
        for(k=0; k<x->np[i]; k++) c[k]= *(temp+x->np[i]-k-1);
        for(jjst=0, j=0; j<x->nsectors; j++)
           for(jst=0; jst<x->basis[j]->length; jst++){
            temp2=x->basis[j]->locate+jst*x->np[j];
	    if(x->inner[j][jst] != 0){
	      for(k=0; k<NPARAMS; k++)yt[k]=0.0;
	      for(p=0; p<subgroup(x->multi)->length; p++){
		for(k=0; k<x->np[j]; k++)
		  g[k]=partrans(temp2+k,subgroup(x->multi)->locate[p]);
		interaction(y1,temp,g, x->np[i],x->np[j]);
		interaction(y2,c,g,x->np[i],x->np[j]);
		for(k=0; k<NPARAMS; k++){
		  yt[k]+=(y1[k]+(element) x->o*y2[k])*multiplet(x->multi)[p];
		}
	      }
	      for(k=0; k<NPARAMS; k++)
		locate[iist*x->length+jjst] += couplings[k]*yt[k]*
		  kfactor/(sqrt((element) 
			 (x->inner[i][ist]*x->inner[j][jst])));
	      jjst++;
            }
	  }
        iist++;
	      /*  printf("\n");  */
      }
    }
  if(iist!=jjst || iist!=x->length)
    fprintf(stderr,"error in makeham: iist=" SIZE_PRINT ", jjst=" SIZE_PRINT ", length=" SIZE_PRINT "\n",
	    iist,jjst,x->length);
#if PRINTIT
  printf("Total of %f CPU seconds to calculate matrix elements\n",
       (float)(clock()-t1)/(float) CLOCKS_PER_SEC);
#endif
  return locate;
}

/*  makehamc(x,y,couplings) returns the full Hamiltonian matrix
    for given bases x and y.  It is not at all fast or efficient: 
    it is for debugging mostly. */

element *makehamc(sectors *x,sectors *y,element *couplings){
  int i,j,k,opi,opj,n,winding=iswinding(x->ht);
  latype p;
  size_t ist,jst,iist,jjst=0,opiist,opjjst;
  complex_e y0[NPARAMS],y1[NPARAMS];
  static element *locate=NULL;
  element yt[NPARAMS],kfactor;
  void (*whichinteraction)(complex_e *, partype *, partype *,int,int);
  partype c[NPMAX],g[NPMAX],*temp,*temp2;
  sectors *xy[2];
#if PRINTIT
  clock_t t1=clock();
#endif
  if(x->kt)kfactor=((element) abs(x->kt))* 0.5; else kfactor=1.0;
  if(x->kt<0 && !x->cyclic)
    whichinteraction=interactiontens;
  else if((PHASE_CONVENTION==1 || PHASE_CONVENTION==0) && !winding){
    whichinteraction=interactionc;
  } else if(PHASE_CONVENTION==2){
    whichinteraction=interactioncu;
  } else if(PHASE_CONVENTION==3 || (PHASE_CONVENTION==0 && winding)){
    whichinteraction=interactioncv;
  } else {
    fprintf(stderr,"Can't find right phase convention in makehamc\n");
    return NULL;
  }
  /* Check that both multiplets have the same subgroup */
  if(subgroup(x->multi)->length!=subgroup(y->multi)->length){
      fprintf(stderr,"wrong multiplets in makehamc, exiting\n");
      return NULL;
    }
  for(i=0; i<subgroup(x->multi)->length; i++)
    if(subgroup(x->multi)->locate[i]!=subgroup(y->multi)->locate[i]){
      fprintf(stderr,"wrong multiplets in makehamc, exiting\n");
      return NULL;
    }
  xy[0]=x; xy[1]=y;
  n=x->length+y->length;
  if(NULL==(locate = realloc(locate,n*n*sizeof(element)))){
    fprintf(stderr,"Not enough memory in makehamc(x,o,kt) for Hamiltonian"
	    "matrix, exiting\n");
    return NULL;
  }
  for(i=0; i<n*n; i++)locate[i]=0.0;
 /*  Actually calculate matrix elements of the Hamiltonian  */
  for(iist=0, i=0; i<x->nsectors; i++)
    for(ist=0; ist<x->basis[i]->length; ist++){
      temp=x->basis[i]->locate+ist*x->np[i];
      for(k=0; k<x->np[i]; k++)c[k]=temp[x->np[i]-k-1];
      for(jjst=0, j=0; j<x->nsectors; j++)
	for(jst=0; jst<x->basis[j]->length; jst++){
	  temp2=x->basis[j]->locate+jst*x->np[j];
	  for(p=0; p<subgroup(x->multi)->length; p++){
	    for(k=0; k<x->np[j]; k++)
	      g[k]=partrans(temp2+k,subgroup(x->multi)->locate[p]);
	      whichinteraction(y0,temp,g,x->np[i],x->np[j]);
	      whichinteraction(y1,c,g,x->np[i],x->np[j]);
#if 0        /* Print out nonzero matrix elements for debugging */
	      for(k=0; k<NPARAMS && 
		    fabs(y0[k].re)+fabs(y0[k].im)+fabs(y0[k].re)+
		    fabs(y1[k].re)+fabs(y1[k].im)<1.0e-14; k++);
	      if(i<=j && ist<=jst && k<NPARAMS){
		printf(" i=%i ist=%i j=%i jst=%i iist=%i jjst=%i\n",
		       i,ist,j,jst,iist,jjst);	
		  for(k=3; k<NPARAMS; k+=2){
		    printf("   %e %e %e %e  \n",
			   y0[k].re,y0[k].im,y1[k].re,y1[k].im);
		  }
		printf("\n");
	      }
#endif
	    opiist=iist;
	    for(opi=0; opi<2; opi++)
	      if(xy[opi]->inner[i][ist]!=0){
		opjjst=jjst;
		for(opj=0; opj<2; opj++)
		  if(xy[opj]->inner[j][jst]!=0){
		    for(k=0; k<NPARAMS; k++){
		      if(opi==opj)
			yt[k]=(y0[k].re+xy[opi]->o*y1[k].re)*
			  multiplet(xy[opj]->multi)[p];
		      else
			yt[k]=(y0[k].im+xy[opi]->o*y1[k].im)*
			  multiplet(xy[opj]->multi)[p]*(opi-opj);
		      if(x->cyclic&&(winding||PHASE_CONVENTION==3))
			yt[k]*=kfactor*xy[opi]->innerno[i][ist]*
			  xy[opj]->innerno[j][jst];
		      else
			yt[k]*=kfactor/sqrt((element) 
		            (xy[opi]->inner[i][ist]*xy[opj]->inner[j][jst]));
		      locate[opiist*n+opjjst] += couplings[k]*yt[k];
		    }
#if 0  /* debugging print */
		    if((opiist==0 && opjjst==3)||(opiist==3 && opjjst==0)){
		      printf("i,j=%i %i opi, opj= %i %i  yt[k]=",
			     opiist,opjjst,opi,opj,yt);
		      for(k=1; k<NPARAMS; k+=2)printf(" %e",yt[k]);
		      printf("\n   y0[k]=");
		      for(k=1; k<NPARAMS; k+=2)
			printf(" %e+i%e", y0[k].re,y0[k].im); 
		      printf("\n   y1[k]=");
		      for(k=1; k<NPARAMS; k+=2)
			printf(" %e+i%e", y1[k].re,y1[k].im); 
		      printf("\n");          
		    }
#endif
		    opjjst++;		    
		  }
		opiist++;		    		  
	      }
	  }
	  if(x->inner[j][jst]!=0)jjst++;
	  if(y->inner[j][jst]!=0)jjst++;
	}
      if(x->inner[i][ist]!=0)iist++;
      if(y->inner[i][ist]!=0)iist++;
      /*  printf("\n");  */
    }
  if(iist!=jjst || iist!=n)
    fprintf(stderr,"error in makeham: iist=" SIZE_PRINT ", jjst=" SIZE_PRINT ", length=" SIZE_PRINT "\n",
	    iist,jjst,x->length);
#if PRINTIT
  printf("Total of %f CPU seconds to calculate matrix elements\n",
	 (float)(clock()-t1)/(float) CLOCKS_PER_SEC);
#endif
  return locate;
}


/*  complexham(x,couplings) returns the full Hamiltonian matrix
    for given basis x.  It is not at all fast or efficient: 
    it is for debugging mostly. */

element **complexham(sectors *x,element *couplings){
  int i,j,k,n=x->length,winding=iswinding(x->ht);
  size_t ist,jst,iist,jjst=0;
  complex_e yes[NPARAMS];
  static element *locate[2]={NULL,NULL};
  element kfactor;
  partype *ci,*cj;
  void (*whichinteraction)(complex_e *, partype *, partype *,int,int);
#if PRINTIT
  clock_t t1=clock();
#endif
  if(x->kt)kfactor=((element) abs(x->kt))* 0.5; else kfactor=1.0;
  if(x->multi != 0 || x->o !=0){
    fprintf(stderr,"Must use trivial multiplet (no lattice symmetries),\n"
	    " or orientation reversal symmetry in complexham, exiting.\n");
    return NULL;
  }
  if(x->kt<0 && !x->cyclic)
    whichinteraction=interactiontens;
  else if((PHASE_CONVENTION==1 || PHASE_CONVENTION==0) && !winding){
    whichinteraction=interactionc;
  } else if(PHASE_CONVENTION==2){
    whichinteraction=interactioncu;
  } else if(PHASE_CONVENTION==3 || (PHASE_CONVENTION==0 && winding)){
    whichinteraction=interactioncv;
  } else {
    fprintf(stderr,"Can't find right phase convention in complexham\n");
    return NULL;
  }
  locate[0] = realloc(locate[0],n*n*sizeof(element));
  locate[1] = realloc(locate[1],n*n*sizeof(element));
  if(locate[0] == NULL || locate[1] == NULL){
    fprintf(stderr,"Not enough memory in complexham(x,o,kt) for Hamiltonian"
           "matrix, exiting\n");
    return NULL;
  }
  for(i=0; i<n*n; i++){
    locate[0][i]=0.0;
    locate[1][i]=0.0;
  }
 /*  Actually calculate matrix elements of the Hamiltonian  */
  for(iist=0, i=0; i<x->nsectors; i++)
    for(ist=0; ist<x->basis[i]->length; ist++)
      if(x->inner[i][ist]!=0){
	ci=x->basis[i]->locate+ist*x->np[i];
	for(jjst=0, j=0; j<x->nsectors; j++)
	  for(jst=0; jst<x->basis[j]->length; jst++)
	    if(x->inner[j][jst]!=0){
	      cj=x->basis[j]->locate+jst*x->np[j];
	      /*   if(x->np[j]==x->np[i]+2)
		      printf("Matrix element for states %i %i\n",iist,jjst);*/
	      whichinteraction(yes,ci,cj,x->np[i],x->np[j]);
#if 1==0 /* Print out nonzero matrix elements */
	      for(k=0; k<NPARAMS && 
		    fabs(yes[k].re)+fabs(yes[k].im)<1.0e-14; k++);
	      if(i<=j && ist<=jst && k<NPARAMS){
		printf(" i=%i ist=%i j=%i jst=%i iist=%i jjst=%i\n",
		       i,ist,j,jst,iist,jjst);	
		  for(k=3; k<NPARAMS; k+=2){
		    printf("  %e %e",
			   yes[k].re,yes[k].im);
		  }
		printf("\n\n");
	      }
#endif
	      for(k=0; k<NPARAMS; k++){
		locate[0][iist*n+jjst] += couplings[k]*yes[k].re*
		  kfactor/sqrt((element) 
			       (innerproduct(ci,ci,x->np[i],x->cyclic,0,0)*
				innerproduct(cj,cj,x->np[j],x->cyclic,0,0)));
		locate[1][iist*n+jjst] += couplings[k]*yes[k].im*
		  kfactor/sqrt((element) 
			       (innerproduct(ci,ci,x->np[i],x->cyclic,0,0)*
				innerproduct(cj,cj,x->np[j],x->cyclic,0,0)));
	      }
	      jjst++;		    
	    }	
	iist++;
      }
  /*  printf("\n");  */
  if(iist!=jjst || iist!=n)
    fprintf(stderr,"error in complexham: iist=" SIZE_PRINT ", jjst=" SIZE_PRINT ", length=" SIZE_PRINT "\n",
	    iist,jjst,x->length);
#if PRINTIT
  printf("Total of %f CPU seconds to calculate matrix elements\n",
	 (float)(clock()-t1)/(float) CLOCKS_PER_SEC);
#endif
  return locate;
}

/*  hamlisth(x,"file") returns the full Hamiltonian matrix
    for a given o = -1,+1 and  kt is 2*K.  The nonzero matrix elements
    are dumped in a file.  multi is the multiplet associated with
    the lattice symmetries.  (multi=0 is no lattice symmetry).
    The subroutine returns the number of
    nonzero matrix elements. 

    This include codes that checks only nonzero matrix elements.       */

size_t hamlisth(sectors *x, char *fname){
  int i,j,k,l,jrest,npi,npj;
  latype p;
  size_t nonzero=0,ist,jst,iist=0,jjst;
  element y1[NPARAMS],y2[NPARAMS],yt[NPARAMS],kfactor;
  partype c[NPMAX],gc[NPMAX], *ipart, *jpart;
  cbase rest={NULL,0,0};
  FILE *fp;
#if PRINTIT
  clock_t t1=clock();
#endif
  if(x->kt)kfactor=((element) abs(x->kt))* 0.5; else kfactor=1.0;
  if((fp = fopen(fname,"w")) == NULL){
    fprintf(stderr,"Trouble opening file %s in windlisth, exiting\n",fname);
    return 0;
  }
/*  Actually calculate matrix elements of the Hamiltonian  */
/*  "i" refers to the left state and "j" refers to the right state
    iist, jjst     number of the state in the final basis.
    i, j           sector number
    ist, jst       current position in the sector
    jrest          current position in "rest"
    npi, npj       number of particles in the current sector
    ipart, jpart   pointer to the current state in the sector
    c[]            is the orientation reversal of the left state      */
/* loop through i sectors  */
  for(i=0; i<x->nsectors; i++){
    npi=x->np[i];
/* loop through states in each i sector  */
    for(ist=0; ist<x->basis[i]->length; ist++){
/* if a state has nonzero norm, then it is really in the basis  */
      if(x->inner[i][ist] != 0){
        jjst = iist++;
	ipart=x->basis[i]->locate+ist*npi;
	for(k=0; k<npi; k++) c[k]=ipart[npi-k-1];
/*  printf("col %i \n",(int) iist);   */
/*  Loop through j sectors  */
        for(j=i; j<x->nsectors && j<i+2; j++){
          npj=x->np[j];
/*  calculate states with nonzero product with i */
          resthk(&rest,x,npj,ipart,npi);
          jrest=0;
/*  loop through j states in each sector  */
          for(jst = (i==j)?ist:0; jst < x->basis[j]->length; jst++){
	    if(x->inner[j][jst] != 0) jjst++;
	    if(jrest<rest.length){
	      jpart=x->basis[j]->locate+jst*npj;
/*  see if j state matches the state in "rest"  */
	      for(l=0; l<npj && rest.locate[jrest*npj+l]==jpart[l]; l++); 
	      if(l==npj){
		jrest++;
/*  if the state has nonzero norm and is in "rest" calculate matrix element */ 
		if(x->inner[j][jst] != 0){
/*  printf("(%i,%i) ",(int) iist,(int) jjst);  */
		  for(k=0; k<NPARAMS; k++)yt[k]=0.0;
		    for(p=0; p<subgroup(x->multi)->length; p++){
		      for(k=0; k<npj; k++)
			gc[k]=partrans(jpart+k,subgroup(x->multi)->locate[p]);
		      interaction(y1,ipart,gc,npi,npj);
		      interaction(y2,c,gc,npi,npj);
		      for(k=0; k<NPARAMS; k++) 
			yt[k] += (y1[k]+(element) x->o*y2[k])*
			  multiplet(x->multi)[p];
		    }
                  for(k=0; k<NPARAMS; k++) yt[k]*=kfactor/
		      sqrt((element) (x->inner[i][ist]* x->inner[j][jst]));
		  for(k=0; k<NPARAMS && fabs(yt[k])<1.0e-14; k++);
		  if(k!=NPARAMS){
/*  if(l != npj)printf("missed %ith nonzero at i,ist=%i,%i"
      " j,jst=%i,%i\n",++miss,i,ist,j,jst);  */
/*  The data file format is to have 1,1 as the first matrix element */
		    if(fprintf(fp,"" SIZE_PRINT " " SIZE_PRINT " ",iist,jjst) < 0) goto err;
		    for(k=0; k<NPARAMS; k++) 
		      if(fprintf(fp,"%.16g ",yt[k]) < 0) goto err;
		    if(fprintf(fp,"\n") < 0) goto err;
		    nonzero++;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  if(fclose(fp) == EOF)
    fprintf(stderr,"Trouble closing file %s in hamlisth\n",fname);
#if PRINTIT
  printf("Total of %f CPU seconds to calculate %i nonzero matrix elements\n",
         (float) (clock()-t1)/(float) CLOCKS_PER_SEC,nonzero);
#endif
  free(rest.locate);
  return nonzero;
 err:
  free(rest.locate);
  fprintf(stderr,"Trouble writing to data file "
	  "%s in function hamlisth, exiting\n",fname);
  return 0;
}

/*  hamvector(outvects,invects,nvects,couplings,x) 
    returns the Hamiltonian acting on the vectors invects. 
    for a given 0 = -1,+1 and  kt is 2*K.  The nonzero matrix elements
    are dumped in a file.  multi is the multiplet associated with
    the lattice symmetries.  (multi=0 is no lattice symmetry).

    This include codes that checks only nonzero matrix elements.       */

void hamvector(sectors *x,element *outvects, element *invects, 
	       size_t nvects,element *couplings){
  int i,j,k,l,jrest,npi,npj;
  latype p;
  size_t ist,jst,iist,jjst,nonzero=0;
  element y1[NPARAMS],y2[NPARAMS],yt[NPARAMS],kfactor,elem;
  partype c[NPMAX],gc[NPMAX], *ipart, *jpart;
  cbase rest={NULL,0,0};
#if PRINTIT
  clock_t t1=clock();
#endif
  if(x->kt)kfactor=((element) abs(x->kt))* 0.5; else kfactor=1.0;
/* initialize output vectors to zero */
  for(iist=0; iist<x->length*nvects; iist++)outvects[iist]=0.0;
/*  Actually calculate matrix elements of the Hamiltonian  */
/*  "i" refers to the left state and "j" refers to the right state
    iist, jjst     number of the state in the final basis.
    i, j           sector number
    ist, jst       current position in the sector
    jrest          current position in "rest"
    npi, npj       number of particles in the current sector
    ipart, jpart   pointer to the current state in the sector
    c[]            is the orientation reversal of the left state      */
/* loop through i sectors  */
  iist=0;
  for(i=0; i<x->nsectors; i++){
    npi=x->np[i];
/* loop through states in each i sector  */
    for(ist=0; ist<x->basis[i]->length; ist++){
/* if a state has nonzero norm, then it is really in the basis  */
      if(x->inner[i][ist] != 0){
        jjst = iist++;
	ipart=x->basis[i]->locate+ist*npi;
	for(k=0; k<npi; k++) c[k]=ipart[npi-k-1];
/*  printf("col %i \n",(int) iist);   */
/*  Loop through j sectors  */
        for(j=i; j<x->nsectors && j<i+2; j++){
          npj=x->np[j];
/*  calculate states with nonzero product with i */
          resthk(&rest,x,npj,ipart,npi);
          jrest=0;
/*  loop through j states in each sector  */
          for(jst = (i==j)?ist:0; jst < x->basis[j]->length; jst++){
	    if(x->inner[j][jst] != 0) jjst++;
	    if(jrest<rest.length){
	      jpart=x->basis[j]->locate+jst*npj;
/*  see if j state matches the state in "rest"  */
	      for(l=0; l<npj && rest.locate[jrest*npj+l]==jpart[l]; l++); 
	      if(l==npj){
		jrest++;
/*  if the state has nonzero norm and is in "rest" calculate matrix element */ 
		if(x->inner[j][jst] != 0){
/*  printf("(%i,%i) ",(int) iist,(int) jjst);  */
		  for(k=0; k<NPARAMS; k++)yt[k]=0.0;
		    for(p=0; p<subgroup(x->multi)->length; p++){
		      for(k=0; k<npj; k++)
			gc[k]=partrans(jpart+k,subgroup(x->multi)->locate[p]);
		      interaction(y1,ipart,gc,npi,npj);
		      interaction(y2,c,gc,npi,npj);
		      for(k=0; k<NPARAMS; k++) 
			yt[k] += (y1[k]+(element) x->o*y2[k])*
			  multiplet(x->multi)[p];
		    }
		  for(k=0; k<NPARAMS && fabs(yt[k])<1.0e-14; k++);
		  if(k!=NPARAMS){
		    nonzero++;
		    elem= 0.0;
		    for(k=0; k<NPARAMS; k++)elem += couplings[k]*yt[k];
		    elem *= kfactor/sqrt((element) 
                                       (x->inner[i][ist]*x->inner[j][jst]));
		    for(k=0; k<nvects; k++){
 /*  This is the 'C' convention with the first matrix element at 0  */
		      outvects[k*x->length+iist-1]+=
			elem*invects[k*x->length+jjst-1];
		      if(iist!=jjst)outvects[k*x->length+jjst-1]+=
				      elem*invects[k*x->length+iist-1];
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free(rest.locate);
#if PRINTIT
  printf("Total of %f CPU seconds to calculate matrix elements\n",
         (float) (clock()-t1)/(float) CLOCKS_PER_SEC);
#endif
}

#if 0 /***************************************************************/
/*  hammer and hamlet are not used (and not even finished)!  */

/*  Variables for communication between routines that call hammer
    and the routine hamlet.                                        */

FILE *hamstr;
struct {elslist *els; size_t chunk;} hels;
struct {element *in;  element *out; int nvects; int n;} hvects; 


/*  hammer(x,which) is an internal routine with 
    loops through the basis and calculates nonzero matrix elements
    of the hamiltonian.  What it then does is determined by 
    which.  It returns the number of nonzero matrix elements.

    This include codes that checks only nonzero matrix elements.       */

size_t hammer(sectors *x, char which){
  int i,j,k,l,jrest,npi,npj,p;
  size_t nonzero=0,ist,jst,iist=0,jjst;
  element y1[NPARAMS],y2[NPARAMS],yt[NPARAMS],kfactor;
  partype c[NPMAX],gc[NPMAX], *ipart, *jpart;
  cbase rest={NULL,0,0};
  FILE *fp;
  if(x->kt)kfactor=((element) x->kt)* 0.5; else kfactor=1.0;
/*  "i" refers to the left state and "j" refers to the right state
    iist, jjst     number of the state in the final basis.
    i, j           sector number
    ist, jst       current position in the sector
    jrest          current position in "rest"
    npi, npj       number of particles in the current sector
    ipart, jpart   pointer to the current state in the sector
    c[]            is the orientation reversal of the left state      */
/* loop through i sectors  */
  for(i=0; i<x->nsectors; i++){
    npi=x->np[i];
/* loop through states in each i sector  */
    for(ist=0; ist<x->basis[i]->length; ist++){
/* if a state has nonzero norm, then it is really in the basis  */
      if(x->inner[i][ist] != 0){
        jjst = iist++;
	ipart=x->basis[i]->locate+ist*npi;
	for(k=0; k<npi; k++) c[k]=ipart[npi-k-1];
/*  printf("col %i \n",(int) iist);   */
/*  Loop through j sectors  */
        for(j=i; j<x->nsectors && j<i+2; j++){
          npj=x->np[j];
/*  calculate states with nonzero product with i */
          resthk(&rest,npj,ipart,npi,x->multi);
          jrest=0;
/*  loop through j states in each sector  */
          for(jst = (i==j)?ist:0; jst < x->basis[j]->length; jst++){
	    jpart=x->basis[j]->locate+jst*npj;
	    if(x->inner[j][jst] != 0) jjst++;
	    if(jrest<rest.length){
	      jpart=x->basis[j]->locate+jst*npj;
/*  see if j state matches the state in "rest"  */
	      for(l=0; l<npj && rest.locate[jrest*npj+l]==jpart[l]; l++); 
	      if(l==npj){
		jrest++;
/*  if the state has nonzero norm and is in "rest" calculate matrix element */ 
		if(x->inner[j][jst] != 0){
/*  printf("(%i,%i) ",(int) iist,(int) jjst);  */
		  for(k=0; k<NPARAMS; k++)yt[k]=0.0;
		    for(p=0; p<order(x->multi); p++){
		      for(k=0; k<npj; k++)
			gc[k]=partrans(jpart[k],subgroup(x->multi)[p]);
		      interaction(y1,ipart,gc,npi,npj);
		      interaction(y2,c,gc,npi,npj);
		      for(k=0; k<NPARAMS; k++) 
			yt[k] += (y1[k]+(element) x->o*y2[k])*
			  multiplet(x->multi)[p];
		    }
		  for(k=0; k<NPARAMS && fabs(yt[k])<1.0e-14; k++);
		  if(k!=NPARAMS){
		    for(k=0; k<NPARAMS; k++) yt[k]*=kfactor/
		      sqrt((element) (x->inner[i][ist]* x->inner[j][jst]));
		    hamlet(iist,jjst,yt,which);
		    nonzero++;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  return nonzero;
}

/*  hamlet(iist,jjst,els,which) is an internal subroutine that
    takes a nonzero matrix element of the Hamiltonian and
    either dumps it into a file, saves it in a vector or,
    calculates part of a Lanczos matrix vector product.    */

void hamlet(size_t iist, size_t jjst, element *els, char which){
  int k;
  element elem=0.0;
  if(which=='f'){
    if(fprintf(hamstr,"%i %i ",iist,jjst) < 0) goto err;
    for(k=0; k<NPARAMS; k++) 
      if(fprintf(hamstr,"%.16g ",els[k]) < 0) goto err;
    if(fprintf(hamstr,"\n") < 0) goto err;
  } else {
    for(k=0; k<NPARAMS; k++) elem += couplings[k]*els[k];
/*  pinching code       */
    pinching
#if HSHIFT!=0
#if HINDEX==1
    for(k=0; k<2; k++) elem += couplings[NPARAMS+k
#endif
#endif
 if (which=='e'){

  } else if (which=='d'){
    for(k=0; k<hvects.nvects; k++){
      hvects.out[k*x->length+iist]+=
	elem*hvects.in[k*hvects.n+jjst];
      if(iist!=jjst)hvects.out[k*x->length+jjst]+=
		      elem*hvects.in[k*hvects.n+iist];
    }
  } else 
    fprintf(stderr,"The choice which=%c is improper for hamlet\n",which);
  return;

err:
  fprintf(stderr,"Trouble writing to data file "
	  "%s in function hamlisth, exiting\n",fname);
  return NULL;
}

#endif /***************************************************************/
















