/*
        Transverse lattice Global definitions

  What form of matrix elements to use in calculations 
      0  use DLCQ matrix elements from file
      1  use internally constructed DLCQ matrix elements.  
      2  use improved matrix elements from file (this is obsolete).
      3  use internally constructed improved matrix elements.  */  

#ifndef IMPROVED
#define IMPROVED 1
#endif

/* maximum value of K for improved matrix elements 
   (This depends on the contents of the data files containing
    the improved matrix elements)  This should be less than or
   equal to pow(2,HSHIFT).                         */

#if IMPROVED==0 || IMPROVED==2
#define KKMAX 20
#else
#define KKMAX 63 /* 500 */
#endif

/*  Number of "helicity" indices and bit shift of HSHIFT.
    These must be compatible with the space available in partype.
    HSHIFT = 0 indicates helicity only (set all momenta to zero).
    HSHIFT must leave at least two bits for the "helicity" index. */

#ifndef HINDEX
#define HINDEX 2
#endif

#define HSHIFT 6 /*10; 6 is maximum possible for partype=char*/

/* maximum number of particles. */

#define NPMAX 9

/* Include new interactions for the heavy source
   potential  */

#define NEWTAU 1 /* flag for new interactions */

/* NPARAMS Number of coupling constants.  
   NMULTIPLETS is the number of
   multiplets defined for lattice symmetries 
   NGROUP  is the order of the lattice symmetry group    */

#if HSHIFT!=0
#if HINDEX == 1
# if NEWTAU
#  define NPARAMS 7
# else
#  define NPARAMS 6
# endif
#  define NMULTIPLETS 4
#  define NGROUP 2
#elif HINDEX == 2
# if NEWTAU
#  define NPARAMS 10
# else
#  define NPARAMS 9
# endif
#  define NMULTIPLETS 19
#  define NGROUP 8
#else

#endif
#endif


/* partype can be defined to be 'char' or 'short'. 
   This must be compatable with the value of HSHIFT and
   KKMAX above. */
/* On the Fujitsu vpp, using char is 2 times as fast
   as using int although int vectorizes... */
#ifdef  __uxp__  
typedef unsigned char partype; /* char or int */
#else
typedef unsigned char /*short or char*/ partype;
#endif

/* Element is the floating point default for matrix elements 
   of the hamiltonian. Complex is the associated
   generalization to complex numbers.  latype is the type used to
   specify lattice transforms.                */

typedef double element;
typedef struct {element re; element im;} complex_e;
typedef unsigned int latype;

/*structure for returning a basis of states including a pointer
  and the number of states. Optionally, chunk contains the memory
  allocated in units of partype.  */

#include <stddef.h>
typedef struct {partype *locate; size_t length; } kbase;
typedef struct {partype *locate; size_t length; size_t chunk;} cbase;

/* Structure to hold a subgroup of the lattice rotation group.
   Elements of the subgroup are assumed to be sorted.
   The elements of latlist cannot be changed                    */

typedef struct {const latype *locate;const latype length;} latlist;

/* Structures to give a basis of states.
   innerno is 1/sqrt(norm); this is used for nonzero
   transverse momentum and winding since the norm is no 
   longer integer valued.  

   cyclic is a flag idicating if a string is closed (TRUE)
   or open (FALSE).

   kt<0 indicates abs(kt) is an upper limit.

   The basis should be initialized to x.nsectors=0
**/

typedef struct {
  int nsectors; 
  int ht[HINDEX]; int npmax;  int kt; 
  int cyclic; int o; int multi;  
  size_t length;
  unsigned int np[NPMAX]; 
  unsigned int nzeros[NPMAX]; 
  unsigned char *inner[NPMAX]; 
  element *innerno[NPMAX]; 
  size_t *zeros[NPMAX];
  kbase *basis[NPMAX]; 
} sectors;


/* Structure to hold matrix elements of a sparse Hamiltonian.
   Here, we use "symmetric column pointer format" as suggested
   in the Cray "man intro_sparse" pages.
   Proper initialization is {NULL,NULL,NULL,0,0,0}
   Here we include only one triangle of the matrix, assuming symmetry.
   The array j is length n+1 and
   els.j[els.n] contains the length of els.e and els.i.  */

typedef struct { 
  element *e;    /* The elements of the matrix */
  size_t *i;     /* The correspoding column of that element */
  size_t *j;     /* The position in the above list of each row of the matrix */
  size_t n;      /* Size of the matrix */
  size_t tchunk; /* Amount of memory allocated for els.e and els.i */ 
  size_t jchunk; /* els.jchunk+1 = Amount of memory allocated for els.j */
} elslist;

/* Flag for choices of phase factor convention for
   nonzero P_1.  The 1 and 3 choices are probably faster,
   and 2 or 3 are necessary for nonzero winding number.
   Setting it to be zero will cause the automatic 
   choice of 1 for n=0 and 3 for n>0.
   These are the fastest two choices.  */

#if HINDEX==1
#define PHASE_CONVENTION 0
#elif HINDEX==2
#define PHASE_CONVENTION 0
#endif

/* Flag to use library routine memcmp() in interaction subroutines
   and routines to compute the basis.                                */
/* Since the differences are generically detectable in the first 
   byte, it is faster to use inline code.  */

#define USEMEMCMP 0

/* ------ External variables and Function prototypes --------------------- */

/* from file: hamilton.c */

void hamels(elslist *,sectors *,element *);
void hamelsc(elslist *,sectors *, sectors *, element *);
                  /* make a list of nonzero matrix elements for nonzero 
		     transverse momentum.  */
#if 0  /* we don't use this any more */
void startstop(elslist *els);  /* For parallel code, find which rows */
#endif
void printels(elslist *,int);  /* print some nonzero matrix elements */
element *makeham(sectors *,element *);
element *makehamc(sectors *,sectors *,element *);
element **complexham(sectors *,element *);
size_t hamlisth(sectors *, char *);
void hamvector(sectors *,element *, element *, size_t, element *);

/* from files: interaction*.c  */

void interaction(element *, partype *, partype *,int,int);
        /* find matrix element of hamiltonian, assuming cyclic permutations */
void makephaselist(element *,int); /*make a list of angles for interactionc */
void interactionc(complex_e *, partype *, partype *,int,int);
        /* find matrix element of hamiltonian, assuming cyclic permutations */
void interactioncu(complex_e *, partype *, partype *, int, int);
        /*  find matrix element for nonzero transverse momentum, nonweighted*/
void interactioncv(complex_e *, partype *, partype *, int, int);
       /*  find matrix element for nonzero transverse momentum, 1st particle*/
void interactiontens(complex_e *, partype *, partype *, int, int);
                 /*  find matrix element for an open string with heavy ends */
void shuffle(sectors *,sectors *,element *, 
	     element *,int);       /*order 0++* and 2++ by particle content */
void maketenslist(element,int,element,element); 
                                     /* initialize longitudinal separations */
void multiply(complex_e *, complex_e *); /* multiply complex numbers*/
void phasetocomplex(complex_e *, element x); 
         /* turn angle into a complex number */
element selfinertia(element *); /* integrand for self-inertia calculations */
element seenorm(element *);  /* norm for improved 0->2 interactions */
element vinst(element *);  /* instantaneous improved 0->2 interaction */
element vtau(element *);   /* tau improved 0->2 interaction */
element vnewtau(element *);   /*  new tau improved 0->2 interaction */

/* from file: improved.c */

const int four,two;       /*number of terms in fourelements and twoelements*/
element mapcouplings(int,element *); /*Find beta & transform matrix elements*/
element *twoelements[KKMAX], *threeelements[KKMAX];
             /* matrix elements of the interaction*/
void printelements(FILE *); /* print memory usage of twoelements and threeelements. */

/* from file: group.c */

latlist *subgroup(int); /* subgroup of the lattice symmetry group
                       associated with given multiplet */
const int *multiplet(int); /* a multiplet (with length order) */
#if 0 /* not used */
latype factorial(int);  /* factorial function */
#endif
int *perms(int); /* permutations of n objects */
partype partrans(partype *, latype); /* undo a transformation */
latype latorder(int *);  /* Helicity indices into canonical form
                           using all lattice symmetries.  */
int isinvariantvector(int *, latlist *); 
                      /*is vector invariant under subgroup? */
int fisinvariantvector(element *, latlist *); 
                        /*is vector invariant under subgroup? */

/* from file: basis.c */

void initbasis(sectors *,int *,int,int,int,int,int);
/*pack basis parameters into an array of integers */
void makebasis(sectors *);
/*make a basis of states with the sectors format. */ 
void freebasis(sectors *);  /* clear memory in basis */
void rebasis(sectors *, element *);
int innerproduct(partype *, partype *, int, int, int,int); 
/* find inner product of 2 states with cyclic 
   permutations and orientation reversals and lattice symmetries.  */
element innerproductc(partype *, partype *, int, int, element *);
element innerproductcu(partype *, partype *, int, int, element *);
element innerproductcv(partype *, partype *, int, int, element *);
/* find inner product for nonzero winding and nonzero P_1*a
   and definite orientation reversal and any of
   the three standard phase conventions.  */ 
void wellorder(partype *,int,int,int, latlist *);     
/* Give state a canoncial order */
void addterms(element *, int *,int,element *,int);/*add terms to sorted list */
size_t addstate(partype *, partype *,size_t,size_t);     
/* Add a state to a basis */
size_t findnext(partype *, partype *,size_t,size_t);     
/*Find a state in a basis */
kbase *mahk(int *,int, int, int,int, latlist *); 
/*Make a basis with helicity and/or momentum*/
void freemahk(int i); /* delete all but the first i bases */ 
size_t max_base; /* total number of partype used to store bases */
void resthk(cbase *,sectors *, int, partype *, int);
   /* find the rest of the states in the basis after the given state */
int iswinding(int *ht); /* Flag to say if ht is nonzero winding */
partype hltoi(partype, int *);/* Turn helicity into internal format integer.*/
#if 1==0 /* not debugged yet */
kbase *mak(int, int, int, int);  /* make a basis with momentum only  */
void free_mak(int);             /* Clear memory for associated function  */ 
#endif


/* --------------- Macro Definitions --------------------------------*/

#define KMASK ((1<<HSHIFT)-1) /* Filter out momentum part */

#if HINDEX==2
/* The following macros are valid for the set of vectors
   (-1,0) (1,0) (0,1) (0,-1)
*/

#define EQUAL(A,B) (A[0]==B[0] && A[1]==B[1])
#define MINUS(A,B) (A[0]==-B[0] && A[1]==-B[1])
#define ORTHO(A,B) (A[0]?B[1]:B[0])
#endif


#if 1 /* Use macro version of hitol. */

      /*   this is used lots of times in constructing 
           the Hamiltonian matrix so a macro should be a
           bit faster. */

#if HINDEX==1
#define hitol(X,H)  ((H)[0]=X&(1<<HSHIFT)?1:-1,X&KMASK)
#elif HINDEX==2
#define hitol(X,H) ((H)[0]=(H)[1]=0, (H)[X>>(HSHIFT+1)]= \
  X&(1<<HSHIFT)?1:-1,X&KMASK)
#else
#define hitol(X,H) (memset(H,0,HINDEX*sizeof(partype)), (H)[X>>(HSHIFT+1)]= \
  X&(1<<HSHIFT)?1:-1,X&KMASK)
#endif
#else  /* use function version of hiltol in file basis.c */
partype hitol(partype, int *);/*Turn internal format integer into helicity*/
#endif










