#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "include.h" 
#include "run.h"
#ifdef USE_MPI
#include "usempi.h"
#endif

#if !defined(USE_MPI) || ONE_PROCESSOR_SPECTRUM

/*
  Given a basis, find the spectrum using laso to diagonalize
  the Hamiltonian.  If s.multi or s.o is different from
  s2.multi or s2.o then nonzero momentum is assumed. 
  Thus one simply uses the same basis twice for zero transverse
  momentum.  vecin retruns a pointer to a list of the associated
  eigenvectors. 

  This routine does not call makephaselist()
  or maketenslist()!   These must be called, when appropriate,
  before calling spectrum. 

  This was formerly a part of the file "extralaso.c"
*/

#define PRINTIT 1 /*Print values, method, time, et cetera in spectrum(...)*/

#define USE_ARPACK 0 /* use ARPACK else use LASO to do Lanczos */

/* 
   On the LINUX machines, it appears that ARPACK
   needs fewer matrix multiplies to converge but
   the internal calculations are much slower than
   LASO.  Thus LASO offers a slight performance 
   advantage.  

   In ARPACK, the arrays qstore and vec may overlap (thus
   saving some memory).  I have not implimented this option 
   since it may or may not hold for DNLASO. 
*/

/*
   MAX_ELS is a guess for the maximum number of nonzero
   matrix elements of the the Hamiltonian.
   If we define this carefully, memory fragmentation on the heap
   can be minimized. 
*/

#ifndef MAX_ELS
#define MAX_ELS 100000
#endif

/*************************************************************************/

/* Memory used by iovect and laso */

static doublereal *qstore=NULL;
static size_t lqstore;

/* list of nonzero matrix elements used by elop and spectrum. */

static elslist els={NULL,NULL,NULL,0,0,0};
static doublereal *spectrumval=NULL, *spectrumvec=NULL;
static size_t lval=0, lvec=0;

void spectrum(sectors *s,         /* basis, filled by initbasis(...) */ 
	      sectors *s2,        /* second basis for nonzero momentum */
	      doublereal **vals,  /* list of eigenvalues */
	      doublereal **vecin, /* list of eigenvectors */
	      integer nval,       /* number of eigenvalues */
	      element *couplings, /* coupling constants */
	      element *angle,     /* angles and momenta, length HINDEX+2 */
	      int globalkt,       /* maximum value of kt, for mapcouplings */
	      integer nfig        /* number of significant digits in answer */
	      ){
  int i,j;
  element beta;
  integer n,ierr=0, nperm=0, maxop=800,bet;
  doublereal *work=NULL;
  char jobz='v',uplo='u';
  /* maxj determines how much memory dnlaso uses.
     The minimum value is 6*nblock and is faster if
     larger.  The limiting factor should be the amount
     of memory available.  */
  integer nblock=3 /*2*/, maxj=40;
#if USE_ARPACK
  char sa[2]={'S','A'},all='A';
  integer ido,iparam[11]={1,0,0,1,0,0,1,0},one=1,ipntr[11],lworkl;
  logical true=1,*select=NULL;
  doublereal tol,*dummy=NULL,*workd,*workl;
#else 
  integer negval,*ind=NULL,post;
#endif

  /*   Timing information */

#if PRINTIT
  clock_t t1;
  time_t t2,tf;

  t1=clock();
  time(&t2);
  printf("    Eigenvalues, %li digits, globalkt=%i\n",nfig,globalkt);
  printf("      K=%i/2, %i-particles, o=%i, multi=%i, multi2=%i (%g %g);\n",
	 s->kt,s->npmax,s->o,s->multi,s2->multi,angle[0],angle[1]);
#endif

/*
  make basis, etc 

  It is more efficient if mapcouplings is called with
  the globally largest value of kt so that the improved
  matrix elements are not recalculated.
*/

  beta=mapcouplings(globalkt,couplings);

  makebasis(s);
  if(s->o != s2->o || s->multi != s2->multi){
    if(s->cyclic && s->kt>0){
      rebasis(s,angle);
      makebasis(s2);
      rebasis(s2,angle);
      makephaselist(angle,s->kt);   
    } else {
      /*bvds July 27, 2000 typo fixed */
      makebasis(s2);
      maketenslist(angle[HINDEX+1],s->kt,angle[HINDEX],beta);
    }      
#if PRINTIT
    time(&tf);
    printf("      %.2f seconds (%li wall) to construct basis, %i states\n",
	   (clock()-t1)/(float) CLOCKS_PER_SEC,tf-t2,
	   s->length+s2->length);
#endif
  } else {
#if PRINTIT
    time(&tf);
    printf("      %.2f seconds (%li wall) to construct basis, %i states\n",
	 (clock()-t1)/(float) CLOCKS_PER_SEC,tf-t2,s->length);
#endif
  }

#if PRINTIT
  t1=clock();
  time(&t2);
#endif
  
  /* set aside initial guess for memory for nonzero matrix elements */

  if(els.tchunk<MAX_ELS||els.tchunk<max_els){
    els.tchunk=(MAX_ELS<max_els)?max_els:MAX_ELS;
    els.i=realloc(els.i,els.tchunk*sizeof(size_t));
    if(NULL==els.i){
      fprintf(stderr,"els.i not allocated\n");
      goto error;
    }
    els.e=realloc(els.e,els.tchunk*sizeof(element));
    if(NULL==els.e){
      fprintf(stderr,"els.e not allocated\n");
      goto error;
    }
  }

  /* Choose Hamiltonian.  If the two bases are equal then use
     real valued, else use complex case (nonzero P_1).
     Nonzero string tension must always be complex valued. */
 
 if(s2->o==s->o && s2->multi==s->multi){
#if PRINTIT
    printf("      Use hamels()");
#endif
    hamels(&els,s,couplings);
  }else{
#if PRINTIT
    printf("      Use hamelsc()");
#endif
    hamelsc(&els,s,s2,couplings);
  }
  n=els.n;
#if PRINTIT
  time(&tf);
  printf(" in %.2f sec (%li wall), %i nonzero, maxop=%i, maxj=%i\n",
	 (clock()-t1)/(float) CLOCKS_PER_SEC,tf-t2,
	 (int) els.j[els.n],(int) maxop, (int) maxj);
#endif

  /* remember largest number of nonzero matrix elements. */
  
  if(max_els<els.j[els.n])max_els=els.j[els.n];    
  
#if 0 /* print out matrix elements of the Hamiltonian */
  printels(&els,100);
#elif 0  /* print out all matrix elements of the Hamiltonian. */
  {
    FILE *f1;
    char fn[]="laso1.out";
    f1=fopen(fn,"w");
    for(i=0; i<els.n+1; i++)
      fprintf(f1,"%i\n",els.j[i]);
    fclose(f1);
    fn[4]='2';
    f1=fopen(fn,"w");
    for(i=0; i<els.j[els.n]; i++)
      fprintf(f1,"%i\n",els.i[i]);
    fclose(f1);
    fn[4]='3';
    f1=fopen(fn,"w");
    for(i=0; i<els.j[els.n]; i++)
      fprintf(f1,"%.16e\n",els.e[i]);
    fclose(f1);
   }
#endif
  
#if 0  /* print out the length of the longest row */
  for(i=0, j=0; i<n; i++)
    if(els.j[i+1]-els.j[i]>j)
      j=els.j[i+1]-els.j[i];
  printf("longest row is %i\n",j);
#endif

/*
  If the matrix is too small for LASO routines, set up memory 
  allocation for LAPACK routine DSYEV.  In this case
  we have to "unpack" the lower triangle of the matrix from 
  els and store it in vec.  Here we put the "upper triangle"
  of the matrix into spectrumvec.  Hence, uplo='u'                 
*/

  if(n==0){
    fprintf(stderr,"Matrix of size zero in spectrum().\n");
    goto error;
  } else if(els.j==NULL){
    fprintf(stderr,"spectrum() detecting an error in hamels or hamelsc\n");
    goto error;
  } else if(n<=6*nblock||n<maxj){
    if(lvec<n*n){
      spectrumvec=realloc(spectrumvec,(lvec=n*n)*sizeof(doublereal));
      if(NULL==spectrumvec){
	fprintf(stderr,"spectrumvec not allocated\n");
	goto error;
      }
    }
    if(n>lval)
      if(NULL==(spectrumval=realloc(spectrumval,(lval=n)*sizeof(doublereal)))){
	fprintf(stderr,"spectrumval not allocated\n");
	goto error;
      }
    if(NULL==(work=malloc((bet=3*n+10)*sizeof(doublereal)))){
      fprintf(stderr,"els.i not allocated\n");
      goto error;
    }
    for(i=0; i<n*n; i++)spectrumvec[i]=0.0;
    /* The following is adapted from the subroutine elop listed below. */
    for(i=0, j=els.j[0]; j<els.j[n]; i++)
      for(; j<els.j[i+1]; j++)
	spectrumvec[els.i[j]*n+i]=els.e[j];
#if 0 /* Print reconstructed matrix for debugging. */
    for(i=0; i<n; i++){
      for(j=0; j<n; j++)printf(" %e",spectrumvec[i*n+j]);
      printf("\n");
    }
#endif
    
/* 
   LAPACK routine dsyev to give eigenvalues and 
   optionally eigenvectors of a symmetric matrix.
   The eigenvectors are returned in spectrumvec.          
*/

    DSYEV(&jobz,&uplo,&n,spectrumvec,&n,spectrumval,work,&bet,&ierr,1,1);
#if PRINTIT
    printf("LAPACK routine DYSEV, ierr=%li\n",ierr);
#endif
    if(ierr!=0)fprintf(stderr,"DSYEV returning error = %li\n",ierr);
    nperm=nval;
    
  } else {
    
    if(maxj<2*nval)maxj=2*nval; /* if maxj is too small, make it bigger */

/*
  Set up memory allocation for laso and NAG Lanczos
  The pointers work,val,spectrumvec,and qstore are defined globally.
  nblock is k for NAG.  

  qstore is the largest memory user for large matrices 
*/

    if(n*maxj>lqstore){
      lqstore=n*maxj;
      if(NULL==(qstore=realloc(qstore,lqstore*sizeof(doublereal)))){
	fprintf(stderr,"qstore of size %i not allocated\n",lqstore);
	goto error;
      }
    }
    if(lvec<n*nval){
      spectrumvec=realloc(spectrumvec,(lvec=n*nval)*sizeof(doublereal));
      if(NULL==spectrumvec){
	fprintf(stderr,"spectrumvec not allocated\n");
	goto error;
      }
    }
    
#if !USE_ARPACK   
    
    if(4*nval>lval)
      if(NULL==(spectrumval=realloc(spectrumval,
				    (lval=4*nval)*sizeof(doublereal)))){
	fprintf(stderr,"spectrumval not allocated\n");
	goto error;
      }
    if(NULL==(ind=malloc(nval*sizeof(integer)))){
      fprintf(stderr,"ind not allocated\n");
      goto error;
    }
    if(nblock==1 && nval>1)
      fprintf(stderr,"nblock=%li is too small\n",nblock);
    bet = nblock + 2*nval+6 + maxj*(2*nblock+1) + maxj +  
      (2*nblock+1)*(nblock+1);
    if(n*nblock>bet)bet=n*nblock;
    bet += 2*n*nblock + 3*nval + maxj*(nblock+1) + 2*nblock*nblock +
      maxj*(nval+1);

    /* calculate memory usage of the dnlaso post-processor, if larger
       than bet, write error.
       in dnlaso, make sure that nperm is always less than nval. */
    post = 2*nval*nval;
    if(n*nblock>post)post=n*nblock;
    post += 2*n*nblock+1 + 2*nval*nval + 2*nval+4 + (2*nblock+1)*(nblock+1);
    if(post>bet){
      fprintf(stderr,"post larger than bet in spectrum\n");
      goto error;
    }

    if(NULL==(work=malloc(bet*sizeof(doublereal)))){
      fprintf(stderr,"work not allocated\n");
      goto error;
    }
    for(i=0; i<nblock*n; i++)work[i]=0.0;
#if PRINTIT
    t1=clock();
    time(&t2);
#endif
    negval=-nval;  /* choose smallest eigenvalues */
    nperm=0; /* no presupplied eigenpairs. */
    
    /* DNLASO(OP, IOVECT, N, NVAL, NFIG, NPERM, NVAL, VAL, NMVEC,
       SPECTRUMVEC, NBLOCK, MAXOP, MAXJ, WORK, IND, IERR)   */
    DNLASO(ELOP_TO_FORTRAN,IOVECT_TO_FORTRAN, &n, &negval, 
	   &nfig, &nperm, &nval, spectrumval, &n, 
	   spectrumvec, &nblock, &maxop, &maxj, work, ind, &ierr);
#if PRINTIT
    time(&tf);
    printf("      Lanczos with DNLASO in %.2f seconds (%li wall), "
	   "%li*%li multiplies\n",
	   (clock()-t1)/(float) CLOCKS_PER_SEC,tf-t2,
	   ind[0],nblock);
#endif
  if(ierr!=0)fprintf(stderr,"DNLASO returning error = %li\n",ierr);

#else  /***** version which uses ARPACK *************/

  if(n>lval)
    if(NULL==(spectrumval=realloc(spectrumval,(lval=n)*sizeof(doublereal)))){
      fprintf(stderr,"spectrumval not allocated\n");
      goto error;
    }
  select=malloc(maxj*sizeof(logical));
  bet=4*n+maxj*(maxj+8);
  if(NULL==(work=malloc(bet*sizeof(doublereal)))){
    fprintf(stderr,"work not allocated\n");
    goto error;
  }
  
  t1=clock();
  tol=pow(10.0,-nfig);
  jobz='I';  /* standard eigenvalue problem */
  iparam[2]=maxop;
  workd=work+n;
  workl=workd+3*n;
  lworkl=bet-4*n;
  ido=0; /* first call to DSAUPD */
  ierr=0; /* use random initial vector */
  
  for(;;){
    DSAUPD(&ido, &jobz, &n, sa, &nval, &tol, work, 
	   &maxj, qstore, &n, iparam, ipntr, workd, workl,
	   &lworkl, &ierr,1,2);

#if 0 /*debugging print */
    printf("in loop, ido=%li ierr=%li,iparam[4]=%li\n",
	   ido,ierr,iparam[4]);
    printf("ipntr= %li %li, n=%li, maxj=%li\n",ipntr[0],ipntr[1],n,maxj);
#endif

    if(ido==1 || ido==-1)
      /* remember FORTRAN uses an offset of 1 */
      elop(&n,&one,workd+ipntr[0]-1,workd+ipntr[1]-1);
    else
      break;
  }

  if(ierr<0){
    fprintf(stderr,"ARPACK DSAUPD returning error = %li\n",ierr);
    goto error;
  }
  
/*
  call dseupd ( rvec, 'All', select, d, v, ldv, sigma, 
  &         bmat, n, which, nev, tol, work, ncv, v, ldv, 
  &         iparam, ipntr, workd,workl, &lworkl, ierr )  
*/
  
  DSEUPD(&true,&all, select, spectrumval, spectrumvec, &n,dummy, 
	 &jobz, &n, sa, &nval, &tol, work, &maxj, qstore,&n, 
	 iparam, ipntr, workd, workl, &lworkl, &ierr,1,1,2);
  
#if PRINTIT
  time(&tf);
  printf("      Lanczos with ARPACK in %.2f seconds (%li wall), "
	 "%li multiplies\n",
	 (clock()-t1)/(float) CLOCKS_PER_SEC,tf-t2,
	 iparam[8]);
#endif

  if(ierr!=0){
    fprintf(stderr,"ARPACK DSEUPD returning error = %li\n",ierr);
    goto error;
  }
  
#endif /************** choose ARPACK or DNLASO ****************/
  }

#if PRINTIT /* Print eigenvalues */
  printf("      ");
  for(j=0, i=3; j<nperm; j++){
    printf("%.10g ",spectrumval[j]);
    if((i++)%6==0 && j+1<nperm)printf("\n      ");
  }
  printf("\n");
#endif

#if 0 /* debugging print statements */
  printf("ierr =% li       number of matrix accesses = %li\n",ierr,ind[0]);
  printf("\n      Eigenvalue       residual norm  "
	 "value error   vector error\n");
  for(j=0; j<nperm; j++){
    printf(" %.12g",spectrumval[j]);
    for(i=1; i<4; i++)printf("  %e",spectrumval[i*nval+j]);
    printf("\n");
  } 
#endif

#if 0 /* debugging statement to print out eigenvectors */
  printf("\n First four components of the eigenvectors:\n");
  for(j=0; j<nperm; j++){
    for(i=0; i<4; i++)printf("  %e",spectrumvec[i+j*n]);
    printf("\n");
  } 
#endif

  free(work);

#if USE_ARPACK
  free(select);
#else
  free(ind);
#endif

  /* A pointer to the list of eigenvectors in vecin */
  *vecin=spectrumvec; 
  *vals=spectrumval;
  return;

  /* Yes, there are rare occasions where a goto is OK....  */
error: 
  fprintf(stderr,"Error in spectrum().\n");
  free(work);

#if USE_ARPACK   
  free(select);
#else
  free(ind);
#endif

  *vecin=NULL;
  *vals=NULL;
  return;
}

/*
  free() all memory used in spectrum 
   On the CRAY, the various quantities should be
   deallocated in reverse order of allocation.
   
   This is only of use for reducing memory fragmentation
   if freemahk() is also called.   This is because resthk() 
   calls mahk().
*/

void freespectrum(void){
  free(spectrumval); free(spectrumvec); free(qstore);
  spectrumval=NULL;  spectrumvec=NULL; qstore=NULL;
  lval=0;            lvec=0;           lqstore=0;
/* Free memory used by the structure els. */
  free(els.j);  free(els.e); free(els.i); 
  els.j=NULL;   els.e=NULL;  els.i=NULL;  
  els.jchunk=0; els.tchunk=0;            
}

/* subroutine to print out memory use of the structure els */

void printhels(FILE *fp){
  fprintf(fp,"MM els.e, els.i have length %i; els.j length is %i\n",
	  els.tchunk,els.jchunk);
  fprintf(fp,"MM (element)=%i and (size_t)=%i bytes, total is %i bytes\n",
	  sizeof(element),sizeof(size_t),(sizeof(element)+sizeof(size_t))*
	  els.tchunk+els.jchunk*sizeof(size_t));
}

element fdot(const element *x,const element *y,int n){
  int i;
  element t=0.0;
  for(i=0; i<n; i++)t += x[i]*y[i];
  return t;
}

int dot(const int *x,const int *y,int n){
  int i,t=0;
  for(i=0; i<n; i++)t += x[i]*y[i];
  return t;
}

/*************************************************************************/

/*  This is the subroutine that stores vectors for dnlaso.
    See the documentation file "laso.doc"      */

#define DEBUGIO 0 /* flag to print out debug statements */

extern fsub iovect(integer *n, integer *m, doublereal *q, integer *j, 
		  integer *k){
  integer i, l, l1;

  if(*k==0){
    for(l=0; l<*m; l++){
      l1 = *j - *m + l;
#if defined(__uxp__)
#pragma loop noalias qstore
#pragma loop disjoint qstore
#endif
      for(i=0; i<*n; i++){
	qstore[i+l1 * *n]=q[i+l * *n];
      }
    }
    return SUBRETURN;
  }else {
    for(l=0; l<*m; l++){
      l1 = *j - *m + l;
#if defined(__uxp__)
#pragma loop noalias q
#pragma loop disjoint q
#endif
      for (i=0; i<*n; i++){
	q[i + l * *n] = qstore[i + l1 * *n];
      }
    }
    return SUBRETURN;
  }
}

/* This prints out the memory used by qstore */

void printlaso(FILE *fp){
  fprintf(fp,"MM laso array qstore, length %i, uses %i bytes.\n",
	  lqstore,lqstore*sizeof(doublereal));
}


/**********************************************************************

    Hamiltonian multiply used by laso and indirectly by the NAG routine.
    This is the largest time user in a long run.

    The running times for matrix diagonalization shown below are 
    for HINDEX=2, MOM=1, and

       int ht[HINDEX]={0,0},np=6,kt=20,cyclic=1,o=1,
           multi=9,multi2=8,nfig=8;
       integer nval=5;

    in the program "runlaso.c" 
    Using DLCQ matrix elements, IMPROVED=1 in "include.h".

    also enter values nval=5 and momentum=1 0.
    couplings: 0.03 1 0.7 0.03 -0.005 3 0.13 1 -2 1
    639674 nonzero matrix elements, 14439 states, Lanczos with DNLASO
    102*3 matrix multiplies.

    Eigenvalues:   
     1.26979644488890e+00
     1.87282569597738e+00
     3.14511664852520e+00
     3.21836991539210e+00
     3.30083925991769e+00

*/



extern fsub elop(integer *n, integer *m, doublereal *p, doublereal *q){
  integer k,i,j;
  size_t *els_j;
#if  defined(__uxp__) /* Use more pointers: els_i for els[i], etc. */
  /* On vpp, it appears that structures inhibit vectorization,
     thus we define pointers that do the same thing  */ 
  size_t *els_i;
  doublereal *els_e,ptemp,qtemp;
#endif
  if(els.n!= *n){
    fprintf(stderr,"dimensions don't agree in elop\n");
    exit(19);
  }

#if 0  /*Try Fortran subroutine instead */
/*  In all of these, I changed the type for elslist from 
    size_t to long

  time:   alk,  vodka, 179 rum; GNU C Power PC
         (generating the matrix takes 35.)

  time:  21 quad IBM SP2 (generating matrix takes 31)  
         (same speed as the C routine!)

  time:  crashes babbage; Hitachi (generating matrix takes 60)
         300 Mflops peak.

  time:   gsm CONVEX (generating matrix takes 52)
         240 Mflops peak.

  time:   vpp; Fujitsu (generating matrix takes 82), however
         above -- with vectorization -- is much faster.

  time:   t3e; CRAY (generating matrix takes 25) 
         600 Mflops peak, 1.2 Gbytes/sec. peak memory access.
*/
  FORELOP(n,m,p,q,els.e,els.i,els.j,&(els.n));

#else /* use C version of elop */

  memset(q,0,*n * *m * sizeof(doublereal));
#if defined(__uxp__) /* Use more pointers:  els_i for els[i], etc. */
   /* Both pragma statements are needed to vectorize on vpp.

      It is difficult to vectorize both the loop over k and
      the loop over j on vpp.  I tried changing the loop 
      increment for k to 1 and moving it inside the i and 
      j loops but the q[els_i[j]] += els_e[j]*p[i]; 
      always remained scaler, although both j and k loop
      variables themselves vectorized.  

      Using els_i and els_e gives modest speedup on CRAY.
      Using pragma _CRI cache_bypass q, p slows things
      down on the CRAY T3E.

      Using pipelining (flag -h pipeline1) gives very slight 
      speedup on CRAY T3E.    

  time: 20 vpp; Fujitsu (generating matrix takes 82)

  time: 42 t3e; CRAY T3E (generating matrix takes 25)

  time: 71 rum; GNU C (generating matrix takes 34), same as below

*/
  for(k=0; k<*m; k++){
    els_j=els.j;
    els_i=els.i;
    els_e=els.e;
    for(i=0; i<*n; i++){
      j=*(els_j++);
      ptemp=p[i]; 
      qtemp=0.0;
      /* see if this row has a diagonal */
      /* Both conditions are needed for the last element of the matrix */
      if(j<*els_j && *els_i==i){
	qtemp += *(els_e++)*ptemp;
	els_i++; j++;
      }
#pragma loop noalias q
#pragma loop disjoint q
#if defined(_CRI)
#pragma _CRI ivdep
#endif
      for(j=j; j<*els_j; j++){
	q[*els_i] += *els_e*ptemp;
	qtemp += *(els_e++)*p[*(els_i++)];
      }
      q[i]+=qtemp;
    }
    p += *n; 
    q += *n;
  }
#else /* Use more pointers:  els_i for els[i], etc. */
/* case for GNU_C:
   The form shown below is chosen because
   it is the simplest.  There are a number of
   possible changes that do not affect speed in
   any substantial manner:
   
   Interchanging order of the loops.
   Using a temporary variable for p[i] and q[i]
   Using pointers for els.i[j] and els.e[j] (slightly slower)
   Using j=els.j[i]; top=els.j[i+1]; instead of j=*(els_j++);
   Introducing jtemp=els.j[i];
   splitting into two loops decreases speed slightly.

  time:  24.4 sec on barak (generating matrix takes 10.4) April 2001
         benchmark stream gives 250 246 280 233 MB/s.
         (Tyan 1834 motherboard, 667MHz Pentium III CPU)

  time:  30 sec on bvds.geneva.edu (generating matrix takes 15) April 2000
         benchmark STREAM gives 282, 279, 330, 248 MB/s.
         (Tyan 1830 motherboard, 400 MHz Pentium II CPU)

  time:  66 alk, 81 vodka, 71 rum; GNU C Power PC
         (generating the matrix takes 35.)

  time:  21 quad IBM SP2 (generating matrix takes 31)  

  time:  62 gsm CONVEX SPP (generating matrix takes 52)
         240 Mflops peak.

  time:  69 babbage; Hitachi (generating matrix takes 60)
         300 Mflops peak.

  time:  108 vpp; Fujitsu (generating matrix takes 82), however
         above -- with vectorization -- is much faster.

  time:  44 t3e.osc.edu; CRAY (generating matrix takes 25)  April 2000
         When ivdep pragma is not used, 58.
         using pragma cache_bypass always slows things down.
         600 Mflops peak, 1.2 Gbytes/sec. peak memory access.

  time:  34 T3E jaromir.psc.edu; CRAY (generating matrix takes 17) April 2000
         Used ivdep pragma below.  The benchmark STREAM gives 
         443, 450, 523, 521 MB/s.

*/
  for(k=0; k<*m; k++){
    for(i=0, els_j=els.j; i<*n; i++){
      j=*(els_j++);
      if(els.i[j]==i && j<*els_j)
	q[i] += els.e[j++]*p[i];
#ifdef _CRI
#pragma _CRI ivdep
#endif
      for(; j<*els_j; j++){
        q[els.i[j]] += els.e[j]*p[i];
	q[i] += els.e[j]*p[els.i[j]];
      }
    }
    p += *n; 
    q += *n;
  }
#endif /* Use more pointers:  els_i for els[i], etc. */

#endif /* use C version of elop */

  return SUBRETURN;
}



/*  The following is a fragment of code which calculates
    matrix elements of the Hamiltonian as one does a Lanczos
    step.  It is substantially slower than precomputaton
    of nonzero matrix elements and should only be used
    in cases where the number of nonzero matrix elements
    uses up too much memory.                              */

/* The following must be defined globally: */
int kt,multi,o;
sectors x;
element couplings[NPARAMS]; 

extern fsub vectop(integer *n, integer *m, doublereal *p, doublereal *q){
/*  printf("starting vectop m=%li\n",*m);  */
  hamvector(&x,q,p,*m,couplings);
/*  printf("finished vectop\n"); */
  return SUBRETURN;
}

#endif






