#include <stdio.h>
#include <math.h>
#include "include.h"
#include "run.h"
#include "usempi.h"

#if ONE_PROCESSOR_SPECTRUM
#ifdef _CRAY  /* this is for the mallopt() command below */
#include <malloc.h>
#endif

#define MS_PRINT 0 /* flag to print out debugging information to file */
                   /* Each process gets a unique file */

#define OFFSET 2 /* offset for tags number for normal tasks */
                 /* 0 quit 
                    1 run mapcouplings */

#define FANCY 0   /* flag for sending format */
                  /* In the fancy format, the entire basis
                     definition is sent to the slave as one
		     big structure.  Else, several "sends" are used. */

/* mpisectors, and mpitype are initialized in
   initmpisectors below: this must be called by the main program.*/

#if FANCY /* flag for sending format */
#define MPITYPELENGTH 4
static int mpilength[MPITYPELENGTH]={1,1,HINDEX+2,1};
static MPI_Datatype mpitype[MPITYPELENGTH]={
  MPI_DATATYPE_NULL,MPI_DATATYPE_NULL,MPI_DOUBLE,MPI_INT};
#endif
static MPI_Datatype mpisectors;
static int mpil2[2]={NPARAMS,1};
static MPI_Datatype mpit2[2]={MPI_DOUBLE,MPI_INT};

/* This subroutine initializes the structure mpisectors
   (which is analogous to the "c" structure sectors)
   and puts the results into the first two
   elements of the global variable mpitype.

   It also initializes the communicator mpiallslaves
*/

void initmpisectors(void){
  int i,seclength[6]={HINDEX,1,1,1,1,1};
  sectors mx;
  MPI_Aint secdisp[6],sec0;
  int master[1]={0};
  MPI_Group g1,g2;

  /* make MPI datatype for the structure sectors */
  /* note that this only works for part of the structure,
     the basis etc. is not included. */
  MPI_Address(mx.ht,secdisp);
  MPI_Address(&mx.npmax,secdisp+1);
  MPI_Address(&mx.kt,secdisp+2);
  MPI_Address(&mx.cyclic,secdisp+3);
  MPI_Address(&mx.o,secdisp+4);
  MPI_Address(&mx.multi,secdisp+5);

  /* we find offsets relative to the "base" of the structure,
     wherever that may be (thus we make no assumptions of order. */
  MPI_Address(&mx,&sec0);
  for(i=0; i<6; i++)secdisp[i]-=sec0;
  MPI_Type_hindexed(6,seclength,secdisp,MPI_INT,&mpisectors);
  MPI_Type_commit(&mpisectors);
  /* use this in the global variable mpitype */
#if FANCY /* flag for sending format */
  mpitype[0] = mpitype[1] = mpisectors;
#endif

#if ONE_PROCESSOR_SPECTRUM

/*
  Initialize the communicator mpiallslaves which
  is a subgroup consisting of just the slaves.
  This is used in file "improved.c"  
*/

  MPI_Comm_group(MPI_COMM_WORLD,&g1);
  MPI_Group_excl(g1,1,master,&g2);
  MPI_Comm_create(MPI_COMM_WORLD,g2,&mpiallslaves);
  MPI_Group_free(&g1);
  MPI_Group_free(&g2);

#endif

}

/* This is the slave process used by MPI 
   it is a "stand-alone" subroutine to calculate the
   lowest eigenvalues for a given basis.  It also
   calculates the basis when necessary. 
   angle contains HINDEX components of transverse momentum,
   the kmax cutoff for nonzero L and L itself.
*/

#define MAXBASIS 150

void slave(void){
  integer nfig=EDIGITS; /* number of significant digits in eigenvalues */
  static sectors a[2];
  int nvals,globalkt,myrank,gme;
  element *val,*vec,couplings[NPARAMS],angle[HINDEX+2];
  size_t gmb;
#if MS_PRINT 
  FILE *fp;
  char out[10];
  int i,j;
#endif  
  MPI_Status status;
  MPI_Datatype xt2;
  MPI_Aint d2[2];
#if FANCY
  MPI_Datatype xtype;
  MPI_Aint disp[MPITYPELENGTH];

  MPI_Address(a,disp);
  MPI_Address(a+1,disp+1);
  MPI_Address(angle,disp+2);
  MPI_Address(&nvals,disp+3);
  MPI_Type_struct(MPITYPELENGTH,mpilength,disp,mpitype,&xtype);
  MPI_Type_commit(&xtype);
#endif
  MPI_Address(couplings,d2);
  MPI_Address(&globalkt,d2+1);
  MPI_Type_struct(2,mpil2,d2,mpit2,&xt2);
  MPI_Type_commit(&xt2);
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);

#if MS_PRINT /* open debugging file with name based on rank */
  sprintf(out,"p%i.tmp",myrank);
  if((fp = fopen(out,"w")) == NULL){
    fprintf(stderr,"Can't open file %s in usempi.c, exiting\n",out);
    goto error;
  }
#endif

  /* on CRAY, be more carefull about memory management.  */

#ifdef _CRAY
      mallopt(M_LOWFIT,1);
#endif

      /* Main loop of program */

  for(;;){

    MPI_Probe(MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&status);

    /* Stop program */
    
    if(status.MPI_TAG==0)goto exit;

    /* Receive coupling constants and global kt. */

    else if(status.MPI_TAG==1)
      MPI_Recv(MPI_BOTTOM,1,xt2,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
   
    /* Calculate spectrum */

    else {    
#if MS_PRINT /* diagnostic print to debugging file*/
      fprintf(fp,"About to receive Job %i\n",status.MPI_TAG-OFFSET);
      fflush(fp);
#endif
#if FANCY /* flag for sending format */
      MPI_Recv(MPI_BOTTOM,1,xtype,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
#else
      MPI_Recv(a,1,mpisectors,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
      MPI_Recv(a+1,1,mpisectors,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
      MPI_Recv(angle,HINDEX+2,MPI_DOUBLE,0,MPI_ANY_TAG,
	       MPI_COMM_WORLD,&status);
      MPI_Recv(&nvals,1,MPI_INT,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
#endif
#if MS_PRINT /* diagnostic print to debugging file */
      fprintf(fp,"Received job %i, nvals=%i\n",
	      status.MPI_TAG-OFFSET,nvals);
      for(j=0; j<2; j++){
	fprintf(fp,"  ht=(");
	for(i=0; i<HINDEX; i++)fprintf(fp," %i",a[j].ht[i]);
	fprintf(fp,"), np=%i kt=%i cyclic=%i o=%i and multi=%i,\n",
		a[j].npmax,a[j].kt,a[j].cyclic,a[j].o,a[j].multi);
      }
      fprintf(fp,"  angles=");
      for(j=0; j<HINDEX+2; j++)fprintf(fp," %g",angle[j]);
      fprintf(fp,"\n");
      for(j=0; j<2; j++)
	if(a[j].cyclic<0||a[j].cyclic>1||a[j].o<-1||a[j].o>1){
	  fprintf(stderr,"error in values sent, exiting\n");
	  goto error;
	}
#endif

/*
    construct one or two bases.  The order for makebasis() 
    and rebasis() is important for memory fragmentation 
    problems on the CRAY. 
    mapcouplings should have already been run.

    Now calculate the actual spectrum 
*/

      spectrum(a,a+1,&val,&vec,nvals,couplings,angle,globalkt,nfig);

      if(val==NULL){
	fprintf(stderr,"slave() %i detecting an error in spectrum()\n",
		myrank);
	goto error;
      }
/*
    Now, for the J^++ sector, we need to order the 0++*
    and 2++ according to particle number content 
*/
      shuffle(a,a+1,vec,val,nvals);

#if MS_PRINT /* diagnositic print */
      fprintf(fp,"  values: ");
      for(i=0; i<nvals; i++)fprintf(fp,"%.10g ",val[i]);
      fprintf(fp,"\n");
#endif

      MPI_Send(val,nvals,MPI_DOUBLE,0,status.MPI_TAG,MPI_COMM_WORLD);

/*
       To prevent memory fragmentation problems on systems
       without virtual memory paging (like the CRAY), we
       deallocate the memory used by spectrum() in reverse order.

       The other two things that can be done  to minimize
       memory fragmentation is to choose
       sufficiently large values for MAX_ELS and MAX_BASE.
*/

#if defined(_CRAY)
      freespectrum();
#endif
      freebasis(a+1);
      freebasis(a);
    }
  
    /* for normal case, just do next loop iteration */
#if MS_PRINT /* if everything is OK, write over old info */
    rewind(fp);
#endif
  }
    
error:  /* On error, print out memory usage and tell the master. */
/* The master will then send out a command that everyone quit. */
  fprintf(stderr,"Error in slave %i after receiving job %i, exiting.\n",
	  myrank,status.MPI_TAG);
#if MS_PRINT
  printhels(fp);
  printelements(fp);
  printlaso(fp);
#endif
#ifdef _CRAY
#if 0 /* Choose one of the following */
  printf("MM Running malloc_stats(0), slave %i:\n",myrank);
  malloc_stats(0);
#elif 0
  printf("MM Running malloc_stats(1), slave %i:\n",myrank);
  malloc_stats(1);
#elif 1  /* the following is very long */
  printf("MM Running malloc_stats(2), slave %i:\n",myrank);
  malloc_stats(2);
#endif
#endif
  /* Just the message tag is read by the master */
  MPI_Send(&myrank,1,MPI_INT,0,0,MPI_COMM_WORLD);
  
exit:  /* Normal exit */
#if FANCY
  MPI_Type_free(&xtype);
#endif
  MPI_Type_free(&xt2);
  MPI_Comm_rank(mpiallslaves,&myrank);
  MPI_Reduce(&max_els,&gme,1,MPI_UNSIGNED_LONG,MPI_MAX,0,mpiallslaves);
  MPI_Reduce(&max_base,&gmb,1,MPI_UNSIGNED_LONG,MPI_MAX,0,mpiallslaves);
  if(myrank==0){
    printf("Global maximum of max_els is %i, max_els=%i\n",
	   gme,max_els);
    printf("Global maximum of max_base is %i, max_base=%i\n",
	   gmb,max_base);
  }
#if MS_PRINT /* debugging file with name based on rank */
  fclose(fp);
#endif
  return;
}

/*  This subroutine gives tasks to the various slaves. 
    Eigenvalues are output in the vector vals whose
    memory should be already allocated.  vals[whichvals[i]+j]
    contains the jth eigenvalue, j=0,1,2... for job i,
    i=1,2,...njob. 

    Since all the jobs are assumed to have the same couplings,
    the generation of improved matrix elements is parallelized.

    master returns 0 on successful completion.
*/


int master(sectors *(*base)[2], element *couplings, 
	    element (*angle)[HINDEX+2], int *nvals,
	    int *whichvals, element *vals, int njob){
  int job,receive,slave,nslaves,globalkt,j;
#if MS_PRINT /* diagnostic print to debugging file */
  FILE *fp;
  int i,k;
  char out[10]="p0.tmp";
#endif  
#if FANCY /* flag for sending format */
  MPI_Datatype xtype;
  MPI_Aint disp[MPITYPELENGTH];
#endif
  MPI_Datatype xt2;
  MPI_Aint d2[2];
  MPI_Status status,status2;

  MPI_Address(couplings,d2);
  MPI_Address(&globalkt,d2+1);
  MPI_Type_struct(2,mpil2,d2,mpit2,&xt2);
  MPI_Comm_size(MPI_COMM_WORLD,&nslaves);
  nslaves--; /* don't include master amongst slaves */
  

#if MS_PRINT /* open debugging file*/
  if((fp = fopen(out,"w")) == NULL){
    fprintf(stderr,"Can't open file %s in usempi.c, exiting\n",out);
    MPI_Type_free(&xt2);
    return 1;
  }
#endif

/*
  Loop through all jobs and find maximum kt.  We calculate improved
  matrix elements based on the global largest value of kt.
  Send out globalkt and the coupling constants to the slaves
*/
  globalkt=0;
  for(j=0, globalkt=0; j<njob; j++)
    if(abs(base[j][0]->kt)>globalkt)globalkt=abs(base[j][0]->kt);
#if 0 /* debugging print */
  printf("global maxmimum kt is %i\n",globalkt);
#endif
  MPI_Type_commit(&xt2);
  for(j=0; j<nslaves; j++)
    MPI_Send(MPI_BOTTOM,1,xt2,j+1,1,MPI_COMM_WORLD);
  MPI_Type_free(&xt2);
  
  /* loop through all jobs that need completion */
  for(job=0, receive=0; receive<njob;){

    /* Receive completed jobs once all slaves have been seeded. */
    if(job<nslaves)
      slave=job+1;  /* slaves are numbered 1,2,3,... */
    else{
      MPI_Probe(MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
      j=status.MPI_TAG-OFFSET; /* remove offset in tag number */
      if(status.MPI_TAG<OFFSET||j>=njob){
	fprintf(stderr,"Master receiving tag=%i from slave %i, exiting\n",
		status.MPI_TAG,status.MPI_SOURCE);
#if MS_PRINT /* debugging file with name based on rank */
	fclose(fp);
#endif
	return 1;
      }
      MPI_Recv(vals+whichvals[j],nvals[j],
	       MPI_DOUBLE,MPI_ANY_SOURCE,
	       MPI_ANY_TAG,MPI_COMM_WORLD,&status2);
#if 0 /* debugging print: print jobs received */
      if(status.MPI_TAG<10){
	printf("Master receiving job %i from slave %i offset %i;\n ",
	       status.MPI_TAG,status.MPI_SOURCE,whichvals[status.MPI_TAG]);
	for(i=0; i<nvals[j]; i++)
	  printf(" %f",vals[whichvals[j]+i]);
	printf("\n");
      }
#endif
      slave=status.MPI_SOURCE;
      receive++;
    }

    /* send jobs to the slaves */
    if(job<njob){
#if MS_PRINT /* debugging print to file: print jobs sent to slaves. */
      fprintf(fp,"Master sending job=%i to slave %i, nvals=%i\n",
	      job,slave,nvals[job]);
      for(k=0; k<2; k++){
	fprintf(fp,"  ht=(");
	for(i=0; i<HINDEX; i++)fprintf(fp," %i",base[job][k]->ht[i]);
	fprintf(fp,"), np=%i kt=%i cyclic=%i o=%i and multi=%i,\n",
		base[job][k]->npmax,base[job][k]->kt,base[job][k]->cyclic,
		base[job][k]->o,base[job][k]->multi);
      }
      fprintf(fp,"  angle=");
      for(i=0; i<HINDEX+2; i++)fprintf(fp," %f",angle[job][i]);
      fprintf(fp," nvals=%i\n",nvals[job]);
      fflush(fp);
#endif
      /* include offset of one in tag number */

#if FANCY /* flag for sending format */
      MPI_Address(base[job][0],disp);
      MPI_Address(base[job][1],disp+1);
      MPI_Address(angle[job],disp+2);
      MPI_Address(nvals+job,disp+3);
      MPI_Type_struct(MPITYPELENGTH,mpilength,disp,mpitype,&xtype);
      MPI_Type_commit(&xtype);
      MPI_Send(MPI_BOTTOM,1,xtype,slave,job+OFFSET,MPI_COMM_WORLD);
      MPI_Type_free(&xtype);
#else
      /* We need two send commands because the base[job][0] and
         base[job][1] are not necessarily contiguous in memory */
      MPI_Send(base[job][0],1,mpisectors,slave,job+OFFSET,MPI_COMM_WORLD);
      MPI_Send(base[job][1],1,mpisectors,slave,job+OFFSET,MPI_COMM_WORLD);
      MPI_Send(angle[job],HINDEX+2,MPI_DOUBLE,slave,
	       job+OFFSET,MPI_COMM_WORLD);
      MPI_Send(nvals+job,1,MPI_INT,slave,job+OFFSET,MPI_COMM_WORLD);
#endif
      job++;
    }
  }

  /* normal return, no error */
#if MS_PRINT /* open debugging file with name based on rank */
  fclose(fp);
#endif
  return 0;  
}

#endif












