#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include "spectrum.h"
#include "jessica.h"
#include "extrapolate.h"
#if defined(USE_MPI)
#include "usempi.h"
#endif

#if HINDEX==2
#define PRINTIT 1 /*  Print timing and convergence information  */

/*
              Glueballs in 3+1 dimensions
              on the transverse lattice

    This program searches coupling constant space for a best fit
    to Poincare symmetries that are broken by the transverse lattice
    regulator.  

    We test for rotational invariance in the transverse plane.  
    In practice, we will choose the x^1 direction and the x^1=x^2 
    directions in order to take advandage of residual reflection 
    symmetries.

    This version also demands degeneracy in the lowest parity doublet.
    It also uses the string tension to determine the lattice spacing.


    We measure:                                                         multi

    p-truncation convergence for n=(1,0) winding mode                   15
    longitudinal potential for n=(0,0) and various L                     3
    longitudinal eigenvalue for n=(1,0) and one L                       15
    longitudinal eigenvalue for n=(1,1) and one L                       17
    winding mode potential for various n=(c,0)                          15
    winding mode potential for one n=(c,c)                              17
    spectrum for P_perp = (0,0) and (c,0)                               9, 8
                                                                        7, 10
    spectrum for P_perp = (c,c)                                         11, 12
                                                                        14, 13


************************************+++++++**********************************/

#define FTEPER 1  /* Number of Teper's lattice states to fit. */
                 /* This can be 0 or 1  */
#define FCSQUARED NPERPVALS /* Number of c^2 values in each sector and
                    direction to use in fit. This can be up to NPERPVALS */
#define FPARITY 3/* Number of parity doublets to fit. */
                 /* Suggested: 0, 2, 3, or 5. */
#define FSCALE 1 /*Whether to include consistantcy of the */
                 /*longitudinal potential with the overall scale.
		   The potential itself is calculated anyway.
		   This can be 0 or 1. */
#define FLONG MAX_LONG /* Number of longitudinal points to fit */
                /* This can be 0 to MAX_LONG. */
#define FWIND 1 /* number of winding spectra to fit
		   This can be 0 or 1. */
#define FEXTRA 0 /* Fit criteria that would not be normally used. */  
                   /*For instance, demanding 
                     a specific value for the lattice spacing.  */
#define NPERP 4 /* Number of J^{P_1 C} sectors for nonzero perp momentum 
		   this must be 4 to cover all multiplets */
#define NPERPVALS 4 /* number of states calculated in each sector */
#define NP1 1 /* Number of nonzero values of P_perp in the (c,0) direction*/
#define NP2 1 /* Number of nonzero values of P_perp in the (c,c) direction*/
#define NSTILL 10 /*Number of J^{P_1 C} sectors for zero transverse momentum spectra
                    This must be 10 to cover all multiplets. */ 
#define NSTILLVALS 4  /*Number of states calculated in each sector */
                      /*This must be large enough to use for nonzero momenta */
#define MAXTH 10 /* max number of terms in fit functions */
#define NLONG 3  /*Data points used to estimate longitudinal potential */
#define NWIND 3 /* Data points used to estimate winding potential */
#define MAX_LONG 3 /* max number of criteria for roundness of long. pot.*/


#if 0 /* small basis */
#if GLUE_PERIODIC 
/* this takes about 4.0 seconds to calculate everything*/
#define NS 5
const static int kt[NS]={14,14,16,18,20}, np[NS]={4,6,4,4,4};
#define WNS 5
static int wkt0[NWIND][WNS]={{12,12,14,16,18},{10,10,12,14,16},
			     {10,10,12,14,16}}, 
  wnp0[NWIND][WNS]={{4,6,4,4,4},{5,7,5,5,5},{6,8,6,6,6}},
  wht0[NWIND][HINDEX]={{2,0},{3,0},{4,0}};
static int wkt[1][WNS]={{10,10,12,14,16}},wnp[1][WNS]={{4,6,4,4,4}},
    wht[1][HINDEX]={{2,2}};
#define LNS 5
static int lkt0[LNS]={12,12,12,14,16}, lnp0[LNS]={2,4,2,2,2},
  lkt[MAX_LONG][LNS]={{8,8,8,10,12},{8,8,8,10,12},
    {8,8,8,10,12}}, 
    lnp[MAX_LONG][LNS]={{3,5,3,3,3},{3,5,3,3,3},{2,4,2,2,2}};
doublereal lkmax0[LNS]={3.0,3.0,2.0,3.0,3.0},
      lkmax[MAX_LONG][LNS]={{3.0,3.0,2.0,3.0,3.0},
        {3.0,3.0,2.0,3.0,3.0},{3.0,3.0,2.0,3.0,3.0}},
	llong0[NLONG]={3.0,4.0,5.0},llong[MAX_LONG]={3.0,1.0,1.0},
	llong0m[NLONG]={0.1,0.1,0.1},llongm[MAX_LONG]={0.1,0.1,0.1};
#define PNS 4 /* number of sectors in p extrapolation */
static int pnp1[PNS]={3,5,7,9},pkt1=18;

#else

/* this takes about 45 seconds to calculate everything*/
#define NS 4
const static int kt[NS]={14,12,14,16}, np[NS]={4,6,6,4};
#define WNS 3
static int wkt0[NWIND][WNS]={{12,14,16},{9,11,13},{10,12,14}}, 
  wnp0[NWIND][WNS]={{6,4,4},{7,5,5},{8,6,6}},
  wht0[NWIND][HINDEX]={{2,0},{3,0},{4,0}};
static int wkt[1][WNS]={{12,14,16}},wnp[1][WNS]={{6,4,4}},
    wht[1][HINDEX]={{2,2}};
#define LNS 4
static int lkt0[LNS]={12,12,12,16}, lnp0[LNS]={2,4,2,2},
  lkt[MAX_LONG][LNS]={{7,7,7,11},{7,7,7,11},
    {8,8,10,12}}, 
    lnp[MAX_LONG][LNS]={{3,5,3,3},{3,5,3,3},{2,4,2,2}};
doublereal lkmax0[LNS]={3.0,3.0,2.0,3.0},
      lkmax[MAX_LONG][LNS]={{3.0,3.0,2.0,3.0},
        {3.0,3.0,2.0,3.0},{3.0,3.0,2.0,3.0}},
	llong0[NLONG]={3.0,4.0,5.0},llong[MAX_LONG]={3.0,1.0,1.0},
	llong0m[NLONG]={0.1,0.1,0.1},llongm[MAX_LONG]={0.1,0.1,0.1};
#define PNS 3 /* number of sectors in p extrapolation */
static int pnp1[PNS]={3,5,7},pkt1=9;
#endif

#elif 0 /************* larger basis ***************************/
/* this takes about 12 minutes to calculate everything*/
#define NS 4
const static int kt[NS]={12,12,14,18}, np[NS]={8,6,6,6};
#define WNS 3
#if GLUE_PERIODIC
static int wkt0[NWIND][WNS]={{16,16,26},{18,18,26},{16,16,26}};
#else 
static int wkt0[NWIND][WNS]={{16,16,26},{17,17,25},{16,16,26}}; 
#endif
static int wnp0[NWIND][WNS]={{6,4,4},{7,5,5},{8,6,6}},
  wht0[NWIND][HINDEX]={{2,0},{3,0},{4,0}};
static int wkt[1][WNS]={{14,14,22}},wnp[1][WNS]={{8,6,6}},
    wht[1][HINDEX]={{2,2}};
#define LNS 7
static int lkt0[LNS]={26,26,26,42,42,70,70}, lnp0[LNS]={2,4,2,2,2,2,2};
#if GLUE_PERIODIC
static int lkt[MAX_LONG][LNS]={{14,14,14,24,24,36,36},
		     {14,14,14,24,24,36,36},
			       {22,22,22,44,44,70,70}};
#else
static int lkt[MAX_LONG][LNS]={{13,13,13,23,23,35,35},
		     {13,13,13,23,23,35,35},
			       {22,22,22,44,44,70,70}}; 
#endif
static int lnp[MAX_LONG][LNS]={{3,5,3,3,3,3,3},{3,5,3,3,3,3,3},
		       {2,4,2,2,2,2,2}};
doublereal lkmax0[LNS]={3.5,3.5,3.0,3.5,4.25,4.0,4.5},
      lkmax[MAX_LONG][LNS]={{3.5,3.5,3.0,4.0,4.5,5.0,6.0},
       {3.5,3.5,3.0,4.0,4.5,5.0,6.0},{5.0,5.0,4.5,5.0,6.0,6.0,7.0}},
      llong0[NLONG]={3.577,5.058,7.154},llong0m[NLONG]={-2.726,-3.856,-5.452},
      llong[MAX_LONG]={1.265,5.058,0.0},llongm[MAX_LONG]={-0.964,-3.856,0.0};
#define PNS 4 /* number of sectors in p extrapolation */
static int pnp1[PNS]={3,5,7,9};
#if GLUE_PERIODIC
static int pkt1=20;
#else
static int pkt1=15;
#endif

#elif 1 /*************** BIG basis *************************/
/* this takes about 6.654 hours to calculate everything on PC's,
   4.14 CPU-hours on the CRAY T3E.
   Geneva 2002, periodic BC's, DLCQ:  49 CPU minutes.  
   barak 2003, periodic BC's, DLCQ: 4.52 CPU-hours
   Opterons 2003, periodic BC's, DLCQ: 90 CPU-minutes
   max_els=3101540, max_base=2255950 (but this could grow...)  */
#define NS 6
/*            was    {16,16,20,26}
              2002:  {20,20,24,30}  
              2003:  {20,20,24,28,32,40}
*/
const static int kt[NS]={20,20,24,28,32,40}, np[NS]={6,8,6,6,6,6};
#define NWIND 3
#if GLUE_PERIODIC
#define WNS 6
static int wkt0[NWIND][WNS]={{20,20,24,28,32,40},{20,20,24,28,32,40},
			     {20,20,24,28,32,40}}; 
static int wkt[1][WNS]={{20,20,24,28,32,40}},wnp[1][WNS]={{6,8,6,6,6,6}}; 
#else
#define WNS 5
static int wkt0[NWIND][WNS]={{20,20,24,28,32},{19,19,23,27,33},
			     {18,18,22,26,32}};
static int wkt[1][WNS]={{16,16,20,24,28}},wnp[1][WNS]={{6,8,6,6,6}}; 
#endif
static int wnp0[NWIND][WNS]={{6,8,6,6,6,6},{5,7,5,5,5,5},{6,8,6,6,6,6}},
  wht0[NWIND][HINDEX]={{2,0},{3,0},{4,0}}, wht[1][HINDEX]={{2,2}};
#define LNS 7
static int lkt0[LNS]={40,40,40,60,60,100,100},
  lnp0[LNS]={2,4,2,2,2,2,2};
#if GLUE_PERIODIC
static int lkt[MAX_LONG][LNS]={{30,30,30,60,60,100,100},
			 {30,30,30,60,60,100,100},
			       {30,30,30,30,30,100,100}};
#else 
static int lkt[MAX_LONG][LNS]={{19,19,19,33,33,69,69},
			 {19,19,19,33,33,69,69},
			       {34,34,34,48,48,70,70}};
#endif
int lnp[MAX_LONG][LNS]={{3,5,3,3,3,3,3},{3,5,3,3,3,3,3},{2,4,2,2,2,2,2}};
doublereal lkmax0[LNS]={4.0,4.0,3.5,3.75,4.25,4.0,4.5},
      lkmax[MAX_LONG][LNS]={{4.0,4.0,3.5,4.0,5.0,5.0,6.0},
       {4.0,4.0,3.5,4.0,5.0,5.0,6.0},{5.5,5.5,5.0,5.5,6.5,6.0,7.0}},
      llong0[NLONG]={3.577,5.058,7.154},llong0m[NLONG]={-2.726,-3.856,-5.452},
      llong[MAX_LONG]={1.265,5.058,0.0},llongm[MAX_LONG]={-0.964,-3.856,0.0};
#define PNS 3 /* number of sectors in p extrapolation */
static int pnp1[PNS]={5,7,9};
#if GLUE_PERIODIC
static int pkt1=24;
#else
static int pkt1=17;
#endif

#endif  /*************** Choose different possible bases *****************/

/* Total number of different bases */
#define N_ALL_BASES ((NSTILL+(NP1+NP2)*NPERP)*NS+(NWIND+FWIND)*WNS+\
                    (FLONG+NLONG)*LNS+PNS)
static specify_calculation all_bases[N_ALL_BASES];
static specify_calculation *s[NP1][NPERP],*t[NP2][NPERP],
  *ws0[NWIND],*ws[FWIND],
  *ls[FLONG],*ls0[NLONG],*ps1,*stillbases[NSTILL];

element par[NPARAMS];

             /* This is the lattice value for the 0^++ in units of the 
                string tension extrapolated to large N.  I assume that
                the large N extrapolation is the same as for the 2+1 glueball.
                The following is the mass formula:
                 1648./440(1-1/4.11* 2.379/3^2)
                The problem here is that the errors are too large to
                really constrain the value for anything other than
		the lowest state.  */
#if FTEPER<2
doublereal teper[1]={12.67},tepererr[1]={1.28};
#endif

static int wpar[NPARAMS],scalemethod;
#if NP1<2 && NP2<2  /* check dimensions */
static doublereal p1a[1][HINDEX]={{0.25,0.0}},
  p2a[1][HINDEX]={{0.25,0.25}};
#endif
static int pht1[HINDEX]={1,0};  /* winding number for p-extrapolation */
int multis[NPERP]={15,15,16,16}, /* multi for nonzero P_perp */
  chargest[NPERP]={1,-1,-1,1}, /* charge conjugation for nonzero P_perp */
      multit[NPERP]={17,17,18,18};

/* The error for: 
     c^2 for each state (either direction), 
     roundness of winding mode potential (two data points), and
     roundness of heavy quark potential,
     parity doublets  (fractional error).           */

#if FCSQUARED <= 4  /* check that array is big enough */
static const element c2err[NPERP][4]={{0.25,2.0,2.0,2.0}, {0.25,2.0,2.0,2.0},
		    {0.25,2.0,2.0,2.0}, {0.25,2.0,2.0,2.0}},
#endif

#if 1 && MAX_LONG==3 /*Normal values for longitudinal and winding errors */
                    /* these errors match what was found in 2+1 ..  */
  lerr0=0.1, lerr[MAX_LONG]={0.1,0.1,0.2}, werr[1]={1.0};
#elif 0 && MAX_LONG==3/*Normal values for longitudinal and winding errors */
  lerr0=0.2, lerr[MAX_LONG]={0.1,0.1,0.2}, werr[1]={1.0};
#elif 0 && MAX_LONG==3 /*larger errors for longitudnal and winding criteria. */
  lerr0=1.0, lerr[MAX_LONG]={1.0,1.0,1.0}, werr[1]={1.0};
#endif

/* These are the parity doublets that are to be compared.
   parityerror is the fractional error assumed for each  
   doublet.
   each element of whichparity contains {charge,multiplet,which state}
   for each parity doublet member. */
#if FPARITY<=5  /* This is just the size of the arrays. */
  static element parityerr[5]={0.25,0.25,0.25,0.25,0.25};
  static struct {int charge; int multi; int state;}
  whichparity[5][2]={{{1,5,0},{1,6,0}},{{-1,5,0},{-1,6,0}},
		      {{-1,7,0},{-1,4,0}},
		      {{1,7,0},{1,5,0}},{{1,7,0},{1,6,0}}};
#endif

/* Error for extra fit criteria. */

#if FEXTRA<2
doublereal extraerr[1]={0.1};
#endif

/* Winding numbers for longitudinal string tension. lht0 must be
   zero since the lattice spacing is unknown.  Also the 
   winding numbers and lattice symmetry multiplet for determining
   roundness of the heavy sourc potential.  */
#if MAX_LONG==3
static int lht[MAX_LONG][HINDEX]={{1,0},{1,0},{1,1}},
  lmulti[MAX_LONG]={15,15,17},wmulti[1]={17};
#endif
static integer wth,lth;
/* These are all the quantities returned by fcn */
typedef struct {doublereal pextrap; element lvalue[FLONG]; 
  element chi2;  /* make negative to force update. */
  element wvalue[FWIND]; 
  element calclvalue[FLONG]; element calcwvalue[FWIND];
  element wfitting[MAXTH]; element lfitting[MAXTH]; element rescale;
  element spectrumvals1[NP1+1][NPERP][NPERPVALS]; 
  element spectrumvals2[NP2+1][NPERP][NPERPVALS];
  element stillvals[NSTILL][NSTILLVALS]; 
  element c2[NP1+NP2][NPERP][NPERPVALS];
} glue_best;
   
glue_best best;

#define CHI_DIM ((NP1+NP2)*NPERP*FCSQUARED+FLONG+FWIND+FPARITY+FTEPER+FSCALE+FEXTRA) 
                    /*Number of dimensions of chi^2 fit*/
#define NOUT 100 /* Output file name length */

#include "jessica3_print.c"

/*********************************************************************/

int MAIN(int argc, char **argv){
  char out[NOUT],*pout;
  /* maxfev is the maximum number of iterations in minimizing chi^2. */
  integer mm=CHI_DIM,npar=1,maxfev,mode=1,ifail=0,info,nfev,ipvt[NPARAMS];
  /*  epsfcn should be a little less than the Lanczos accuracy. (say 10 times)
      ftol should be a little more than epsfcn
      xtol is the desired precision of the coupling constants  
      factor controls the maximum initial step size.  If the
        starting parameters are near the minimum, this must
        be decreased to maybe 1. 
        (recommended range:  0.1 to 100, default 100). */
  /*  epsfcn is the variable used to determine the step size 
      for the finite difference approximation.  Since the 
      extrapolations magnify any error in the eigenvalues,
      this should be comfortably larger than the errors in
      the lanczos routine.  However, it should also be smaller
      than xtol.                                             */
  doublereal xtol=0.01,gtol=0.0,ftol=5.0e-3,epsfcn=5.0e-5,factor=1.0;
  doublereal vpar[NPARAMS],fvec[CHI_DIM],diag[NPARAMS],fjac[CHI_DIM*NPARAMS];
  doublereal qtf[NPARAMS],wa1[NPARAMS],wa2[NPARAMS],wa3[NPARAMS],wa4[CHI_DIM];
  int count_bases=0;
  element angle[HINDEX+2];
  element chi2;
#if NSTILL==6
  int multistill[NSTILL]={3, 5, 5, 6, 6, 7},
    /*    ostill[NSTILL]={1, 1,-1, 1,-1,-1};*/
    charge_still[NSTILL]={1, 1,-1, 1,-1, 1};
#elif NSTILL==10
  int multistill[NSTILL]={3, 3, 4, 4, 5, 5, 6, 6, 7, 7},
    /* ostill[NSTILL]=   {1,-1, 1,-1, 1,-1, 1,-1, 1,-1};*/
    charge_still[NSTILL]={1,-1, 1,-1, 1,-1, 1,-1,-1, 1};
#elif NSTILL==12
  int multistill[NSTILL]={3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8},
    /*    ostill[NSTILL]={1,-1, 1,-1, 1,-1, 1,-1, 1,-1, 1,-1};*/
    charge_still[NSTILL]={1,-1, 1,-1, 1,-1, 1,-1,-1, 1,-1, 1};
#endif
  int ht_zero[HINDEX]={0,0}; /* zero winding for ordinary states */
  int i,j,k,l;
  FILE *fp=NULL;
#ifdef USE_MPI
  int myrank,nslaves;
#endif

#ifdef BABBAGE  /* Routine to initialize Fortran on Babbage */
  if(hf_fint((char*)NULL)) exit(1);
#endif

  /******************* make basis structures  ****************************/
  /* Do this before starting mpi, so everyone knows what global_kt is */

  for(k=0; k<HINDEX+2; k++)angle[k]=0;

  /*  Calculate the extrapolated spectrum in the P_perp = (c,0) direction */
  for(l=0; l<NP1; l++){
    for(k=0; k<HINDEX; k++)angle[k]=p1a[l][k];
    for(j=0; j<NPERP; j++){
      s[l][j]=all_bases+count_bases;
      for(i=0; i<NS; i++)
	init_spectrum(all_bases+count_bases++,TYPE_GLUE,NULL,ht_zero,np[i],
		     kt[i],chargest[j],0,multis[j],1,angle,NPERPVALS+1,0);
    }
  }
  /*  Calculate the extrapolated spectrum in the P_perp = (c,c) direction */
  for(l=0; l<NP2; l++){
    for(k=0; k<HINDEX; k++)angle[k]=p2a[l][k];
    for(j=0; j<NPERP; j++){
      t[l][j]=all_bases+count_bases;
      for(i=0; i<NS; i++)
	init_spectrum(all_bases+count_bases++,TYPE_GLUE,NULL,ht_zero,np[i],
		      kt[i],chargest[j],0,multit[j],1,angle,NPERPVALS+1,0);
    }
  }
 
  /*  Calculate the ordinary extrapolated spectra. */
  for(j=0; j<NSTILL; j++){
    stillbases[j]=all_bases+count_bases;
    for(i=0; i<NS; i++)
      init_spectrum(all_bases+count_bases++,TYPE_GLUE,NULL,ht_zero,np[i],kt[i],
		   charge_still[j],0,multistill[j],0,NULL,NSTILLVALS,0);
  }
  for(i=0; i<NWIND; i++){
    ws0[i]=all_bases+count_bases;
    for(j=0; j<WNS; j++)
      /* assume o=1 is lowest state */
      init_spectrum(all_bases+count_bases++,TYPE_GLUE,NULL,wht0[i],wnp0[i][j],
		   wkt0[i][j],0,1,15,0,NULL,1,0);
  }

  for(i=0; i<FWIND; i++){
    ws[i]=all_bases+count_bases;
    for(j=0; j<WNS; j++)
      /* assume o=1 is lowest state */
      init_spectrum(all_bases+count_bases++,TYPE_GLUE,NULL,wht[i],wnp[i][j],
		   wkt[i][j],0,1,wmulti[i],0,NULL,1,0);
  }

  /*  Calculate the longitudinal string tension  */
  for(i=0; i<NLONG; i++){
    ls0[i]=all_bases+count_bases;
    /* orientation reversal is not a good symmetry for L != 0 */
    for(j=0; j<LNS; j++){
      angle[HINDEX]=lkmax0[j];
      init_spectrum(all_bases+count_bases++,TYPE_SOURCE,NULL,ht_zero,lnp0[j],
		    lkt0[j],0,0,3,1,angle,1,0);
    }
  }
  
  /*  Calculate potential for n=(c,0), L nonzero  */
  for(i=0; i<FLONG; i++){
    /* orientation reversal is not a good symmetry for L != 0 */
    ls[i]=all_bases+count_bases;
    for(j=0; j<LNS; j++){
      angle[HINDEX]=lkmax[i][j];
      init_spectrum(all_bases+count_bases++,TYPE_SOURCE,NULL,lht[i],lnp[i][j],
		   lkt[i][j],0,0,lmulti[i],1,angle,1,0);
    }
  }
  
  /* Calculate the convergence in p-truncation */
  ps1=all_bases+count_bases;
  for(j=0; j<PNS; j++)
    /* assume o=1, multi=15 is lowest state */
    init_spectrum(all_bases+count_bases++,TYPE_GLUE,NULL,pht1,pnp1[j],pkt1,
		 0,1,15,0,NULL,1,0);
  
  if(count_bases!=N_ALL_BASES){
    fprintf(stderr,__FILE__":%i  Number of bases don't match\n",__LINE__);
    goto cleanup;
  }

  /***********************************************************************/

  /*
    initialize parallelization via MPI when applicable.
     in this case we use a simple master/slave setup to
     distribute the construction of the Hamiltonian and the
     diagonalization of the various bases.      
  */

#ifdef USE_MPI
  MPI_Init(&argc, &argv);     /* initialize MPI */
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  init_mpi_types(GROUP_SIZE);
#if GROUP_SIZE==0
  mpi_group=MPI_COMM_WORLD;   /* communicator used by spectrum(...) */
#endif
  if(myrank!=WORLD_MASTER){
    slave(GROUP_SIZE);
    MPI_Finalize();   /* cleanup MPI */
    return 0;
  }
#endif

  /* Choose scale method and open output file */

  printf("Output file name:  ");
  fgets(out,NOUT,stdin);
  if((pout=strchr(out,'\n'))!=NULL)*pout='\0';
  printf("%s\n",out);
  if((fp = fopen(out,"w")) == NULL){
    fprintf(stderr,__FILE__":  Can't open file %s, exiting\n",out);
    goto cleanup;
  }
  
  printf("Input maximum number of iterations:  ");
  scanf("%i",&j); maxfev=j; /* the type of maxfev might vary */
  
  printf("Choose method for determining the overall scale:\n"
	 "    0 use longitudinal string tension\n"
	 "    1 fit lowest state in spectrum to lattice value\n"
	 "    2 demand minimum chi^2 for c^2 measurements\n"
	 "    3 find scale by demanding overall minimum chi^2.\n");
  scanf("%i",&scalemethod);  
  print_jessica_header(fp,xtol,chargest, multis, multit, charge_still, 
		       multistill,ht_zero); 
   


  /******************* Main loop of program ********************************/

  for(;;){
    printf("How many to fit (0 to %i, -1=stop)\n",NPARAMS-2);
    scanf("%i",&i); npar=i;  /* the type of npar may vary */
    if(npar<0)break;
#if 1 /* Input starting parameters */
    printf("Input %i starting parameters\n",NPARAMS);
    for(i=0; i<NPARAMS; i++)scanf("%lf",par+i);
#endif
    
    best.chi2=-1.0;  /* the first time, always update best */
    if(npar==0){
      info=0; nfev=1;
      fcn(&mm,&npar,vpar,fvec,&info);
    }
    else{
      printf("Input which %i couplings (0=m^2, ...)\n",(int) npar);
      for(i=0;i<npar;i++)scanf("%i",wpar+i);
      for(i=0;i<npar;i++)vpar[i]=par[wpar[i]];

      /*    Least squares fitting routine    
	    subroutine lmdif(fcn,m,n,x,fvec,ftol,xtol,gtol,
	    maxfev,epsfcn,diag,mode,factor,nprint,info,nfev,
	    fjac,ldfjac,ipvt,qtf,wa1,wa2,wa3,wa4)  */
      LMDIF(FCN_TO_FORTRAN,&mm,&npar,vpar,fvec,&ftol,&xtol,&gtol,
            &maxfev,&epsfcn,diag,&mode,&factor,&ifail,&info,&nfev,
            fjac,&mm,ipvt,qtf,wa1,wa2,wa3,wa4);

      if(info<1||info>4) 
	fprintf(stderr,__FILE__":  Error in lmdif info=%i\n",(int) info);
      if(info==0)goto cleanup;
      /*Map best values into par[] */
      for(i=0; i<npar; i++)par[wpar[i]]=vpar[i];
    }
    if(info<0){
      fprintf(stderr,__FILE__":  Error in fcn info=%li\n",(long)info);
      if(info<=-8)goto cleanup;
    }
    chi2=fdot(fvec,fvec,mm);
    print_jessica(fp, &best, chi2, info, nfev,npar);
    print_jessica(stdout, &best, chi2, info, nfev,npar);
  }
  
cleanup:
  fclose(fp);

  /* Free up memory used.  This is mainly for debugging */
  free_spectrum();

#ifdef USE_MPI
  /************* Tell slaves to shut down. **************/
  MPI_Comm_size(MPI_COMM_WORLD,&nslaves);
  for(i=0; i<nslaves; i++){
    if(i==WORLD_MASTER)continue;
    /* only the message tag=0 is read by the slaves. */
    MPI_Send(&i,1,MPI_INT,i,TAG_OFFSET,MPI_COMM_WORLD);
  }
  MPI_Finalize();             /* cleanup MPI */
#else
    printf("Group maximum matrix size: n=%lu, nonzero=" 
	   SIZE_MEM_PRINT ", " SIZE_MEM_PRINT " bytes\n",
	   (lu) max_els.n,max_els.nonzero,max_els.memory);
    printf("Maxumum matrix-vector multiplies: %i\n",max_els.multiply);
    printf("Memory used to construct bases: %lu bytes\n",
	   (lu) reduced_basis_memory);
#endif

  return 0;
}


/*****************************************************************

  The following routine calculates the fit criteria for
  a given set of fit parameters.  It returns the vector
  fvecc which is then squared to give the chi^2.

  The values that are calculated are reused from previous
  iterations if the relevant couplings are unchanged.
  This includes:

         pextrap, wfitting, lfitting, wvalue1, lvalue1,
	 lvalue2, spectrumvals1, spectrumvals2, stillvals
         
  Information about the spectra are returned in the structure
  best.  The members of the structure best 
  are from the lowest chi^2 iteration of fcn (which is
  not necessarily the last iteration of fcn.)

  On error, fcn sets *iflag to a negative integer.
  *flag<=-8 are "fatal" errors.

*******************************************************************/


extern fsub fcn(integer *m, integer *npar, doublereal *xp, 
	       doublereal *fvecc, integer *iflag){
  /* fitc is the index for the fit criterion */
  int i,k,j,l,fitc;
  int count_it;
  integer nwind=NWIND,nlong=NLONG,th=0;
  element *vals=NULL; /* this is realloc'ed by the extrapolation routines */
  element sum,temp,rescale,c2[NP1+NP2][NPERP][NPERPVALS],zz;
  doublereal mlong0[NLONG],mlong[MAX_LONG];
  static doublereal spectrumvals1[NP1+1][NPERP][NPERPVALS],
    spectrumvals2[NP2+1][NPERP][NPERPVALS],
    stillvals[NSTILL][NSTILLVALS];
  static element pextrap,calclvalue[FLONG],calcwvalue[FWIND];
  static element wfitting[MAXTH],lfitting[MAXTH],lvalue[FLONG],
    wvalue[FWIND],wvalue20;  
  clock_t tif;
  time_t clock1,clock2;
#if NWIND>NLONG
  element weights[NWIND*3];
#else
  element weights[NLONG*3];
#endif

  element *all_vals[N_ALL_BASES];
#if defined(USE_MPI) &&  GROUP_SIZE>0
#else
  int job;
  full_basis basis={0};
  eigensystem eigen=INITIAL_EIGENSYSTEM;
#endif
#if defined(USE_MPI) && GROUP_SIZE==0
  int nslaves;
  MPI_Comm_size(MPI_COMM_WORLD,&nslaves);
#endif
  tif=clock();
  time(&clock1);

  for(i=0; i<*npar; i++)par[wpar[i]]=xp[i];

#if PRINTIT
  printf("  Parameters: ");
  for(i=0; i<NPARAMS; i++)printf(" %.10g",par[i]);
  printf("\n");
#endif

  /* Find longitudinal values for a given mass */

  for(i=0;i<nlong;i++){
    mlong0[i]=llong0[i]+sqrt(par[0]/par[1])*llong0m[i];
      for(j=0; j<LNS; j++)
	ls0[i][j].angle[HINDEX+1]=mlong0[i];
  }

  for(i=0;i<FLONG;i++){
    mlong[i]=llong[i]+sqrt(par[0]/par[1])*llongm[i];
    for(j=0; j<LNS; j++)
	ls[i][j].angle[HINDEX+1]=mlong[i];
  }

  /* allocate memory to store eigenvalues */

  for(i=0; i<N_ALL_BASES; i++)
    all_vals[i]=malloc(all_bases[i].nval*sizeof(element));


#if defined(USE_MPI) && GROUP_SIZE>0

  /************ use master-slave approach **********/

  if(master(all_bases,par,all_vals,N_ALL_BASES)){
     fprintf(stderr,__FILE__":  master() is returning an error, exiting\n");
    *iflag=-9;
    return SUBRETURN;
  }
#else  /**************** calculate spectra in parallel ****************/

  for(job=0; job<N_ALL_BASES; job++){
#if defined(USE_MPI) && GROUP_SIZE==0
    for(i=0; i<nslaves; i++){
      if(i==WORLD_MASTER)continue;
#if 0  /* debug print */
      printf("job=%i slave=%i\n",job,i);
#endif
      MPI_Send(par,NPARAMS,MPI_DOUBLE,i,job,MPI_COMM_WORLD);
      MPI_Send(all_bases+job,1,mpi_specify_calculation,i,job,
	       MPI_COMM_WORLD);
    }      
#endif
    spectrum(all_bases+job,par,&eigen,&basis);
    for(i=0; i<all_bases[job].nval; i++)
      all_vals[job][i]=eigen.values[i];
  }
#endif
  /******** take results and calculate extrapolations etc. **************/

  /* Calculate the convergence in p-truncation */
  pextrap=pextrapolate(PNS,ps1,all_vals+(ps1-all_bases));

  /*  Calculate the transverse string tension.  The wht matrix
      must be reformatted into a form compatable with
      the FORTRAN routince WWEIGHTS.    */
  wweights(weights,&wth,nwind,wht0);
  PSEUDOINVERSE(weights,&wth,&nwind);
  for(j=0; j<wth; j++)wfitting[j]=0.0;
  for(i=0; i<nwind; i++){
    extrapolate(&vals,WNS,ws0[i],all_vals+(ws0[i]-all_bases),&th,1,&pextrap);
    /* winding n=(2,0) eigenvalue saved */ 
    if(dot(wht0[i],wht0[i],HINDEX)==4)wvalue20=vals[0];
    for(j=0; j<wth; j++)
      wfitting[j] += vals[0]*weights[i+j*nwind];
  }
  
  /*  Calculate roundness of winding modes */
  for(i=0; i<FWIND; i++){
    extrapolate(&vals,WNS,ws[i],all_vals+(ws[i]-all_bases),&th,1,&pextrap);
    wvalue[i]=vals[0];
  }
  
  /*  Calculate the longitudinal string tension  */
  lweights(weights,&lth,nlong,mlong0);
  PSEUDOINVERSE(weights,&lth,&nlong);
  for(j=0; j<lth; j++)lfitting[j]=0.0;
  for(i=0; i<nlong; i++){
    lextrapolate(&vals,LNS,ls0[i],all_vals+(ls0[i]-all_bases),
		  &th,1,&pextrap,lkmax0);
    for(j=0; j<lth; j++)
      lfitting[j]+=vals[0]*weights[i+j*nlong];
    }
#if 0 /* debugging print */
  printf("  sigma=%f const=%f 1/L %f\n",
	   lfitting[0],lfitting[1],lfitting[2]);
#endif
  
  /*  Calculate roundness of heavy source potential.  */
  for(i=0; i<FLONG; i++){
    lextrapolate(&vals,LNS,ls[i],all_vals+(ls[i]-all_bases),
		  &th,1,&pextrap,lkmax[i]);
    lvalue[i]=vals[0];
  }

  /*  Calculate the ordinary extrapolated spectra. */
  for(i=0; i<NSTILL; i++){
    extrapolate(&vals,NS,stillbases[i],all_vals+(stillbases[i]-all_bases),
		 &th,NSTILLVALS,&pextrap);
    for(k=0; k<NSTILLVALS; k++)stillvals[i][k]=vals[k];
  }
  
  /*  Calculate the extrapolated spectrum in the P_perp = (c,0) direction */
  /*  Sort states before adding (and calculate 1 extra value) */
  for(l=0; l<NP1; l++)
    for(i=0; i<NPERP; i++){
      extrapolate(&vals,NS,s[l][i],all_vals+(s[l][i]-all_bases),
		   &th,NPERPVALS+1,&pextrap);
      /* sort and add states: we assume NPERPVALS+1 eigenvalues */
      k=0;
      addterms(spectrumvals1[l+1][i],&k,NPERPVALS,vals,NPERPVALS+1);
#if 0 /* debugging print */
      printf("  l=%i, spectrum %i P_perp=(%f,%f)\n",l,i,p1a[l][0],p1a[l][1]);
#endif
    } 
  
  /*  Calculate the extrapolated spectrum in the P_perp = (c,c) direction */
  /*  Sort states before adding (and calculate 1 extra value) */
  for(l=0; l<NP2; l++)
    for(i=0; i<NPERP; i++){
      extrapolate(&vals,NS,t[l][i],all_vals+(t[l][i]-all_bases),
		   &th,NPERPVALS+1,&pextrap);
      /* sort and add states: we assume NPERPVALS+1 eigenvalues */
      k=0;
      addterms(spectrumvals2[l+1][i],&k,NPERPVALS,vals,NPERPVALS+1);
#if 0 /* debugging print */
      printf("  l=%i spectrum %i P_perp=(%f,%f)\n",l,i,p2a[l][0],p2a[l][1]);
#endif
    }

  /***************** Free up memory used. ************************/
#if !(defined(USE_MPI) && GROUP_SIZE>0)
  free_eigensystem(&eigen);
  free_basis(&basis);  /* not necessary if we don't use eigenvectors */
#endif
  for(i=0; i<N_ALL_BASES; i++)free(all_vals[i]);


  /*  Put ordinary spectrum values into nonzero P_perp array*/
  /*  Here we sort the states before adding.  */
  /*  The ordering of the states must be compatable with */
  /*  the ordering obtained when calculating nonzero p_perp */
  for(i=0; i<NPERP; i++){
    for(j=0, k=l=0; j<NSTILL; j++){
      if(stillbases[j]->charge!=chargest[i])continue;
      /* assume that the two spin 1 multiplets are degenerate */
      if(((stillbases[j]->multi==3 || stillbases[j]->multi==5 ||
	    stillbases[j]->multi==7) && multis[i]==15) ||
	  ((stillbases[j]->multi==4 || stillbases[j]->multi==6 ||
	    stillbases[j]->multi==7) && multis[i]==16))
	addterms(spectrumvals1[0][i],&k,NPERPVALS,stillvals[j],NSTILLVALS);
     if(((stillbases[j]->multi==3 || stillbases[j]->multi==6 ||
	    stillbases[j]->multi==7) && multit[i]==17) ||
	  ((stillbases[j]->multi==4 || stillbases[j]->multi==5 ||
	    stillbases[j]->multi==7) && multit[i]==18))
	addterms(spectrumvals2[0][i],&l,NPERPVALS,stillvals[j],NSTILLVALS);
    }
    if(k!=NPERPVALS || l!=NPERPVALS){
       fprintf(stderr,__FILE__":  Incorrect multiplet mapping from states "
	       "in ordinary\n"
	      "spectrum to those in nonzero P_perp, exiting\n");
       fprintf(stderr,"          k=%i, l=%i and NPERPVALS=%i\n",
	       k,l,NPERPVALS);
      *iflag=-12;
      return SUBRETURN;
    }	
  }
    
#if 0 /*debugging print of the entire spectrum */
  for(k=0; k<NPERPVALS; k++)
    for(i=0; i<NPERP; i++){
      for(j=0; j<NP1+1; j++)
	printf(" %g",spectrumvals1[j][i][k]);
      for(j=0; j<NP1+1; j++)
	printf(" %g",spectrumvals2[j][i][k]);
      printf("\n");
    } 
#endif
      
  /* There are several possible methods for setting
     the overall scale of the Hamiltonian  */
  
  switch(scalemethod){         
  case 0:          /*Find scale by long string tension */
    rescale=fabs(1.0/lfitting[0]);
    break;
  case 1:    /*Find scale by looking only at lowest state */
    rescale=teper[0]/stillvals[0][0];
    break;
  case 2:    /* Find scale by minimizing chi^2 for c^2 measurement */
    if(NP1<1){ 
      *iflag=-1;
      rescale=-1;
      break;
    }
    sum=0.0; zz=0.0;
    for(i=0; i<NPERP; i++)
      for(j=0; j<FCSQUARED && j<NPERPVALS; j++){
	rescale=spectrumvals1[1][i][j]-spectrumvals1[0][i][j];
	sum += rescale/pow(c2err[i][j],2);
	zz += pow(rescale/c2err[i][j],2);
      }
    rescale=sum/zz*fdot(p1a[0],p1a[0],HINDEX)/wfitting[0];
    if(rescale>0.0)
      rescale=sqrt(rescale);
    else {
      rescale=0.0;
      *iflag = -1;
    }
    break;
  case 3:    /* Find scale by minimizing overall chi^2 */
    /* This does not include the consistancy of the L nonzero,
       n=1 datum because it is too complicated */
    /* initial guess is from Teper's lowest state */
    printf("Find scale by minimizine chi^2 rescale=");
    for(rescale=teper[0]/stillvals[0][0], count_it=0; ; count_it++){
      sum=0.0; zz=0.0;
      for(j=0; j<NPERP; j++)
	for(i=0; i<FCSQUARED && i<NPERPVALS; i++){
	  for(k=0; k<NP1; k++){
	    c2[k][j][i]=wfitting[0]*pow(rescale,2)*
	      (spectrumvals1[k+1][j][i]-spectrumvals1[0][j][i])/
	      fdot(p1a[k],p1a[k],HINDEX);
	    temp=2.0*c2[k][j][i]/rescale;
	    sum += temp*(c2[k][j][i]-1.0)/pow(c2err[j][i],2);
	    zz +=  pow(temp/c2err[j][i],2);
	  }
	  for(k=0; k<NP2; k++){
	    c2[k+NP1][j][i]=wfitting[0]*pow(rescale,2)*
	      (spectrumvals2[k+1][j][i]-spectrumvals2[0][j][i])/
	      fdot(p2a[k],p2a[k],HINDEX);
	    temp=2.0*c2[k+NP1][j][i]/rescale;
	    sum += temp*(c2[k+NP1][j][i]-1.0)/pow(c2err[j][i],2);
	    zz +=  pow(temp/c2err[j][i],2);
	  }
	}
#if FTEPER==1
      sum += stillvals[0][0]*(stillvals[0][0]*rescale-teper[0])/
	pow(tepererr[0],2);
      zz += pow(stillvals[0][0]/tepererr[0],2);
#endif
#if FSCALE
      sum += lfitting[0]*(rescale*lfitting[0]-1.0)/pow(lerr0,2);
      zz += pow(lfitting[0]/lerr0,2);
#endif
      temp=rescale;
      /* Here I include a damping factor which will
	 slow convergence but hopefully stablize it. */
      rescale -= 0.25*sum/zz;
#if 0 /* debugging print */
      if(count_it<100)
	printf("rescale:  count_it=%i rescale=%.16g\n",count_it,rescale);
#endif
      /* converge as well as we can, assuming round-off errors */
      if(fabs(temp-rescale)<=(2.0+2.0*FCSQUARED)*ELEMENT_EPSILON)break;
      if(count_it>1000){
	fprintf(stderr,__FILE__":  Scale determination not converged, "
		"rescale=%g\n",rescale);
	rescale=fabs(1.0/lfitting[0]); /* give up and use long. tension */
	break;
      }
    }
    printf("%g, %i iterations\n",rescale,count_it);
    break;
  default:
    fprintf(stderr,__FILE__":  in fcn, scale method=%i is wrong.\n",
	    scalemethod);
    *iflag=-14;
    rescale=teper[0]/stillvals[0][0];
  }
  /* If rescale is negative for some reason, 
     set it to something large and positive.        */
  if(rescale<0.0){
    rescale=50.0;
    fprintf(stderr,__FILE__":  rescale is negative, "
	    "set recale=%f.\n",rescale);
  }
  
  /**********************************************************************/

  fitc=0;  /* start counting fit criteria */

  /*  
      Calculate the desired dispersion using the overall scale 
      determined above.
      Find error of speed of light squared
      with coupling g^2 N and "a" as determined above.
      One can weight the value for our value of sigma.
  */

  for(i=0; i<NPERP; i++)
    for(j=0; j<NPERPVALS; j++){
      for(k=0; k<NP1; k++){
	temp=wfitting[0]*pow(rescale,2)/fdot(p1a[k],p1a[k],HINDEX);
	c2[k][i][j] = (spectrumvals1[k+1][i][j]-spectrumvals1[0][i][j])*temp;
      }
      for(k=0; k<NP2; k++){
	temp=wfitting[0]*pow(rescale,2.0)/fdot(p2a[k],p2a[k],HINDEX);
	c2[k+NP1][i][j] = 
	  (spectrumvals2[k+1][i][j]-spectrumvals2[0][i][j])*temp;
      }
    }
  for(i=0; i<NPERP; i++)
    for(j=0; j<FCSQUARED && j<NPERPVALS; j++)
      for(k=0; k<NP1+NP2; k++)
	fvecc[fitc++] = (c2[k][i][j]-1.0)/c2err[i][j];
  
  /* If the 0^++ or the n=(2,0) is tachyonic, modify the 
     0^++ c^2 errors to reflect this */

  if(stillvals[0][0]<0.0){
      fprintf(stderr,__FILE__":  Lowest state gone tachyonc.\n");
      fvecc[0]=100.0*fabs(stillvals[0][0]);
  }
#if 0  /*  Add penalty if the winding modes become tachyonic.  */
  if(wvalue20<0.0){
      fprintf(stderr,__FILE__":  Winding mode gone tachyonc.\n");
      fvecc[1]=100.0*fabs(wvalue20);
  }
#endif

  /* Add roundness of winding modes to fitting criterion. */
  for(i=0; i<FWIND; i++){
    wweights(weights,&wth,1,&wht[i]);
    calcwvalue[i]=fdot(weights,wfitting,wth);
    fvecc[fitc++]=(wvalue[i]-calcwvalue[i])/werr[i];
  }

  /* Add roundness of heavy source to fitting criterion.  
     Here, sum is the total
     distance in units of G^2 N. */
  for(i=0; i<FLONG; i++){
  calclvalue[i]=pow(mlong[i],2)+wfitting[0]*
    (element) dot(lht[i],lht[i],HINDEX)*pow(rescale,2);
  if(calclvalue[i]>0.0){
    calclvalue[i]=sqrt(calclvalue[i]);
    lweights(weights,&lth,1,&calclvalue[i]);
    calclvalue[i]=fdot(weights,lfitting,lth);
    fvecc[fitc++]=(lvalue[i]-calclvalue[i])/lerr[i];
  } else {
     fprintf(stderr,__FILE__":  value for calclvalue[%i]=%g bad, mlong=%g wfitting[0]=%g\n"
	    "  dot=%g rescale=%g\n",i,calclvalue[i],mlong[i],
	    wfitting[0],(element) dot(lht[i],lht[i],HINDEX),rescale);
    fvecc[fitc++]=100.0*fabs(calclvalue[i])+10.0;
  }
#if 1==0 /* debugging print  */
  printf("fitc=%i fvecc=%f calclvalue[%i]=%f lerr=%f lvalue=%f\n",
	 fitc,fvecc[fitc],calclvalue[i],lerr[i],lvalue[i]);
#endif
  }

  /*  Consistancy of longitudinal string tension with rescale */
#if FSCALE
  fvecc[fitc++]=(rescale*lfitting[0]-1.0)/lerr0;
#endif

  /* Add parity doublets to fitting criterion using fractional
     difference.   */

  for(k=0; k<FPARITY; k++){
    for(i=0; i<NSTILL; i++)
      if(stillbases[i]->multi==whichparity[k][0].multi && 
	 stillbases[i]->charge==whichparity[k][0].charge)break;
    for(j=0; j<NSTILL; j++)
      if(stillbases[j]->multi==whichparity[k][1].multi && 
	 stillbases[j]->charge==whichparity[k][1].charge)break;
    if(i<NSTILL && j<NSTILL)
      fvecc[fitc++]=(stillvals[i][whichparity[k][0].state]-
		     stillvals[j][whichparity[k][1].state])/
	(0.5*(stillvals[i][whichparity[k][0].state]+
	      stillvals[j][whichparity[k][1].state])*parityerr[k]);
    else {
      fprintf(stderr,__FILE__":  Incorrect multiplets in ordinary spectra, "
	      " exiting\n");
      *iflag=-15;
      return SUBRETURN;
    }
  }

  /* Add fit to teper's value for the lowest state
     we assume that the lowest state is in the first sector
     of stillvals.   */
#if FTEPER==1
    fvecc[fitc++]=(rescale*stillvals[0][0]-teper[0])/tepererr[0];
#endif

    /* Demand that the lattice spacing be a specific value. */
#if FEXTRA==1
    fvecc[fitc++]=(1.0+2.0*sqrt(par[0]/par[1])-wfitting[0]*rescale)/
      extraerr[0];
#endif
    
  /* after including all fit criteria, make sure we 
     have actually filled the fvecc vector. */
  if(fitc!=*m){
    fprintf(stderr,__FILE__":  Number of fit criteria wrong.\n");
    *iflag=-75;
    return SUBRETURN;
  }
    
    /* check if this is the best chi^2 so far,
       if so, copy values into the structure best.
    */
  sum=fdot(fvecc,fvecc,*m);
  if(*iflag>=0 && (best.chi2<0.0 || sum<best.chi2)){
    best.chi2=sum;
    best.pextrap=pextrap; 
    for(i=0; i<FWIND; i++){
      best.wvalue[i]=wvalue[i];
      best.calcwvalue[i]=calcwvalue[i];
    }
    for(i=0; i<FLONG; i++){
      best.lvalue[i]=lvalue[i];
      best.calclvalue[i]=calclvalue[i];
    }
    best.rescale=rescale;
    for(i=0; i<NPERP; i++)
      for(j=0; j<NPERPVALS; j++)
	for(k=0; k<NP1+NP2; k++)
	  best.c2[k][i][j]=c2[k][i][j];
    for(i=0; i<NPERP; i++)
      for(k=0; k<NPERPVALS; k++){
	for(j=0; j<NP1+1; j++)
	  best.spectrumvals1[j][i][k]=spectrumvals1[j][i][k];
	for(j=0; j<NP2+1; j++)
	  best.spectrumvals2[j][i][k]=spectrumvals2[j][i][k];
      }
    for(i=0; i<NSTILL; i++)
      for(j=0; j<NSTILLVALS; j++)
	best.stillvals[i][j]=stillvals[i][j];
    for(i=0; i<wth; i++)best.wfitting[i]=wfitting[i];
    for(i=0; i<lth; i++)best.lfitting[i]=lfitting[i];
  }

#if PRINTIT
  printf("  fvecc values =");
  for(i=0, k=3; i<*m; i++){
    printf(" %g",fvecc[i]);
    if((k++)%6==0 && i+1<*m)printf("\n  ");
  }
  time(&clock2);
  printf("\n  Chi^2=%g in %.2f CPU seconds (%li wall).\n\n",
	 fdot(fvecc,fvecc,*m),
	 (float)(clock()-tif)/(float) CLOCKS_PER_SEC,clock2-clock1);
  fflush(stdout);
#endif

  free(vals); /* vals is realloc()ed by extrapolation routines */

  return SUBRETURN; /* normal return from subroutine */

}

#endif /* HINDEX==2 */



