#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include "spectrum.h"
#include "extrapolate.h"
#if defined(USE_MPI)
#include "usempi.h"
#endif

#if HINDEX==2
#define PRINTIT 1 /*  Print timing and convergence information  */

/*
                 Mesons on the transverse lattice

    This program searches coupling constant space for a best fit
    to Poincare symmetries that are broken by the transverse lattice
    regulator.  

    We test for rotational invariance in the transverse plane.  
    In practice, we will choose the x^1 direction and the x^1=x^2 
    directions in order to take advandage of residual reflection 
    symmetries.

    We measure:                                                         multi

    p-truncation convergence for n=(1,0) winding mode                   15
    longitudinal potential for n=(0,0) and various L                     3
    longitudinal eigenvalue for n=(1,0) and one L                       15
    longitudinal eigenvalue for n=(1,1) and one L                       17
    winding mode potential for various n=(c,0)                          15
    winding mode potential for one n=(c,c)                              17
    spectrum for P_perp = (0,0) and (c,0)                               9, 8
                                                                        7, 10
    spectrum for P_perp = (c,c)                                         11, 12
                                                                        14, 13


************************************+++++++**********************************/

#define USE_LMDIF 1 /* Use lmdif or PORT3 routine for minimization */
#define FCSQUARED 2 /* Number of c^2 values in each sector and
                    direction to use in fit. This can be up to NPERPVALS */
#define FPARITY 0  /*1,3 Number of parity doublets to fit. */
#define NPERP 4 /* Number of J^{P_1 C} sectors for nonzero perp momentum 
		   this must be 4 to cover all multiplets */
#define NPERPVALS 2 /* number of states calculated in each sector */
#define NP1 0 /* Number of nonzero values of P_perp in the (c,0) direction*/
#define NP2 1 /* Number of nonzero values of P_perp in the (c,c) direction*/
              /* This can be 1.  */
#define NSTILL 10 /*Number of sectors for zero transverse momentum spectra
                    This must be 10 to cover all multiplets. */ 
#define NSTILLVALS 2  /*Number of states calculated in each sector*/
                      /*This must be large enough to use for nonzero 
                        momenta */
#define MAXTH 10 /* max number of terms in fit functions */
#define FIX_MASSES 3 /*2,5 number of masses to fit to experimental values */
#define FEXTRA 1 /* Fit criteria that would not be normally used. */  
                   /*For instance, demanding 
                     a specific value for the lattice spacing.  */
#define N_LOG_VALS 1 /* number of log states to calculate, must be 0,1 */
/* if these are zero, take values from lattice glueball results */

#define NLONG 0  /*Data points used to estimate longitudinal potential */
#define NWIND 0 /* Data points used to estimate winding potential */
#define PNS 0 /* number of sectors in p extrapolation */

/************************************+++++++**********************************/

#if 0 /* basis simon is using test */
/* This takes about 85 seconds to calculate everything */ 
#define NS 4
const static int kt[NS]={6,8,10,12}, np[NS]={4,4,4,4};
#if PNS<=3 /* check dimensions */
static int pnp1[3]={3,5,7}, pkt1=11;
#endif
#define LOG_NS 4 /* number of sectors to find log divergence */
static int log_np[LOG_NS]={4,4,4,4}, log_kt[LOG_NS]={12,16,20,24};

#elif 0 /* one link truncation */
/* This takes about 4 CPU seconds to calculate everything */ 
#undef FPARITY 
#define FPARITY 1 
#undef FIX_MASSES
#define FIX_MASSES 2
#undef FEXTRA
#define FEXTRA 0
#define NS 3
const static int kt[NS]={12,14,18}, np[NS]={3,3,3};
#if PNS<=3 /* check dimensions */
static int pnp1[3]={3,5,7}, pkt1=11;
#endif
#define LOG_NS 4 /* number of sectors to find log divergence */
static int log_np[LOG_NS]={3,3,3,3}, log_kt[LOG_NS]={20,24,28,36};

#elif 0 /* smaller test basis, two links */
/* This takes about 42 CPU seconds to calculate everything */ 
#define NS 3
const static int kt[NS]={12,14,18}, np[NS]={4,4,4};
#if PNS<=3 /* check dimensions */
static int pnp1[3]={3,5,7}, pkt1=11;
#endif
#define LOG_NS 4 /* number of sectors to find log divergence */
static int log_np[LOG_NS]={4,4,4,4}, log_kt[LOG_NS]={20,24,28,36};

#elif 1 /* three links */
/* One step is 128 CPU minutes, Aug. 2002 */ 
#define BIG /* use rescale, etc from three/ttraj_1_composite.out */
#define NS 3
const static int kt[NS]={20,24,30}, np[NS]={5,5,5};
#if PNS<=3 /* check dimensions */
static int pnp1[3]={3,5,7}, pkt1=11;
#endif
#define LOG_NS 4 /* number of sectors to find log divergence */
static int log_np[LOG_NS]={5,5,5,5}, log_kt[LOG_NS]={20,24,30,40};

#elif 0 /* four links */
/* This takes about 620 CPU seconds to calculate everything */ 
#define NS 3
const static int kt[NS]={12,14,18}, np[NS]={6,6,6};
#if PNS<=3 /* check dimensions */
static int pnp1[3]={3,5,7}, pkt1=11;
#endif
#define LOG_NS 0 /* number of sectors to find log divergence */
#if LOG_NS <= 4
static int log_np[4]={6,6,6,6}, log_kt[4]={20,24,28,36};
#endif

#elif 0 /* a little bigger basis */
/* This takes about 22 minutes to calculate everything */ 
#define NS 3
const static int kt[NS]={24,28,36}, np[NS]={4,4,4};
#if PNS<=3 /* check dimensions */
static int pnp1[3]={3,5,7}, pkt1=11;
#endif
#define LOG_NS 4 /* number of sectors to find log divergence */
static int log_np[LOG_NS]={4,4,4,4}, log_kt[LOG_NS]={36,40,50,70};

#endif  /************* Choose different possible bases *********************/

/* Total number of different bases */
#define CHI_DIM ((NP1+NP2)*NPERP*FCSQUARED+FPARITY+FIX_MASSES+FEXTRA) 
                    /*Number of dimensions of chi^2 fit*/
#define N_ALL_BASES ((NSTILL+(NP1+NP2)*NPERP)*NS+PNS+LOG_NS)
static specify_calculation all_bases[N_ALL_BASES];
static specify_calculation *s[NP1?NP1:1][NPERP];
static specify_calculation *t[NP2?NP2:1][NPERP];
static specify_calculation *ps1,*stillbases[NSTILL],*log_base;

element par[NPARAMS];
int new_best=0;  /* flag to force fcn to reset best chi^2 results */

static int wpar[NPARAMS],scalemethod;
#if NP1<2 && NP2<2  /* check dimensions */
static doublereal p1a[1][HINDEX]={{0.25,0.0}}, p2a[1][HINDEX]={{0.25,0.25}};
#endif
int multis[NPERP]={15,15,16,16},  /* multi for nonzero P_perp */
    chargest[NPERP]={1,-1,-1,1},  /* charge conjugation for nonzero P_perp */
    multit[NPERP]={17,17,18,18};
static int pht1[HINDEX]={1,0};  /* winding number for p-extrapolation */

/* The error for: 
     c^2 for each state (either direction), 
     roundness of winding mode potential (two data points), and
     roundness of heavy quark potential,
     parity doublets  (fractional error).           */

#if FCSQUARED <= 4  /* check that array is big enough */
/* it makes sense to have the same error in each direction */
const element c2err[NPERP][4]={{5.0,5.0,5.0,5.0}, {0.2,0.2,5.0,5.0}, 
					{0.2,5.0,5.0,5.0},{0.2,5.0,5.0,5.0}};
#endif

/* These are the parity doublets that are to be compared. */
#if FPARITY<=5  /* This is just the size of the arrays. */
/* relative error for chi^2 fit */
static element parityerr[5]={0.3,0.5,0.5,0.5,0.5};
static struct {
  int charge; 
  int multi; 
  int state;}
whichparity[5][2]={
  {{-1,7,0},{-1,3,0}}, /* components of rho */
  {{1,5,0},{1,6,0}},{{-1,5,0},{-1,6,0}}, /* lowest J_z=2 doublets */
  {{1,7,0},{1,5,0}},{{1,7,0},{1,6,0}}};
#endif

/* experimental meson masses, here I include the lowest state
   in each J_z^{P_1,C} sector that is not otherwise determined
   by a J multiplet degeneracy.  The idea is that spin
   multiplets should be fixed by FPARITY above. */
#if FIX_MASSES<=8   /* This is just the size of the array. */
struct {
  doublereal value;  /* mass^2 in units of the string tension */
  doublereal error;  /* mass^2 error in units of the string tension 
			note that there is a conversion to get to mass error */
  int charge;
  int multi;
  int state;
} experiment[8]={{0.1006,0.1,1,4,0},  /* pion */
                 {3.191,0.6,-1,3,0},  /* J_z=0 of the rho */
                 {3.191,0.6,-1,7,0},  /* J_z=1 of the rho */
                 {4.991,2.0, 1,3,0},  /* a_0(983) SD says this is
					 a bad choice */
		 {7.878,2.0,-1,4,0},  /* J_z=0 of the b_1(1235) */
		 {8.200,2.0, 1,7,0},  /* J_z=1 of the a_1(1260) */
		 {9.000,2.0, 1,5,0},  /* J_z=2 of the a_2(1320) */
		 {14.75,4.0,-1,5,0}}; /* J_z=2 of the rho_2(1690) */
#endif

static integer wth,lth;

/* function minimization routine below */
extern fsub fcn(integer *, integer *, doublereal *,doublereal *,integer *);
extern fsub fcn3(integer *m, integer *npar, doublereal *xp, integer *nf,
		 doublereal *fvecc, integer *ui, doublereal *ur, fsub *uf);
#include "christiaan_print.c"
meson_best best;
   
#define NOUT 100 /* Output file name length */

/*********************************************************************/

int MAIN(int argc, char **argv){
  char out[NOUT],*pout;
  int count_bases=0;
  element angle[HINDEX+2];
  element chi2;
#if NSTILL==6
  int multistill[NSTILL]={3, 5, 5, 6, 6, 7},
    /*    ostill[NSTILL]={1, 1,-1, 1,-1,-1};*/
    charge_still[NSTILL]={1, 1,-1, 1,-1, 1};
#elif NSTILL==10
  int multistill[NSTILL]={3, 3, 4, 4, 5, 5, 6, 6, 7, 7},
    /* ostill[NSTILL]=   {1,-1, 1,-1, 1,-1, 1,-1, 1,-1};*/
    charge_still[NSTILL]={1,-1, 1,-1, 1,-1, 1,-1,-1, 1};
#elif NSTILL==12
  int multistill[NSTILL]={3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8},
    /*    ostill[NSTILL]={1,-1, 1,-1, 1,-1, 1,-1, 1,-1, 1,-1};*/
    charge_still[NSTILL]={1,-1, 1,-1, 1,-1, 1,-1,-1, 1,-1, 1};
#endif
  int i,j,k,l,tempi;
  FILE *fp=NULL;
#ifdef USE_MPI
  int myrank,nslaves;
#endif
  integer maxfev;  /* maximum number of iterations in minimizing chi^2. */
  integer mm=CHI_DIM,npar=1,info,nfev;
  /*
    epsfcn should be a little less than the Lanczos accuracy. (say 10 times)
    ftol should be a little more than epsfcn
    xtol is the desired precision of the coupling constants  
    factor controls the maximum initial step size.  If the
    starting parameters are near the minimum, this must
    be decreased to maybe 1. 
    (recommended range:  0.1 to 100, default 100). 
    
    epsfcn is the variable used to determine the step size 
    for the finite difference approximation.  Since the 
    extrapolations magnify any error in the eigenvalues,
    this should be comfortably larger than the errors in
    the lanczos routine.  However, it should also be smaller
    than xtol. 

    lmdif seems to converge faster, but the port routine
    seems to get a lower minimum value.
  */
  doublereal xtol=0.001,ftol=5.0e-5,epsfcn=5.0e-7;
  doublereal vpar[NPARAMS],fvec[CHI_DIM];
#if USE_LMDIF
  integer mode=1,ifail=0,ipvt[NPARAMS];
  doublereal diag[NPARAMS],fjac[CHI_DIM*NPARAMS],gtol=0.0,factor=1.0;
  doublereal qtf[NPARAMS],wa1[NPARAMS],wa2[NPARAMS],wa3[NPARAMS],wa4[CHI_DIM];
#else  /* Use PORT3 library routine */
#define LIV 100
#define LV 300
	integer iv[LIV],liv=LIV,lv=LV,one=1;
	doublereal v[LV];
#endif

#ifdef BABBAGE  /* Routine to initialize Fortran on Babbage */
  if(hf_fint((char*)NULL)) exit(1);
#endif

 /******************* make basis structures  ****************************/
  /* Do this before starting mpi, so everyone knows what global_kt is */

  for(k=0; k<HINDEX+2; k++)angle[k]=0;

  /*  Calculate the extrapolated spectrum in the P_perp = (c,0) direction */
  for(l=0; l<NP1; l++){
    for(k=0; k<HINDEX; k++)angle[k]=p1a[l][k];
    for(j=0; j<NPERP; j++){
      s[l][j]=all_bases+count_bases;
      for(i=0; i<NS; i++)
	initspectrum(all_bases+count_bases++,TYPE_MESON,NULL,np[i],kt[i],
		     chargest[j],0,multis[j],1,angle,NPERPVALS+1,0);
    }
  }
  /*  Calculate the extrapolated spectrum in the P_perp = (c,c) direction */
  for(l=0; l<NP2; l++){
    for(k=0; k<HINDEX; k++)angle[k]=p2a[l][k];
    for(j=0; j<NPERP; j++){
      t[l][j]=all_bases+count_bases;
      for(i=0; i<NS; i++)
	initspectrum(all_bases+count_bases++,TYPE_MESON,NULL,np[i],kt[i],
		     chargest[j],0,multit[j],1,angle,NPERPVALS+1,0);
    }
  }
 
  /*  Calculate the ordinary extrapolated spectra. */
  for(j=0; j<NSTILL; j++){
    stillbases[j]=all_bases+count_bases;
    for(i=0; i<NS; i++)
      initspectrum(all_bases+count_bases++,TYPE_MESON,NULL,np[i],kt[i],
		   charge_still[j],0,multistill[j],0,NULL,NSTILLVALS,0);
  }

  /* Calculate the convergence in p-truncation */
  ps1=all_bases+count_bases;
  for(j=0; j<PNS; j++)
    /* assume o=1, multi=15 is lowest state */
    initspectrum(all_bases+count_bases++,TYPE_GLUE,pht1,pnp1[j],pkt1,
		 0,1,15,0,NULL,1,0);  

  /* Calculate the log divergence */
  log_base=all_bases+count_bases;
  for(j=0; j<LOG_NS; j++)
    initspectrum(all_bases+count_bases++,TYPE_MESON,NULL,log_np[j],log_kt[j],
		 1,0,4,0,NULL,N_LOG_VALS,0);  

  if(count_bases!=N_ALL_BASES){
    fprintf(stderr,__FILE__":%i  Number of bases don't match\n",__LINE__);
    goto cleanup;
  }

 /***********************************************************************/

  /* 
     initialize parallelization via MPI when applicable.
     in this case we use a simple master/slave setup to
     distribute the construction of the Hamiltonian and the
     diagonalization of the various bases.      
  */

#ifdef USE_MPI
  MPI_Init(&argc, &argv);     /* initialize MPI */
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  init_mpi_types(GROUP_SIZE);
#if GROUP_SIZE==0
  mpi_group=MPI_COMM_WORLD;   /* communicator used by spectrum(...) */
#endif
  if(myrank!=WORLD_MASTER){
    slave(GROUP_SIZE);
    MPI_Finalize();   /* cleanup MPI */
    return 0;
  }
#endif

  /* Choose scale method and open output file */

  printf("Output file name:  ");
  fgets(out,NOUT,stdin);
  if((pout=strchr(out,'\n'))!=NULL)*pout='\0';
  printf("%s\n",out);
  if((fp = fopen(out,"w")) == NULL){
    fprintf(stderr,__FILE__":  Can't open file %s, exiting\n",out);
    goto cleanup;
  }

  printf("Input maximum number of iterations:  ");
  scanf("%i",&tempi); maxfev=tempi;

  printf("Choose method for determining the overall scale:\n"
         "    0 use scale from glueball calculations\n"
	 "    1 find scale by demanding overall minimum chi^2.\n");
  scanf("%i",&scalemethod);  
  
  /*  Print out the results in the output file **/
  print_christiaan_header(fp,xtol,chargest,multis,multit,
			 charge_still,multistill);


  /******************* Main loop of program ********************************/

  for(;;){
    printf("How many to fit (0 to %i, -1=stop)\n",NPARAMS-2);
    scanf("%i",&tempi); npar=tempi;
    if(npar<0)break;
#if 1 /* Input starting parameters */
    printf("Input %i starting parameters\n",NPARAMS);
    for(i=0; i<NPARAMS; i++)scanf("%lf",par+i);
#endif

    new_best=1; /* flag to force fcn to reset best chi^2 results */

    if(npar==0){
      info=0; nfev=1;
      fcn(&mm,&npar,vpar,fvec,&info);
      chi2=fdot(fvec,fvec,mm);
    }
    else{
      printf("Input which %li couplings (0=m^2, ...)\n",npar);
      for(i=0;i<npar;i++)scanf("%i",wpar+i);
      for(i=0;i<npar;i++)vpar[i]=par[wpar[i]];

/*    Least squares fitting routine    
      subroutine lmdif(fcn,m,n,x,fvec,ftol,xtol,gtol,
      *                  maxfev,epsfcn,diag,mode,factor,nprint,info,nfev,
 *                 fjac,ldfjac,ipvt,qtf,wa1,wa2,wa3,wa4)  */
#if USE_LMDIF
      LMDIF(FCN_TO_FORTRAN,&mm,&npar,vpar,fvec,&ftol,&xtol,&gtol,
            &maxfev,&epsfcn,diag,&mode,&factor,&ifail,&info,&nfev,
            fjac,&mm,ipvt,qtf,wa1,wa2,wa3,wa4);
      if(info<1||info>4) 
	fprintf(stderr,__FILE__":  Error in lmdif info=%li\n",info);
      if(info==0)goto cleanup;
      chi2=fdot(fvec,fvec,mm);
#else   /* use PORT3 library routine  */
      DIVSET(&one,iv,&liv,&lv,v);  /* set default values */
      /* don't print diagnostics */
      iv[13]=0; iv[18]=0; iv[19]=0; iv[22]=0; iv[23]=0;  
      v[31]=ftol; v[32]=xtol; 
      v[42]=epsfcn;  /* step size for finding derivative */
      iv[16]=maxfev; iv[17]=maxfev;  /* set maximum iterations */
      DN2F(&mm,&npar,vpar,FCN3_TO_FORTRAN,iv,&liv,&lv,v,&info,NULL,NULL);
      if(info<0 || iv[0]>6)fprintf(stderr,__FILE__":  Error in dn2f, "
				   "iv[0]=%li info=%li\n",iv[0],info);
      if(iv[0]>14 || info<-7)goto cleanup;  /* fatal errors */ 
      if(info==0)info=iv[0];  /* if no flag from fcn, use code from fn2f */
      nfev=iv[5]+iv[29];
      chi2=2.0*v[9];
#endif
      /*Map best values into par[] */
      for(i=0; i<npar; i++)par[wpar[i]]=vpar[i];
    }
    if(info<0){
       fprintf(stderr,__FILE__":  Error in fcn info=%li\n",info);
      if(info<=-8)goto cleanup;
    }
    /* print the results of the minimization */
    print_christiaan(fp,&best,chi2,info,nfev,npar);
    print_christiaan(stdout,&best,chi2,info,nfev,npar);

  }
  
  
cleanup:
  fclose(fp);

  /* Free up memory used.  This is mainly for debugging */
  free_spectrum();

#ifdef USE_MPI
  /************* Tell slaves to shut down. **************/
  MPI_Comm_size(MPI_COMM_WORLD,&nslaves);
  for(i=0; i<nslaves; i++){
    if(i==WORLD_MASTER)continue;
    /* only the message tag=0 is read by the slaves. */
    MPI_Send(&i,1,MPI_INT,i,TAG_OFFSET,MPI_COMM_WORLD);
  }
  MPI_Finalize();             /* cleanup MPI */
#endif
#if defined(USE_MPI) && GROUP_SIZE != 0
#else /* only works for one group */
  printf("Group maximum matrix size: n=" SIZE_T_PRINT ", nonzero=" 
	 SIZE_MEM_PRINT ", " SIZE_MEM_PRINT " bytes\n",
	 max_els.n,max_els.nonzero,max_els.memory);
  printf("Maxumum matrix-vector multiplies: %i\n",max_els.multiply);
  printf("Memory used to construct bases: " SIZE_T_PRINT " bytes\n",
	 reduced_basis_memory);
#endif
  return 0;
}


/*****************************************************************

  The following routine calculates the fit criteria for
  a given set of fit parameters.  It returns the vector
  fvecc which is then squared to give the chi^2.

  The values that are calculated are reused from previous
  iterations if the relevant couplings are unchanged.
  This includes:

         pextrap, wfitting, lfitting, wvalue1, lvalue1,
	 lvalue2, spectrumvals1, spectrumvals2, stillvals
         
  Information about the spectra are returned in the structure
  best.  The members of the structure best 
  are from the lowest chi^2 iteration of fcn (which is
  not necessarily the last iteration of fcn.)

  On error, fcn sets *iflag to a negative integer.
  *flag<=-8 are "fatal" errors.

*******************************************************************/


extern fsub fcn(integer *m, integer *npar, doublereal *xp, 
	       doublereal *fvecc, integer *iflag){
  /* fitc is the index for the fit criterion */
  int i,k,j,l,fitc;
  int count_it;
  integer th=0;
  element *vals=NULL; /* this is realloc'ed by the extrapolation routines */
  element sum,temp,rescale,c2[NP1+NP2][NPERP][NPERPVALS];
  static element bestchi2=0;
  static doublereal spectrumvals1[NP1+1][NPERP][NPERPVALS],
    spectrumvals2[NP2+1][NPERP][NPERPVALS],
    stillvals[NSTILL][NSTILLVALS];
  element log_k_vals[N_LOG_VALS][MAXTH];
  static element pextrap;
#ifndef BIG
  /* use glueball results for these parameters  */
  element glueball_rescale=7.563874923;
  element wfitting[MAXTH]={0.280322, -0.531459, -0.246649};
  element lfitting[MAXTH]={0.148495, -0.166355, 0.340103};
#else
  /* use glueball results for these parameters  */
  element glueball_rescale=8.864492516;
  element wfitting[MAXTH]={0.266230,-0.684240,-0.108548};
  element lfitting[MAXTH]={0.127505, -0.234912, 0.471316};
#endif
  clock_t tif;
  time_t clock1,clock2;

  element *all_vals[N_ALL_BASES];
#if defined(USE_MPI) &&  GROUP_SIZE>0
#else
  int job;
  full_basis basis={0};
  eigensystem eigen=INITIAL_EIGENSYSTEM;
#endif
#if defined(USE_MPI) && GROUP_SIZE==0
  int nslaves;
  MPI_Comm_size(MPI_COMM_WORLD,&nslaves);
#endif
  tif=clock();
  time(&clock1);

  for(i=0; i<*npar; i++)par[wpar[i]]=xp[i];

#if PRINTIT
  printf("  Parameters: ");
  for(i=0; i<NPARAMS; i++)printf(" %.10g",par[i]);
  printf("\n");
#endif


  /* allocate memory to store eigenvalues */

  for(i=0; i<N_ALL_BASES; i++)
    all_vals[i]=malloc(all_bases[i].nval*sizeof(element));


#if defined(USE_MPI) && GROUP_SIZE>0

  /********************* use master-slave approach ******************/

  if(master(all_bases,par,all_vals,N_ALL_BASES)){
     fprintf(stderr,__FILE__":  master() is returning an error, exiting\n");
    *iflag=-9;
    return SUBRETURN;
  }
#else  /**************** calculate spectra in parallel ****************/

  for(job=0; job<N_ALL_BASES; job++){
#if defined(USE_MPI) && GROUP_SIZE==0
    for(i=0; i<nslaves; i++){
      if(i==WORLD_MASTER)continue;
#if 0  /* debug print */
      printf("job=%i slave=%i\n",job,i);
#endif
      MPI_Send(par,NPARAMS,MPI_DOUBLE,i,job,MPI_COMM_WORLD);
      MPI_Send(all_bases+job,1,mpi_specify_calculation,i,job,
	       MPI_COMM_WORLD);
    }      
#endif
    spectrum(all_bases+job,par,&eigen,&basis);
    for(i=0; i<all_bases[job].nval; i++)
      all_vals[job][i]=eigen.values[i];
  }
#endif
  /******** take results and calculate extrapolations etc. **************/


  /* Calculate the convergence in p-truncation */
#if PNS
  pextrap=pextrapolate(PNS,ps1,all_vals+(ps1-all_bases));
#else
  pextrap=-1.764202771;
#endif


#if N_LOG_VALS && LOG_NS
  th=-1;  /* flag to include log(K) term in fit */
  extrapolate(&vals,LOG_NS,log_base,all_vals+(log_base-all_bases),
	      &th,N_LOG_VALS,&pextrap);
    for(k=0; k<N_LOG_VALS; k++)
      for(i=0; i<th; i++)log_k_vals[k][i]=vals[k+i*N_LOG_VALS];
#endif


  /*  Calculate the transverse string tension.  The wht matrix
      must be reformatted into a form compatable with
      the FORTRAN routince WWEIGHTS.    */
#if NWIND
  wweights(weights,&wth,nwind,wht0);
  PSEUDOINVERSE(weights,&wth,&nwind);
  for(j=0; j<wth; j++)wfitting[j]=0.0;
  for(i=0; i<nwind; i++){
    extrapolate(&vals,WNS,ws0[i],all_vals+(ws0[i]-all_bases),&th,1,&pextrap);
    /* winding n=(2,0) eigenvalue saved */ 
    if(dot(wht0[i],wht0[i],HINDEX)==4)wvalue20=vals[0];
    for(j=0; j<wth; j++)
      wfitting[j] += vals[0]*weights[i+j*nwind];
  }
#else
  wth=3;
#endif

  /*  Calculate the longitudinal string tension  */
#if NLONG
  lweights(weights,&lth,nlong,mlong0);
  PSEUDOINVERSE(weights,&lth,&nlong);
  for(j=0; j<lth; j++)lfitting[j]=0.0;
  for(i=0; i<nlong; i++){
    lextrapolate(&vals,LNS,ls0[i],all_vals+(ls0[i]-all_bases),
		  &th,1,&pextrap,lkmax0);
    for(j=0; j<lth; j++)
      lfitting[j]+=vals[0]*weights[i+j*nlong];
    }
#else
  lth=3;
#endif

  /*  Calculate the ordinary extrapolated spectra. */
  for(i=0; i<NSTILL; i++){
    extrapolate(&vals,NS,stillbases[i],all_vals+(stillbases[i]-all_bases),
		 &th,NSTILLVALS,&pextrap);
    for(k=0; k<NSTILLVALS; k++)stillvals[i][k]=vals[k];
  }
  
  /*  Calculate the extrapolated spectrum in the P_perp = (c,0) direction */
  /*  Sort states before adding (and calculate 1 extra value) */
  for(l=0; l<NP1; l++)
    for(i=0; i<NPERP; i++){
      extrapolate(&vals,NS,s[l][i],all_vals+(s[l][i]-all_bases),
		   &th,NPERPVALS+1,&pextrap);
      /* sort and add states: we assume NPERPVALS+1 eigenvalues */
      k=0;
      addterms(spectrumvals1[l+1][i],&k,NPERPVALS,vals,NPERPVALS+1);
#if 0 /* debugging print */
      printf("  1:  l=%i, spectrum %i\n",l,i);
#endif
    } 
  
  /*  Calculate the extrapolated spectrum in the P_perp = (c,c) direction */
  /*  Sort states before adding (and calculate 1 extra value) */
  for(l=0; l<NP2; l++)
    for(i=0; i<NPERP; i++){
      extrapolate(&vals,NS,t[l][i],all_vals+(t[l][i]-all_bases),
		   &th,NPERPVALS+1,&pextrap);
      /* sort and add states: we assume NPERPVALS+1 eigenvalues */
      k=0;
      addterms(spectrumvals2[l+1][i],&k,NPERPVALS,vals,NPERPVALS+1);
#if 0 /* debugging print */
      printf("  2:  l=%i spectrum %i\n",l,i);
#endif
    }

  /* Free up memory used.*/
#if !(defined(USE_MPI) && GROUP_SIZE>0)
  free_eigensystem(&eigen);
  free_basis(&basis);  /* not necessary if we don't use eigenvectors */
#endif
  for(i=0; i<N_ALL_BASES; i++)free(all_vals[i]);

  /******************************************************************/

  /*  Put ordinary spectrum values into nonzero P_perp array*/
  /*  Here we sort the states before adding.  */
  /*  The ordering of the states must be compatable with */
  /*  the ordering obtained when calculating nonzero p_perp */
  for(i=0; i<NPERP; i++){
    for(j=0, k=l=0; j<NSTILL; j++){
      if(stillbases[j]->charge!=chargest[i])continue;
      /* assume that the two spin 1 multiplets are degenerate */
      if(((stillbases[j]->multi==3 || stillbases[j]->multi==5 ||
	    stillbases[j]->multi==7) && multis[i]==15) ||
	  ((stillbases[j]->multi==4 || stillbases[j]->multi==6 ||
	    stillbases[j]->multi==7) && multis[i]==16))
	addterms(spectrumvals1[0][i],&k,NPERPVALS,stillvals[j],NSTILLVALS);
     if(((stillbases[j]->multi==3 || stillbases[j]->multi==6 ||
	    stillbases[j]->multi==7) && multit[i]==17) ||
	  ((stillbases[j]->multi==4 || stillbases[j]->multi==5 ||
	    stillbases[j]->multi==7) && multit[i]==18))
	addterms(spectrumvals2[0][i],&l,NPERPVALS,stillvals[j],NSTILLVALS);
    }
    if(k!=NPERPVALS || l!=NPERPVALS){
       fprintf(stderr,__FILE__":%i  Incorrect multiplet mapping from states "
	       "in ordinary\n"
	      "spectrum to those in nonzero P_perp, exiting\n",__LINE__);
       fprintf(stderr,"          k=%i, l=%i and NPERPVAL=%i\n",
	       k,l,NPERPVALS);
      *iflag=-12;
      return SUBRETURN;
    }	
  }
    
#if 0 /*debugging print of the entire spectrum */
  for(i=0; i<NPERP; i++){
    for(k=0; k<NPERPVALS; k++){
      for(j=0; j<NP1+1; j++)
	printf(" %g",spectrumvals1[j][i][k]);
      printf("    ");
    }
    printf("\n");
  } 
  for(i=0; i<NPERP; i++){
    for(k=0; k<NPERPVALS; k++){
      for(j=0; j<NP2+1; j++)
	printf(" %g",spectrumvals2[j][i][k]);
      printf("    ");
    }
    printf("\n");
  } 
#endif
      
  /* There are several possible methods for setting
     the overall scale of the Hamiltonian  */
  
  switch(scalemethod){         
  case 0: 
    rescale=glueball_rescale; /* Get scale from glueball results */
    break;
  case 1:
    /* Find scale by minimizing overall chi^2 */
    /* This does not include the consistancy of the L nonzero,
       n=1 datum because it is too complicated */
    /* initial guess is from glueball data */
    rescale=glueball_rescale;
    for(count_it=0; ; count_it++){
      element sum=0.0, zz=0.0;
      for(j=0; j<NPERP; j++)
	for(i=0; i<FCSQUARED && i<NPERPVALS; i++){
	  for(k=0; k<NP1; k++){
	    c2[k][j][i]=wfitting[0]*pow(rescale,2)*
	      (spectrumvals1[k+1][j][i]-spectrumvals1[0][j][i])/
	      fdot(p1a[k],p1a[k],HINDEX);
	    temp=2.0*c2[k][j][i]/rescale;
	    sum += temp*(c2[k][j][i]-1.0)/pow(c2err[j][i],2);
	    zz +=  pow(temp/c2err[j][i],2);
	  }
	  for(k=0; k<NP2; k++){
	    c2[k+NP1][j][i]=wfitting[0]*pow(rescale,2)*
	      (spectrumvals2[k+1][j][i]-spectrumvals2[0][j][i])/
	      fdot(p2a[k],p2a[k],HINDEX);
	    temp=2.0*c2[k+NP1][j][i]/rescale;
	    sum += temp*(c2[k+NP1][j][i]-1.0)/pow(c2err[j][i],2);
	    zz +=  pow(temp/c2err[j][i],2);
	  }
	}
    /* Add fit to the experimental value for the mesons */
    for(j=0; j<FIX_MASSES; j++)
      for(i=0; i<NSTILL; i++)
	if(stillbases[i]->multi==experiment[j].multi && 
	   stillbases[i]->charge==experiment[j].charge){
	  sum += stillvals[i][experiment[j].state]*
	    (rescale*stillvals[i][experiment[j].state]-
	     experiment[j].value)/pow(experiment[j].error,2);
	  zz += pow(stillvals[i][experiment[j].state]/experiment[j].error,2);
	  break;
	}
#if FSCALE
    sum += lfitting[0]*(rescale*lfitting[0]-1.0)/pow(lerr0,2);
    zz += pow(lfitting[0]/lerr0,2);
#endif
    temp=rescale;
    /* Here I include a damping factor which will
       slow convergence but hopefully stablize it. */
    rescale -= 0.25*sum/zz;
#if 0 /* debugging print */
      if(j<100)printf("rescale:  j=%i rescale=%.16g\n",j,rescale);
#endif
      /* converge as well as we can, assuming round-off errors */
      if(fabs(temp-rescale)<=4*(1+FCSQUARED)*ELEMENT_EPSILON)break;
      if(count_it>1000){
	 fprintf(stderr,__FILE__":  Scale determination not converged, "
		 "rescale=%g,\n        use glueball results",rescale);
	 rescale=glueball_rescale; /* Get scale from glueball results */
	break;
      }
    }
    break;
  default:
    fprintf(stderr,__FILE__":  in fcn, scale method=%i is wrong.\n",
	    scalemethod);
    *iflag=-14;
    rescale=1.0;
  }
  /* If rescale is negative for some reason, 
     set it to something large and positive.  */
  if(rescale<0.0){
    rescale=50.0;
     fprintf(stderr,__FILE__":  rescale is negative, set recale=%f.\n",
	     rescale);
  }

  /**********************************************************************/

  fitc=0;  /* start counting fit criteria */

  /*  Calculate the desired dispersion using the overall scale 
      and lattice spacing from glueball calculations.
      Find error of speed of light squared
      with coupling g^2 N and "a" as determined above.
      One can weight the value for our value of sigma. */

  for(i=0; i<NPERP; i++)
    for(j=0; j<NPERPVALS; j++)
      for(k=0; k<NP1; k++){
	temp=wfitting[0]*pow(rescale,2)/fdot(p1a[k],p1a[k],HINDEX);
	c2[k][i][j]=(spectrumvals1[k+1][i][j]-spectrumvals1[0][i][j])*temp;
	if(j<FCSQUARED)fvecc[fitc++] = (c2[k][i][j]-1.0)/c2err[i][j];
      }
  for(i=0; i<NPERP; i++)
    for(j=0; j<NPERPVALS; j++)
      for(k=0; k<NP2; k++){
	temp=wfitting[0]*pow(rescale,2.0)/fdot(p2a[k],p2a[k],HINDEX);
	c2[k+NP1][i][j] = 
	  (spectrumvals2[k+1][i][j]-spectrumvals2[0][i][j])*temp;
	if(j<FCSQUARED)fvecc[fitc++] = (c2[k+NP1][i][j]-1.0)/c2err[i][j];
      }

  /* Add parity doublets to fitting criterion using fractional
     difference.   */

  for(k=0; k<FPARITY; k++){
    for(i=0; i<NSTILL; i++)
      if(stillbases[i]->multi==whichparity[k][0].multi && 
	 stillbases[i]->charge==whichparity[k][0].charge)break;
    for(j=0; j<NSTILL; j++)
      if(stillbases[j]->multi==whichparity[k][1].multi && 
	 stillbases[j]->charge==whichparity[k][1].charge)break;
    if(i<NSTILL && j<NSTILL)
      fvecc[fitc++]=(stillvals[i][whichparity[k][0].state]-
		     stillvals[j][whichparity[k][1].state])/
	(0.5*(stillvals[i][whichparity[k][0].state]+
	      stillvals[j][whichparity[k][1].state])*parityerr[k]);
    else {
      fprintf(stderr,__FILE__":  Incorrect multiplets in ordinary spectra, "
	      " exiting\n");
      *iflag=-15;
      return SUBRETURN;
    }
  }

  /* Add fit to the experimental value for the mesons */
  for(j=0; j<FIX_MASSES; j++)
    for(i=0; i<NSTILL; i++)
      if(stillbases[i]->multi==experiment[j].multi && 
	 stillbases[i]->charge==experiment[j].charge){
	fvecc[fitc++]=(rescale*stillvals[i][experiment[j].state]-
		       experiment[j].value)/experiment[j].error;
	break;
      }

    /* Demand the size of the log(K) term be small compared to
       finite part. */
  temp=0.5*(element) log_base[1].kt;  /* typical value of K */
  for(i=0; i<FEXTRA && i<N_LOG_VALS && LOG_NS; i++)
    fvecc[fitc++]=10.0*log_k_vals[i][0]*log(temp)/log_k_vals[i][1];

  /* If the lowest state in any sector is tachyonic, modify the 
     c^2 errors to reflect this */

  for(i=0; i<NSTILL; i++)
    if(stillvals[i][0]<0.0){
      fprintf(stderr,__FILE__":  Lowest state gone tachyonc.\n");
      fvecc[i%fitc]=100.0*fabs(stillvals[i][0]);
    }
    
  /* after including all fit criteria, make sure we 
     have actually filled the fvecc vector. */
  if(fitc!=*m){
    fprintf(stderr,__FILE__":  Fit criteria %i wrong, should be %li\n",
	    fitc,*m);
    *iflag=-75;
    return SUBRETURN;
  }
    
  /*
    check if this is the best chi^2 so far,
    if so, copy values into the structure best.
  */

  sum=fdot(fvecc,fvecc,*m);
  if(*iflag>=0 && (new_best || sum<bestchi2)){
    new_best=0; 
    bestchi2=sum;
    best.pextrap=pextrap; 
    best.rescale=rescale;
    for(i=0; i<NPERP; i++)
      for(j=0; j<NPERPVALS; j++)
	for(k=0; k<NP1+NP2; k++)
	  best.c2[k][i][j]=c2[k][i][j];
    for(i=0; i<NPERP; i++)
      for(k=0; k<NPERPVALS; k++){
	for(j=0; j<NP1+1; j++)
	  best.spectrumvals1[j][i][k]=spectrumvals1[j][i][k];
	for(j=0; j<NP2+1; j++)
	  best.spectrumvals2[j][i][k]=spectrumvals2[j][i][k];
      }
    for(i=0; i<NSTILL; i++)
      for(j=0; j<NSTILLVALS; j++)
	best.stillvals[i][j]=stillvals[i][j];
    for(i=0; i<wth; i++)best.wfitting[i]=wfitting[i];
    for(i=0; i<lth; i++)best.lfitting[i]=lfitting[i];
    for(i=0; i<N_LOG_VALS && LOG_NS; i++)best.extra[i]=log_k_vals[i][0];
  }

#if PRINTIT
  printf("  fvecc values =");
  for(i=0, k=3; i<*m; i++){
    printf(" "PDIGITS,fvecc[i]);
    if((k++)%6==0 && i+1<*m)printf("\n  ");
  }
  time(&clock2);
  printf("\n  Chi^2=%g in %.2f CPU seconds (%li wall).\n\n",
	 fdot(fvecc,fvecc,*m),
	 (float)(clock()-tif)/(float) CLOCKS_PER_SEC,clock2-clock1);
  fflush(stdout);
#endif

  free(vals); /* vals is realloc()ed by extrapolation routines */

  return SUBRETURN; /* normal return from subroutine */

}

/*  
    Interface function for fcn()
    returns iflag=0 if there is no error
    else, reterun iflag from fcn()
*/

extern fsub fcn3(integer *m, integer *npar, doublereal *xp, integer *nf,
		 doublereal *fvecc, integer *iflag, doublereal *ur, fsub *uf){
  *iflag=0;
  fcn(m,npar,xp,fvecc,iflag);
  if(*iflag<-8)*nf=0;  /* errors with iflag<-8 are "fatal" */ 
  return SUBRETURN;  /* normal return from subroutine */
}

#endif /* HINDEX==2 */

