#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "observable.h"
#ifdef USE_MPI
#include "usempi.h"
#endif

/*
  calculate probablilites for
  operators diagonal in the FOCK space basis.
	
  to add a new operator:
    1. Add to the "enum observable" list in observable.h
    2. Add a name to the print strings in observable_string below.
    3. Add to both switch() statements in find_observable below.
*/

void find_observable(
		     int which_state, /* which eigenstate, starting with 0 */
		     observable_result *x, /* result of calculation */
		     eigensystem *eigen, /* the results for the eigensystem */
		     full_basis *basis   /* pointer to the basis created */
		){
  int i;
  int quark_k,link_k;
  doublereal temp;
  partype *bpoint;
  /* The beginning basis element for the portion of the 
     eigenvector stored by this processor.
     In the non-mpi case, this a dummy variable. */
  int displ=0;  
#ifdef USE_MPI
  doublereal *result;
  int *recvsendarray;
  int my_id, comm_size;
  MPI_Comm_rank(mpi_group, &my_id);
  MPI_Comm_size(mpi_group, &comm_size);
  recvsendarray=malloc(comm_size*sizeof(int));
  i=eigen->vectors.length;
  MPI_Allgather(&i, 1, MPI_INT, recvsendarray, 1, MPI_INT, mpi_group);   
  for(i=0; i<my_id; i++)
    displ+=recvsendarray[i];
#endif

  /* set aside memory */

  switch(x->observable){
  case O_IDENTITY:
    x->length=1;
    break;
  case O_NUMBER:
    x->length=basis->nptrunc+1;
    break;
  case O_QUARK_DISTRIBUTION:
    if(basis->type!=TYPE_MESON){
      fprintf(stderr,__FILE__":  need meson to find quark distribution\n");
      goto end;
    }
    x->length=basis->kt/2;
    break;
  case O_QUARK_LINK_DISTRIBUTION:
    if(basis->type!=TYPE_MESON){
      fprintf(stderr,__FILE__":  need meson to find quark distribution\n");
      goto end;
    }
    x->length=basis->kt/2*basis->kt/2;
    break;
  default:
    x->length=0;
    fprintf(stderr,__FILE__":  %i not a valid observable index\n",
	    x->observable);
  }
  x->result=realloc(x->result,x->length*sizeof(*(x->result)));
  for(i=0; i<x->length; i++)x->result[i]=0.0;


  /* 
     now, go through the basis and add the weight for each state 
  */
  
  bpoint=basis->s+basis->nptrunc*displ;
  for(i=0; i<eigen->vectors.length; i++, bpoint+=basis->nptrunc){
    if(eigen->vectors.r!=NULL){
      temp=pow(eigen->vectors.r[which_state*eigen->vectors.length+i],2);
    }
    else {
      temp=pow(eigen->vectors.c[which_state*eigen->vectors.length+i].re,2)+
	pow(eigen->vectors.c[which_state*eigen->vectors.length+i].im,2);
    }

    switch(x->observable){
    case O_IDENTITY:
      x->result[0]+=temp;
      break;
    case O_NUMBER:
      x->result[basis->np[i+displ]]+=temp;
      break;
    case O_QUARK_DISTRIBUTION:
      if(basis->charge || basis->o){
	temp*=0.5;
	x->result[KMASK&(bpoint[basis->np[i+displ]-1])]+=temp;
      }
      x->result[KMASK&(bpoint[0])]+=temp;
      break;
    case O_QUARK_LINK_DISTRIBUTION:
      if(basis->np[i+displ]<3)break; /* can't measure if there is no link */
      if(basis->charge || basis->o){
	temp*=0.5;
	quark_k=KMASK&(bpoint[basis->np[i+displ]-1]);
	link_k=KMASK&(bpoint[basis->np[i+displ]-2]);
	x->result[quark_k*(basis->kt/2)+link_k]+=temp;
      }
      quark_k=KMASK&(bpoint[0]);
      link_k=KMASK&(bpoint[1]);
      x->result[quark_k*(basis->kt/2)+link_k]+=temp;
      break;

    default:
      fprintf(stderr,__FILE__":  %i not a valid observable index\n",
	      x->observable);

    }

  }

#ifdef USE_MPI
  /* have auxilliary array because MPI can handle the input
     and output arrays being the same */
  result=malloc(x->length*sizeof(*result));
  for(i=0; i<x->length; i++)result[i]=x->result[i];
  MPI_Allreduce(result,x->result, x->length, MPI_DOUBLE, 
		 MPI_SUM, mpi_group);
  free(result);
#endif

 end:
#ifdef USE_MPI
  free(recvsendarray); 
#endif
  return;
}

/*
  print an array of double numbers as a Mathematica formatted list 
  x has length n
*/

void print_real_list(FILE *fp,element *x,int n){
  int i;
  fprintf(fp,"{");
  for(i=0; i<n; i++){
    if(i)fprintf(fp,",");
    if(i%6==5)fprintf(fp,"\n  ");
    fprintf(fp,"%.12f",x[i]);
  }
  fprintf(fp,"};\n");
}

void print_complex_list(FILE *fp,complex_element *x,int n){
  int i;
  fprintf(fp,"{");
  for(i=0; i<n; i++){
    if(i)fprintf(fp,",");
    if(i%3==2)fprintf(fp,"\n  ");
    fprintf(fp,"%.12f%+.12f*I",x[i].re,x[i].im);
  }
  fprintf(fp,"};\n");
}

/* free memory used used to store wavefunction results */

void free_observable_result(observable_result *result){
  result->observable=-1;
  result->length=0;
  free(result->result);
  result->result=NULL;
}

/* string containing operator names */

const char *observable_string(enum observable t){
  static char string[4][50]={"the identity operator",
			     "the number operator","quark distribution",
			     "quark-link distribution"};
  return string[t];
}
