#include <stdio.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include "usempi.h"
#undef GROUP_SIZE /* This should be passed as a function argument */

#define MS_PRINT 0 /* flag to print out debugging information to file */
                   /* Each process gets a unique file */

#ifdef _CRAY  /* this is for the mallopt() command below */
#include <malloc.h>
#endif

/* 
   This subroutine initializes the data structures and
   communication groups used by mpi
*/

void init_mpi_types(int group_size){
  int i;
#define SPECIFY_SIZE 11
  int len[SPECIFY_SIZE];
  MPI_Datatype type[SPECIFY_SIZE];
  specify_calculation mx;
  MPI_Aint disp[SPECIFY_SIZE],base;
  MPI_Group g1,g2;
  int master[1]={WORLD_MASTER};

  /* make MPI datatype for the structure specify_calculation */

  MPI_Address(&mx.type,disp);          len[0]=1;        type[0]=MPI_INT;
  MPI_Address(&mx.ht,disp+1);          len[1]=HINDEX;   type[1]=MPI_INT;
  MPI_Address(&mx.nptrunc,disp+2);     len[2]=1;        type[2]=MPI_UNSIGNED;
  MPI_Address(&mx.kt,disp+3);          len[3]=1;        type[3]=MPI_UNSIGNED;
  MPI_Address(&mx.charge,disp+4);      len[4]=1;        type[4]=MPI_INT;
  MPI_Address(&mx.o,disp+5);           len[5]=1;        type[5]=MPI_INT;
  MPI_Address(&mx.multi,disp+6);       len[6]=1;        type[6]=MPI_INT;
  MPI_Address(&mx.momflag,disp+7);     len[7]=1;        type[7]=MPI_INTEGER;
  MPI_Address(&mx.angle,disp+8);       len[8]=HINDEX+2; type[8]=MPI_DOUBLE;
  MPI_Address(&mx.nval,disp+9);        len[9]=1;        type[9]=MPI_INTEGER;
  MPI_Address(&mx.vectorflag,disp+10); len[10]=1;       type[10]=MPI_INTEGER;

  /* we find offsets relative to the "base" of the structure,
     wherever that may be (thus we make no assumptions of order. */
  MPI_Address(&mx,&base);
  for(i=0; i<SPECIFY_SIZE; i++)disp[i]-=base;
  MPI_Type_struct(SPECIFY_SIZE,len,disp,type,&mpi_specify_calculation);
  MPI_Type_commit(&mpi_specify_calculation);
  
/*
  Initialize the communicator mpiallslaves which
  is a subgroup consisting of just the slaves.
  This is used in file "improved.c"  
*/

  if(group_size==0)
    mpiallslaves=MPI_COMM_WORLD;
  else {
    MPI_Comm_group(MPI_COMM_WORLD,&g1);
    MPI_Group_excl(g1,1,master,&g2);
    MPI_Comm_create(MPI_COMM_WORLD,g2,&mpiallslaves);
    MPI_Group_free(&g2);
    MPI_Group_free(&g1);
  }
}

/* This is the slave process used by MPI 
   it is a "stand-alone" subroutine to calculate the
   lowest eigenvalues for a given basis.  It also
   calculates the basis when necessary. 
   angle contains HINDEX components of transverse momentum,
   the kmax cutoff for nonzero L and L itself.
*/

#define MAXBASIS 150

void slave(int group_size){
  specify_calculation a;
  full_basis basis={0};
  eigensystem eigen=INITIAL_EIGENSYSTEM;
  int world_rank,group_rank=0;
  element couplings[NPARAMS];
#if MS_PRINT 
  FILE *fp;
  char out[50];
  int i,lastjob=0;
#else
  FILE *fp=stderr;  /* send error messages to stderr */
#endif  
  MPI_Status status, status2;

  if(group_size==1)
    group_rank=0;
  else
    MPI_Comm_rank(mpi_group,&group_rank);
  MPI_Comm_rank(MPI_COMM_WORLD,&world_rank);

#if MS_PRINT /* open debugging file with name based on rank */
  sprintf(out,"aslave%i.tmp",world_rank);
  if((fp = fopen(out,"w")) == NULL){
     fprintf(stderr,__FILE__":  Can't open file %s, exiting\n",out);
    goto error;
  }
#endif

  for(;;){        /* Main loop of program */
    
    MPI_Probe(WORLD_MASTER,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
    if(status.MPI_TAG==TAG_OFFSET)goto exit;    /* Stop program */
    else{      /* Calculate spectrum */  
#if MS_PRINT /* if starting new calculations, write over old info */
      if(status.MPI_TAG<lastjob)rewind(fp);
      lastjob=status.MPI_TAG;
      fprintf(fp,"About to receive Job %i\n",status.MPI_TAG);
#endif
      MPI_Recv(couplings,NPARAMS,MPI_DOUBLE,WORLD_MASTER,status.MPI_TAG,
	       MPI_COMM_WORLD,&status2);
      MPI_Recv(&a,1,mpi_specify_calculation,WORLD_MASTER,status.MPI_TAG,
	       MPI_COMM_WORLD,&status2);
      if(status2.MPI_ERROR!=MPI_SUCCESS)goto error;
#if MS_PRINT /* diagnostic print to debugging file */
      fprintf(fp,"Received job %i\n",status.MPI_TAG);
      fprintf(fp,"  %s spectrum:  h=(",state_type_string(a.type));
      for(i=0; i<HINDEX; i++)fprintf(fp," %i",a.ht[i]);
      fprintf(fp,"), %i-particles, K=%i/2, charge=%i, o=%i, multi=%i",
	     a.nptrunc,a.kt,a.charge,a.o,a.multi);
      if(a.momflag){
	if(a.type==TYPE_SOURCE)
	  fprintf(fp,"\n      (%.12g %.12g)",a.angle[HINDEX],a.angle[HINDEX+1]);
      else
	fprintf(fp,"\n      (%.12g %.12g)",a.angle[0],a.angle[1]);
      }
      fprintf(fp,"\n");
      fflush(fp);
#endif

      spectrum(&a,couplings,&eigen,&basis);

      if(eigen.values==NULL){
	fprintf(fp,__FILE__":  error in spectrum(...)\n");
	goto error;
      }

#if MS_PRINT /* diagnositic print */
      fprintf(fp,"  values: ");
      for(i=0; i<a.nval; i++)fprintf(fp,"%.10g ",eigen.values[i]);
      fprintf(fp,"\n");
      fflush(fp);
#endif

      /* We don't do anything with any eigenvectors */
      if(group_rank==0)
	MPI_Send(eigen.values,a.nval,MPI_DOUBLE,WORLD_MASTER,
		 status.MPI_TAG,MPI_COMM_WORLD);
    }
  }
    
error:  /* On error, print out memory usage and tell the master. */
/* The master will then send out a command that everyone quit. */
   fprintf(fp,__FILE__":  Error in slave %i (group rank %i) after "
	   "receiving job %i, exiting.\n",
	   world_rank,group_rank,status.MPI_TAG);
   fflush(fp);
  /* Just the message tag is read by the master */
  MPI_Bsend(&world_rank,1,MPI_INT,0,0,MPI_COMM_WORLD);
  
exit:  /* Normal exit */
  free_eigensystem(&eigen);
  free_basis(&basis);
#if MS_PRINT /* debugging file with name based on rank */
  fclose(fp);
#endif
  return;
}

/*  
    This subroutine gives tasks to the various slaves. 
    Eigenvalues are output in the vector vals whose
    memory should be already allocated.  vals[whichvals[i]+j]
    contains the jth eigenvalue, j=0,1,2... for job i,
    i=1,2,...njob. 

    Since all the jobs have the same couplings,
    the generation of improved matrix elements is parallelized.

    master returns 0 on successful completion.
*/


int master(specify_calculation *base, element *couplings, 
	   element **vals, int njob){
  int job,receive,slave,nslaves,j;
  int ierr=0;  /* zero is normal return */
#if MS_PRINT /* diagnostic print to debugging file */
  FILE *fp;
  char out[50]="amaster.tmp";
#else
  FILE *fp=stderr;
#endif  
  MPI_Status status,status2;

  MPI_Comm_size(MPI_COMM_WORLD,&nslaves);
  nslaves--;   /* don't include master amongst slaves */
  

#if MS_PRINT /* open debugging file*/
  if((fp = fopen(out,"w")) == NULL){
     fprintf(stderr,__FILE__":  Can't open file %s, exiting\n",out);
    ierr+=1; goto end;
  }
#endif

  /* For the matrix elements to be generated in parallel,
     we must have more jobs than slaves */
  if(njob<nslaves){
    fprintf(fp,__FILE__":  need njob=%i >= nslaves=%i\n",njob,nslaves);
    ierr+=2; goto end;
  }

  /* loop through all jobs that need completion */
  for(job=0, receive=0; receive<njob;){

    /* Receive completed jobs once all slaves have been seeded. */
    if(job<nslaves)
      slave=job+1;  /* slaves are numbered 1,2,3,... */
    else{
#if 0  /* this may take up too much CPU when waiting */
      MPI_Probe(MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
#else 
      for(;;){
        MPI_Iprobe(MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&j,&status);
        if(j)break;
        usleep(1); /* cause process to sleep for one turn */
      }
#endif
      j=status.MPI_TAG;
      if(j>=njob){
	 fprintf(fp,__FILE__":  Master receiving tag=%i from slave %i, "
		 "exiting\n",status.MPI_TAG,status.MPI_SOURCE);
	 ierr+=4; goto end;
      }
      MPI_Recv(vals[j],base[j].nval,MPI_DOUBLE,status.MPI_SOURCE,
	       status.MPI_TAG,MPI_COMM_WORLD,&status2);
      if(status2.MPI_ERROR != MPI_SUCCESS){
	fprintf(fp,__FILE__":  master() receive error!\n");
	ierr+=8; goto end;
      }
      if(base[j].vectorflag){
	fprintf(fp,__FILE__":  master(...) can't receive vectors!\n");
	ierr+=16;  goto end;
      }
      slave=status.MPI_SOURCE;
      receive++;
    }

    /* send jobs to the slaves */
    if(job<njob){
#if MS_PRINT /* debugging print to file: print jobs sent to slaves. */
      fprintf(fp,"Master sending job=%i to slave %i\n",job,slave);
      fflush(fp);
#endif
      MPI_Send(couplings,NPARAMS,MPI_DOUBLE,slave,job,MPI_COMM_WORLD);
      MPI_Send(base+job,1,mpi_specify_calculation,slave,job,MPI_COMM_WORLD);
      job++;
    }
  }

end:
#if MS_PRINT /* open debugging file with name based on rank */
  fprintf(fp,"Done, extiting\n");
  fclose(fp);
#endif
  return ierr;  
}
