#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "spectrum.h"
#ifdef USE_MPI
#include "usempi.h"
#endif

#define PROB(A) (pow((A)->re,2)+pow((A)->im,2))

/*
  calculate probablilites for
  operators diagonal in the FOCK space basis.
	
  to add a new operator:
    1. Add to the "observable" list in spectrum.h
    2. Add a name to the print strings in observable_string below.
    3. Add to both switch() statements in find_observable below.
*/

void find_observable(
		     observable_result *x, /* result of calculation */
		     observable observable, /* the observable to calculate */
		     int which_state, /* which eigenstate, starting with 0 */
		     eigensystem *eigen, /* the results for the eigensystem */
		     full_basis *basis   /* pointer to the basis created */
		){
  int i;
  int j;
  int phase;
  partype rotstate[NPMAX];
  int quark_k,link_k;
  latype p;
  complex_element amplitude;
  complex_element ztemp;
  complex_element zsum={0.0,0.0};
  doublereal temp;
  partype *bpoint;
  unsigned char *nppoint;
  /* The beginning basis element for the portion of the 
     eigenvector stored by this processor.
     In the non-mpi case, this a dummy variable. */
  int displ=0;  
  const latlist *subg=latgroups+subgroup[basis->multi];
#ifdef USE_MPI
  int *recvsendarray;
  int my_id, comm_size;    
#endif

  if(observable==O_NONE || observable==O_VECTORS){
    x->length=0;
    return;
  }

  /* initialize MPI stuff */

#ifdef USE_MPI
  MPI_Comm_rank(mpi_group, &my_id);
  MPI_Comm_size(mpi_group, &comm_size);
  recvsendarray=malloc(comm_size*sizeof(int));
  i=eigen->vectors.length;
  MPI_Allgather(&i, 1, MPI_INT, recvsendarray, 1, MPI_INT, mpi_group);   
  for(i=0; i<my_id; i++)
    displ+=recvsendarray[i];
  free(recvsendarray); 
#endif

  /* set aside memory */

  switch(observable){
  case O_IDENTITY:
    x->length=1;
    break;
  case O_NUMBER:
    x->length=basis->nptrunc+1;
    break;
  case O_QUARK_DISTRIBUTION:
    if(basis->type==TYPE_MESON || basis->type==TYPE_BARYON){
      x->length=basis->kt/2;
    } else {
      fprintf(stderr,__FILE__":  need meson or baryon for quark distribution\n");
      goto end;
    }
    break;
  case O_QUARK_LINK_DISTRIBUTION:
    if(basis->type!=TYPE_MESON){
      fprintf(stderr,__FILE__":  need meson for quark distribution\n");
      goto end;
    }
    x->length=basis->kt/2*basis->kt/2;
    break;
  case  O_VALENCE_IDENTITY:
  case  O_VALENCE_SIGMA2:
  case  O_VALENCE_SIGMA3:
    if(basis->type!=TYPE_MESON){
      fprintf(stderr,__FILE__":  need meson for valence amplitude\n");
      goto end;
    }
    x->length=1;
    break;
  default:
    x->length=0;
    fprintf(stderr,__FILE__":  %i not a valid observable index\n",
	    observable);
  }
  x->result=realloc(x->result,x->length*sizeof(*(x->result)));
  if(x->result==NULL){
    fprintf(stderr,__FILE__":  allocate error\n");
    exit(32);
  }
  for(i=0; i<x->length; i++)x->result[i]=0.0;


  /* 
     now, go through the basis and add the weight for each state 
  */
#if 0
  printf(__FILE__":  basis length %li vector length %li\n",
	 basis->length,eigen->vectors.length);
#endif
  bpoint=basis->s+basis->nptrunc*displ;
  nppoint=basis->np+displ;
  for(i=0; i<eigen->vectors.length; 
      i++, bpoint+=basis->nptrunc, nppoint++){
    /* skip non-dynamical states */
    if(state_constrained(bpoint,basis->type,basis->np[i+displ])){
      continue;
    }
    if(eigen->vectors.r!=NULL){
      amplitude.re=eigen->vectors.r[which_state*eigen->vectors.length+i];
      amplitude.im=0.0;
    }
    else {
      amplitude.re=eigen->vectors.c[which_state*eigen->vectors.length+i].re;
      amplitude.im=eigen->vectors.c[which_state*eigen->vectors.length+i].im;
    }

    switch(observable){
    case O_IDENTITY:
      x->result[0]+=PROB(&amplitude);
      break;
    case O_NUMBER:
      x->result[*nppoint]+=PROB(&amplitude);
      break;
    case O_QUARK_DISTRIBUTION:
      if(basis->type==TYPE_BARYON){
	for(j=0; j<*nppoint; j++)
	  if(bpoint[j-1]==EPSILON)
	    x->result[KMASK&(bpoint[j])]+=PROB(&amplitude);
      } else if(basis->type==TYPE_MESON){
	temp=PROB(&amplitude);
	if(basis->charge || basis->o){
	  temp*=0.5;
	  x->result[KMASK&(bpoint[*nppoint-1])]+=temp;
	} 
	x->result[KMASK&(bpoint[0])]+=temp;
      } else if(basis->type == TYPE_HEAVY_LIGHT)
	x->result[KMASK&(bpoint[0])]+=PROB(&amplitude);
      break;
    case O_QUARK_LINK_DISTRIBUTION:
      if(*nppoint<3)break; /* can't measure if there is no link */
      temp=PROB(&amplitude);
      if(basis->charge || basis->o){
	temp*=0.5;
	quark_k=KMASK&(bpoint[*nppoint-1]);
	link_k=KMASK&(bpoint[*nppoint-2]);
	x->result[quark_k*(basis->kt/2)+link_k]+=temp;
      }
      quark_k=KMASK&(bpoint[0]);
      link_k=KMASK&(bpoint[1]);
      x->result[quark_k*(basis->kt/2)+link_k]+=temp;
      break;
    case  O_VALENCE_IDENTITY:
      if(*nppoint>2)break; /* only quark-antiquark sector */
      for(p=0; p<subg->length; p++){
	phase=statelat(rotstate,bpoint,basis->type,*nppoint,subg->e[p]);
	ZZPROD(&ztemp,&amplitude,&identity(rotstate[0],rotstate[1]));
	phasemultiply(&ztemp,phase+multiplet[basis->multi][p]);
	ZTIMES(&zsum,&ztemp,
	       1.0/sqrt((element) basis->inner[i+displ]));
      }
      break;
    case  O_VALENCE_SIGMA2:
      if(*nppoint>2)break; /* only quark-antiquark sector */
      for(p=0; p<subg->length; p++){
	phase=statelat(rotstate,bpoint,basis->type,*nppoint,subg->e[p]);
	ZZPROD(&ztemp,&amplitude,&sigma2(rotstate[0],rotstate[1]));
	phasemultiply(&ztemp,phase+multiplet[basis->multi][p]);
	ZTIMES(&zsum,&ztemp,
	       1.0/sqrt((element) basis->inner[i+displ]));
      }
      break;
    case  O_VALENCE_SIGMA3:
      if(*nppoint>2)break; /* only quark-antiquark sector */
      for(p=0; p<subg->length; p++){
	phase=statelat(rotstate,bpoint,basis->type,*nppoint,subg->e[p]);
	ZZPROD(&ztemp,&amplitude,&sigma3(rotstate[0],rotstate[1]));
	phasemultiply(&ztemp,phase+multiplet[basis->multi][p]);
	ZTIMES(&zsum,&ztemp,
	       1.0/sqrt((element) basis->inner[i+displ]));
      }
     break;

    default:
      fprintf(stderr,__FILE__":  %i not a valid observable index\n",
	      observable);
    }
  }

#ifdef USE_MPI  /* This must come before final rescaling */
  {
  /* have auxilliary array because MPI can't handle the input
     and output arrays being the same */
    doublereal *result=malloc(x->length*sizeof(*result));
    doublereal zsumin[2]={zsum.re,zsum.im}, zsumout[2];
    for(i=0; i<x->length; i++)result[i]=x->result[i];
    MPI_Allreduce(result,x->result, x->length, MPI_DOUBLE, 
		  MPI_SUM, mpi_group);
    MPI_Allreduce(zsumin,zsumout,2, MPI_DOUBLE, 
		  MPI_SUM, mpi_group);
    zsum.re=zsumout[0]; zsum.im=zsumout[1];
    free(result);
  }
#endif

  /* final rescaling */

  switch(observable){
  case  O_VALENCE_IDENTITY:
    /* b^dagger I d^dagger is odd under charge conjugation */
    x->result[0]=sqrt(PROB(&zsum)*(1-basis->charge)*2.0/
		      ((element) basis->kt*subg->length)); /* kt = 2*K */
    break;
  case  O_VALENCE_SIGMA2:
    /* b^dagger \sigma^2 d^dagger is even under charge conjugation */
    x->result[0]=sqrt(PROB(&zsum)*(1+basis->charge)*2.0/
		      ((element) basis->kt*subg->length)); /* kt = 2*K */
    break;
  case  O_VALENCE_SIGMA3:
    /* b^dagger \sigma^3 d^dagger is odd under charge conjugation */
    x->result[0]=sqrt(PROB(&zsum)*(1-basis->charge)*2.0/
		      ((element) basis->kt*subg->length)); /* kt = 2*K */
    break;
  default:
    break;
  }

 end:
#if 0 /* debug print */
  printf("Observable %i:  ",observable);
  for(i=0; i<x->length; i++)
    printf("%g ",x->result[i]);
  printf("\n");
#endif
  return;
}

/*
  print an array of double numbers as a Mathematica formatted list 
  x has length n.  The string s is appended after the list.
*/

void print_real_list(FILE *fp,element *x,int n,char *s){
  int i;
  fprintf(fp,"{");
  for(i=0; i<n; i++){
    if(i)fprintf(fp,",");
    if(i%6==5)fprintf(fp,"\n  ");
    fprintf(fp,"%.12f",x[i]);
  }
  fprintf(fp,"};%s",s);
}

void print_complex_list(FILE *fp,complex_element *x,int n){
  int i;
  fprintf(fp,"{");
  for(i=0; i<n; i++){
    if(i)fprintf(fp,",");
    if(i%3==2)fprintf(fp,"\n  ");
    fprintf(fp,"%.12f%+.12f*I",x[i].re,x[i].im);
  }
  fprintf(fp,"};\n");
}

/* free memory used used to store wavefunction results */

void free_observable_result(observable_result *result){
  result->length=0;
  free(result->result);
  result->result=NULL;
}

/* string containing operator names */

const char *observable_string(observable t){
  static char string[7][50]={
    "","",
    "the identity operator",
    "the number operator","quark distribution",
    "quark-link distribution","quark valence amplitude"};
  return string[t];
}
