#include <stdio.h>
#include "blocksparse.h"
#include "lapack.h"
#include "heap_sort.h"

#define DEBUG 0 /* put in some checks */

/*
  Turn a sparse symmetric matrix stored in comprssed row storage
  (CRS) format into a block-diagonal matrix.
  Finding the blocks is equivalent to finding the
  maximal disconnected subgraphs of a graph.
*/

int sparse_to_real_block_matrix(elslist *els, struct real_block_matrix *a){
  size_v row,col;
  size_v colp;  /* pointer to the current column in elslist */
  size_v block;
  int k;
  int row_count;
  integer err;
  real_block *blistp;
  size_v *which_block;
  char uplo='U';
  /* Initially, the matrix blocks are described as a linked list */
  struct {  
    size_v first;  /* first element in list */
    size_v last;   /* last element in list */
    size_v n;     /* length of list */
  } *block_row, *brp, *brp2;
  size_v *link_row_list; /* linked list of rows */

  /* 
     allocate memory and 
     mark each row as a separate subgraph 
  */  

  link_row_list=malloc(els->n*sizeof(*link_row_list));
  if(link_row_list==NULL){
    fprintf(stderr,__FILE__":%i\n",__LINE__);
    goto error;
  }
  block_row=malloc(els->n*sizeof(*block_row));
  if(block_row==NULL){
    fprintf(stderr,__FILE__":%i\n",__LINE__);
    goto error;
  }
  for(row=0; row<els->n; row++){
    block_row[row].n=1;
    block_row[row].first=row;
    block_row[row].last=row;
    link_row_list[row]=row;
  }
  which_block=malloc(els->n*sizeof(*which_block));
  if(which_block==NULL){
    fprintf(stderr,__FILE__":%i\n",__LINE__);
    goto error;
  }
  for(row=0; row<els->n; row++)which_block[row]=row;

  /* consolidate connected subgraphs */

  for(row=0; row<els->n; row++)
    for(colp=els->j[row]; colp<els->j[row+1]; colp++){
      col=els->i[colp];
      if(col==row)continue;
      if(which_block[row]!=which_block[col]){
        brp=block_row+which_block[row];
        brp2=block_row+which_block[col];
	link_row_list[brp->last]=brp2->first;
	brp->last=brp2->last;
	brp->n+=brp2->n;
	col=brp2->first;
	do{
	  which_block[col]=which_block[row];
	  if(col==brp2->last)break;
	  col=link_row_list[col];
	}while(1);
        brp2->n=0;
      }
    }

  free(which_block); which_block=NULL;

#if DEBUG /* print out block list */
  printf("row, link,    n, first, last\n");
  for(row=0; row<els->n; row++)
    printf("%5i %5i %5i %5i %5i\n",row,link_row_list[row],
	   (int)block_row[row].n,
	   block_row[row].first,block_row[row].last);
#endif
  
  /* allocate block matrix structure */
  
  a->largest_block_size=0;
  a->nblocks=0;
  for(block=0; block<els->n; block++){
    if(block_row[block].n>0)a->nblocks++; /* count non-empty blocks */
    if(block_row[block].n>a->largest_block_size)
      a->largest_block_size=block_row[block].n;
  }
  a->blist=malloc(a->nblocks*sizeof(*a->blist));
  if(a->blist==NULL){
    fprintf(stderr,__FILE__":%i\n",__LINE__);
    goto error;
  }

#if DEBUG 
  printf("largest:  %i  nblocks:  %i\n",(int) a->largest_block_size,
	 (int) a->nblocks);
#endif

  /* fill up block matrices */
  
  for(block=0, brp=block_row, blistp=a->blist; 
      block<a->nblocks; 
      brp++, blistp++, block++){
    while(brp->n==0)brp++;
    err=allocate_real_block(blistp,brp->n);
    if(err){
      fprintf(stderr,__FILE__":%i\n",__LINE__);
      goto error;
    }
    row=brp->first;
    k=0;
    do{
      blistp->row[k++]=row;
      if(row==brp->last)break;
      row=link_row_list[row];
    }while(k<brp->n);
    heapSort(blistp->row,blistp->n);
#if DEBUG
    printf("block %i length %i\n",(int) block,(int) blistp->n);
    for(k=0; k<blistp->n; k++)
      printf(" %i",blistp->row[k]);
    printf("\n");
#endif
    /* row_count is current row within this block */
    for(row_count=0; row_count<blistp->n; row_count++){
      int col_count; /* the current column within this block */
      row=blistp->row[row_count]; /* current row in original matrix */
      for(colp=els->j[row]; colp<els->j[row+1]; colp++){
	/* don't assume els->i is sorted */
	col_count=find_sorted(blistp->row,blistp->n,els->i[colp]);
#if DEBUG	
	/* error condition for find_sorted */
	if(col_count==blistp->n){
	  fprintf(stderr,__FILE__":%i\n",__LINE__);
	  goto error;
	}
	/* row_count <= col_count */
	if(row_count>col_count){
	  if(col_count==blistp->n){
	    fprintf(stderr,__FILE__":%i\n",__LINE__);
	    goto error;
	  }
	}
#endif
	blistp->a[row_count+col_count*(col_count+1)/2]=els->r[colp];
      }
    }
    /*
      CALL DSPTRF( UPLO, N, AP, IPIV, INFO )
    */
    DSPTRF(&uplo,&blistp->n,blistp->a,blistp->ipiv,&err);
    if(err){
      fprintf(stderr,__FILE__":  DSPTRF returning %i\n",(int) err);
      goto error;
    }
  }
  
  free(link_row_list); link_row_list=NULL;
  free(block_row); block_row=NULL;

  return 0;

 error:
  free(which_block);
  free(link_row_list);
  free(block_row);
  if(err==0)err=-128;

  return err;
  }

/*
  allocate memory for one block
*/

int allocate_real_block(real_block *b, integer n){
  int i;
  if(n<1)return -1;
  b->n=n;
  b->ipiv=malloc(n*sizeof(*(b->ipiv)));
  if(b->ipiv==NULL)return -2;
  b->a=malloc((n*(n+1))/2*sizeof(*(b->a)));
  for(i=0; i<(n*(n+1))/2; i++)b->a[i]=0.0;
  if(b->a==NULL)return -4;
  b->row=malloc(n*sizeof(*(b->row)));
  if(b->row==NULL)return -8;
  return 0;
}

/*
  free memory allocated for one block
*/

void free_real_block_matrix(struct real_block_matrix *a){
  size_v block;
  real_block *b=a->blist;

  for(block=0; block<a->nblocks; b++, block++){
    free(b->ipiv);
    free(b->a);   
    free(b->row);
  }
  free(a->blist); b=NULL;
  a->nblocks=0;
}
 
